//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

float			air_control = 1.0 / 256.0;

#define HEAL_RADIUS_DIST	255.0

#define TXT_USEPUZZLEFAILED			"YOU CANNOT USE THIS HERE"

void SetPlayerRunState(player_t* player);
boolean UndoPlayerMorph(player_t *player);
void MovePsprites(player_t *player);
void MorphPlayerThink(player_t *player);
void DeathPlayerThink(player_t* player);
mobj_t *SpawnSpeedMobj(mobj_t* pmo);
void ArtiPig(mobj_t *mo);
void ArtiSummon(player_t *player);
void ArtiPoisonBag(player_t *player);
void ArtiTeleportOther(player_t *player);
void ArtiBlastRadius(player_t *player);
void PlayerUseArtifact(player_t *player, int arti);
void PlayerNextArtifact(player_t *player);
void PlayerRemoveArtifact(player_t *player, int slot);
void P_PoisonDamage(player_t *player, mobj_t *source, int damage,
	boolean playPainSound);
void P_RemoveMobjFromTIDList(mobj_t *mobj);
boolean P_TeleportMove(mobj_t* thing, float x, float y);
boolean P_Teleport(mobj_t *thing, float x, float y, angle_t angle,
	boolean useFog);
boolean G_CheckSpot(int playernum, mthing_t* mthing);
boolean P_UsePuzzleItem(player_t *player, int itemType);
void P_UseLines(player_t* player);
int P_GetThingFloorType(mobj_t *thing);

mobj_t 		LavaInflictor;

boolean		onground;

// Y-adjustment values for full screen (4 weapons)
float PSpriteSY[NUMCLASSES * NUMWEAPONS] =
{
	0.0, -12.0, -10.0, 10.0, // Fighter
	-8.0, 10.0, 10.0, 0.0, // Cleric
	9.0, 20.0, 20.0, 20.0, // Mage
	10.0, 10.0, 10.0, 10.0 // Pig
};

//==========================================================================
//
//  MovePlayer
//
//==========================================================================

float	sv_maxspeed = 400.0;
float	sv_accelerate = 10.0;

void MovePlayer(player_t* player)
{
	int 		fly;
	float		forward;
	float		side;
	TVec		vforward;
	TVec		vright;
	TVec		vup;
	TVec		wishvel;

	forward = player->forwardmove;
	side = player->sidemove;

	//	Adjust speed depending on class. Originaly this is done in input
	// code, but to simplify it, we do it here.
	//	Fighter moves faster, cleric at original speed, mage - slower,
	// pig - slower, but faster than mage
	switch (player->pclass)
	{
	 case PCLASS_FIGHTER:
		forward *= 1.18;
		side *=	1.3;
		break;

	 case PCLASS_CLERIC:
		break;

	 case PCLASS_MAGE:
		forward *= 0.9;
		side *= 0.9;
		break;

	 case PCLASS_PIG:
		forward *= 0.97;
		side *= 0.97;
		break;
	}

	if (player->powers[pw_speed] && !player->morphTics)
	{
		// Adjust for a player with a speed artifact
		forward *= 1.5;
		side *= 1.5;
	}
	if (!(player->powers[pw_flight] && !(player->mo->origin.z <= player->mo->floorz)) &&
		P_GetThingFloorType(player->mo) == FLOOR_ICE) // Friction_Low
	{
		forward *= 0.5;
		side *= 0.5;
	}

    // Do not let the player control movement
    //  if not onground.
	onground = (player->mo->origin.z <= player->mo->floorz)
		|| (player->mo->flags2 & MF2_ONMOBJ);

	AngleVectors(&player->mo->angles, &vforward, &vright, &vup);

	wishvel = forward * vforward + side * vright;
	wishvel.z = 0.0;

	if (!(onground || player->mo->flags2 & MF2_FLY))
	{
		//	not on ground, so little effect on velocity
		forward *= air_control;
		side *= air_control;
	}
	player->mo->velocity += 5.0 * frametime * wishvel;

    if (forward || side)
    {
		SetPlayerRunState(player);
	}

	fly = ftoi(player->flymove) >> 4;
	if (fly > 7)
	{
		fly -= 16;
	}
	if (fly && player->powers[pw_flight])
	{
		if (fly != TOCENTER)
		{
			player->flyheight = fly*2;
			if (!(player->mo->flags2&MF2_FLY))
			{
				player->mo->flags2 |= MF2_FLY;
				player->mo->flags |= MF_NOGRAVITY;
				if (player->mo->velocity.z <= -39.0 * 35.0)
				{
					// stop falling scream
					S_StopSound(player->mo);
				}
			}
		}
		else
		{
			player->mo->flags2 &= ~MF2_FLY;
			player->mo->flags &= ~MF_NOGRAVITY;
		}
	}
	else if (fly > 0)
	{
		PlayerUseArtifact(player, arti_fly);
	}
	if (player->mo->flags2 & MF2_FLY)
	{
		player->mo->velocity.z = itof(player->flyheight) * 35.0;
		if (player->flyheight)
		{
			player->flyheight /= 2;
		}
	}
	if ((player->buttons & BT_JUMP) && onground && !player->jumpTics)
	{
		if (player->morphTics)
		{
			player->mo->velocity.z = 6.0 * 35.0;
		}
		else
		{
			player->mo->velocity.z = 9.0 * 35.0;
		}
		player->mo->flags2 &= ~MF2_ONMOBJ;
		player->jumpTics = 18;
	}
}

//==========================================================================
//
//	CheckWaterJump
//
//==========================================================================

void CheckWaterJump(player_t *player)
{
	TVec	start;
	TVec	end;
	TVec	vforward;

	// check for a jump-out-of-water
	AngleVector(&player->mo->angles, &vforward);
	start = player->mo->origin;
	start.z += player->mo->height * 0.5 + 8.0;
	vforward.z = 0.0;
	vforward = Normalize(vforward);
	end = start + vforward * 24.0;
	if (!TraceLine(start, end))
	{
		// solid at waist
		start.z = player->mo->origin.z + player->mo->height;
		end = start + vforward * 24.0;
		player->movedir = trace_plane_normal * -50.0;
		if (TraceLine(start, end))
		{
			// open at eye level
			player->mo->flags |= MF_WATERJUMP;
			player->mo->velocity.z = 350.0;
			player->mo->reactiontime = 70;	// safety net
		}
	}
}

//==========================================================================
//
//  WaterMove
//
//==========================================================================

void WaterMove(player_t* player)
{
	float		forward;
	float		side;
	TVec		vforward;
	TVec		vright;
	TVec		vup;
	TVec		wishvel;

    // Do not let the player control movement
    //  if not onground.
	onground = (player->mo->origin.z <= player->mo->floorz)
		|| (player->mo->flags2 & MF2_ONMOBJ);

	AngleVectors(&player->mo->angles, &vforward, &vright, &vup);

	forward = player->forwardmove;
	side = player->sidemove;

	//	Adjust speed depending on class. Originaly this is done in input
	// code, but to simplify it, we do it here.
	//	Fighter moves faster, cleric at original speed, mage - slower,
	// pig - slower, but faster than mage
	switch (player->pclass)
	{
	 case PCLASS_FIGHTER:
		forward *= 1.18;
		side *=	1.3;
		break;

	 case PCLASS_CLERIC:
		break;

	 case PCLASS_MAGE:
		forward *= 0.9;
		side *= 0.9;
		break;

	 case PCLASS_PIG:
		forward *= 0.97;
		side *= 0.97;
		break;
	}

	if (player->powers[pw_speed] && !player->morphTics)
	{
		// Adjust for a player with a speed artifact
		forward *= 1.5;
		side *= 1.5;
	}

	wishvel = vforward * forward + vright * side;
	if (!forward && !side/* && !cmd.upmove*/)
		wishvel.z -= 60.0;		// drift towards bottom
//	else
//		wishvel.z += cmd.upmove;

	player->mo->velocity += 3.5 * frametime * wishvel;

    if (forward || side)
    {
		SetPlayerRunState(player);
	}

	if (player->buttons & BT_JUMP)
	{
		if (player->mo->watertype == CONTENTS_WATER)
			player->mo->velocity.z = 100.0;
		else if (player->mo->watertype == CONTENTS_SLUDGE)
			player->mo->velocity.z = 80.0;
		else
			player->mo->velocity.z = 50.0;
	}
	CheckWaterJump(player);
}

//==========================================================================
//
//	WaterJump
//
//==========================================================================

void WaterJump(player_t *player)
{
	if (!player->mo->reactiontime || !player->mo->waterlevel)
	{
		player->mo->flags &= ~MF_WATERJUMP;
		player->mo->reactiontime = 0;
	}
	player->mo->velocity.x = player->movedir.x;
	player->mo->velocity.y = player->movedir.y;
}

//==========================================================================
//
// 	ThrustPlayer
//
// 	Moves the given origin along a given angle.
//
//==========================================================================

void ThrustPlayer(player_t* player, angle_t angle, float move)
{
	if (!(player->powers[pw_flight] && !(player->mo->origin.z <= player->mo->floorz)) &&
		P_GetThingFloorType(player->mo) == FLOOR_ICE) // Friction_Low
	{
		move *= 0.5;
	}
	player->mo->velocity.x += move * cos(angle) * frametime;
	player->mo->velocity.y += move * sin(angle) * frametime;
}

//==========================================================================
//
// 	CalcHeight
//
// 	Calculate the walking / running height adjustment
//
//==========================================================================

void CalcHeight(player_t* player)
{
    int		angle;
    float	bob;
    
    // Regular movement bobbing
    // (needs to be calculated for gun swing
    // even if not on ground)
    // OPTIMIZE: tablify angle
    // Note: a LUT allows for effects
    //  like a ramp with low health.
   	player->bob = (player->mo->velocity.x * player->mo->velocity.x +
   					player->mo->velocity.y * player->mo->velocity.y)
   					* frametime * frametime;
    
   	player->bob /= 4.0;
	if (player->bob > MAXBOB)
		player->bob = MAXBOB;

	if (player->mo->flags2 & MF2_FLY && !onground)
	{
		player->bob = 0.5;
	}

    angle = ANG180 / 10 * level->tictime;
    bob = player->bob / 2.0 * sin(angle);

    
    // move viewheight
    if (player->playerstate == PST_LIVE)
    {
		player->viewheight += player->deltaviewheight;

		if (player->viewheight > VIEWHEIGHT)
		{
		    player->viewheight = VIEWHEIGHT;
		    player->deltaviewheight = 0.0;
		}

		if (player->viewheight < VIEWHEIGHT / 2.0)
		{
		    player->viewheight = VIEWHEIGHT / 2.0;
		    if (player->deltaviewheight <= 0.0)
				player->deltaviewheight = 0.00001;
		}
	
		if (player->deltaviewheight)
		{
		    player->deltaviewheight += 0.25;
		    if (!player->deltaviewheight)
				player->deltaviewheight = 0.00001;
		}
    }
	if (player->morphTics)
	{
		player->vieworg.z = player->mo->origin.z + player->viewheight - 20.0;
	}
	else
	{
		player->vieworg.z = player->mo->origin.z + player->viewheight + bob;
	}
	if (player->mo->floorclip && player->playerstate != PST_DEAD
		&& player->mo->origin.z <= player->mo->floorz)
	{
		player->vieworg.z -= player->mo->floorclip;
	}
	if (player->vieworg.z < player->mo->floorz + 4.0)
	{
		player->vieworg.z = player->mo->floorz + 4.0;
	}

    if (player->vieworg.z > player->mo->ceilingz - 4.0)
		player->vieworg.z = player->mo->ceilingz - 4.0;
}

//==========================================================================
//
//  PlayerInSpecialSector
//
//	Called every tic frame that the player origin is in a special sector.
//
//==========================================================================

float pushTab[3] =
{
	5.0 * 35.0 * 35.0 / 32.0,
	10.0 * 35.0 * 35.0 / 32.0,
	25.0 * 35.0 * 35.0 / 32.0
};

void PlayerInSpecialSector(player_t *player)
{
	sector_t *sector;

	sector = player->mo->subsector->sector;
	if (player->mo->origin.z != sector->floorheight)
	{
		// Player is not touching the floor
		return;
	}
	switch (sector->special)
	{
		case 9: // SecretArea
			player->secretcount++;
			sector->special = 0;
			break;

		case 201: case 202: case 203: // Scroll_North_xxx
			ThrustPlayer(player, ANG90, pushTab[sector->special-201]);
			break;
		case 204: case 205: case 206: // Scroll_East_xxx
			ThrustPlayer(player, 0, pushTab[sector->special-204]);
			break;
		case 207: case 208: case 209: // Scroll_South_xxx
			ThrustPlayer(player, ANG270, pushTab[sector->special-207]);
			break;
		case 210: case 211: case 212: // Scroll_West_xxx
			ThrustPlayer(player, ANG180, pushTab[sector->special-210]);
			break;
		case 213: case 214: case 215: // Scroll_NorthWest_xxx
			ThrustPlayer(player, ANG90+ANG45, pushTab[sector->special-213]);
			break;
		case 216: case 217: case 218: // Scroll_NorthEast_xxx
			ThrustPlayer(player, ANG45, pushTab[sector->special-216]);
			break;
		case 219: case 220: case 221: // Scroll_SouthEast_xxx
			ThrustPlayer(player, ANG270+ANG45, pushTab[sector->special-219]);
			break;
		case 222: case 223: case 224: // Scroll_SouthWest_xxx
			ThrustPlayer(player, ANG180+ANG45, pushTab[sector->special-222]);
			break;

		case 40: case 41: case 42: case 43: case 44: case 45:
		case 46: case 47: case 48: case 49: case 50: case 51:
			// Wind specials are handled in (P_mobj):P_XYMovement
			break;

		case 26: // Stairs_Special1
		case 27: // Stairs_Special2
			// Used in (P_floor):ProcessStairSector
			break;

		case 198: // Lightning Special
		case 199: // Lightning Flash special
		case 200: // Sky2
			// Used in (R_plane):R_Drawplanes
			break;
//		default:
//			I_Error("PlayerInSpecialSector: "
//				"unknown special %i", sector->special);
	}
}

//============================================================================
//
//  PlayerOnSpecialFlat
//
//============================================================================

void PlayerOnSpecialFlat(player_t *player, int floorType)
{
	if (player->mo->origin.z != player->mo->floorz)
	{
		// Player is not touching the floor
		return;
	}
	switch (floorType)
	{
		case FLOOR_LAVA:
			if (!(level->tictime & 31))
			{
				DamageMobj(player->mo, &LavaInflictor, NULL, 10);
				StartSoundName(player->mo, "LavaSizzle");
			}
			break;
		default:
			break;
	}
}

//==========================================================================
//
//	PlayerInContents
//
//==========================================================================

void PlayerInContents(player_t *player)
{
	if (!player->mo->waterlevel)
	{
		return;
	}
	switch (player->mo->watertype)
	{
	 case CONTENTS_LAVA:
		if (!(level->tictime & 31))
		{
			DamageMobj(player->mo, &LavaInflictor, NULL, 10);
			StartSoundName(player->mo, "LavaSizzle");
		}
		break;
	}
}

//==========================================================================
//
//  PlayerImpulse
//
//==========================================================================

void PlayerImpulse(player_t *player)
{
	int			impulse;
	int 		i;

	impulse = player->impulse;
    player->impulse = 0;
	player->arti = 0;

	if (impulse <= NUMWEAPONS)
    {
		// The actual changing of the weapon is done when the weapon
		// psprite can do it (A_WeaponReady), so it doesn't happen in
		// the middle of an attack.
		if (player->weaponowned[impulse - 1] &&
			impulse - 1 != player->readyweapon &&
			!player->morphTics)
		{
			player->pendingweapon = impulse - 1;
		}
	}
    else if (impulse == 13)
    {
       	InventoryLeft(player);
	}
    else if (impulse == 14)
    {
       	InventoryRight(player);
	}
    else if (impulse == 15)
	{
       	InventoryUse(player);
	}
    else if (impulse == 16)
    {
		DamageMobj(player->mo, NULL, NULL, 10000);
    }
	// Hexen artifact hot keys
	if (!player->arti)
    {
		if (impulse == 11)
		{
			player->arti = NUMARTIFACTS;
		}
		else if (impulse == 12 && (player->mo->health < MAXHEALTH))
		{
			player->arti = arti_health;
		}
		else if (impulse == 10)
		{
			player->arti = arti_poisonbag;
		}
		else if (impulse == 9)
		{
			player->arti = arti_blastradius;
		}
		else if (impulse == 8)
		{
			player->arti = arti_teleport;
		}
		else if (impulse == 7)
		{
			player->arti = arti_teleportother;
		}
		else if (impulse == 6)
		{
			player->arti = arti_egg;
		}
		else if (impulse == 5 && !player->powers[pw_invulnerability])
		{
			player->arti = arti_invulnerability;
		}
    }

	if (player->arti)
	{
		// Use an artifact
		if (player->arti == NUMARTIFACTS)
		{
			// use one of each artifact (except puzzle artifacts)
			for (i = 1; i < arti_firstpuzzitem; i++)
			{
				PlayerUseArtifact(player, i);
			}
		}
		else
		{
			PlayerUseArtifact(player, player->arti);
		}
	}
}

//==========================================================================
//
//  PlayerThink
//
//==========================================================================

void PlayerThink(player_t* player)
{
	int 		floorType;
    int			newweapon;
	mobj_t 		*pmo;
	mobj_t 		*speedMo;

    // fixme: do this in the cheat code
    if (player->cheats & CF_NOCLIP)
		player->mo->flags |= MF_NOCLIP;
    else
		player->mo->flags &= ~MF_NOCLIP;
    
    if (player->mo->flags & MF_JUSTATTACKED)
    {
		player->forwardmove = 100.0;
		player->sidemove = 0.0;
		player->mo->flags &= ~MF_JUSTATTACKED;
    }

	player->worldTimer++;
	InventoryTick(player);

    if (player->playerstate == PST_DEAD)
    {
		DeathPlayerThink(player);
		return;
    }
    
	if (player->jumpTics)
	{
		player->jumpTics--;
	}
	if (player->morphTics)
	{
		MorphPlayerThink(player);
	}

    // 	Move around.
    // 	Reactiontime is used to prevent movement for a bit after a teleport.
    if (player->mo->reactiontime)
	{
		player->mo->reactiontime--;
		if (player->mo->flags & MF_WATERJUMP)
			WaterJump(player);
	}
    else
	{
		if (player->mo->waterlevel >= 2)
			WaterMove(player);
		else
			MovePlayer(player);
		pmo = player->mo;
		if (player->powers[pw_speed] && !(level->tictime & 1) &&
			vlen(pmo->velocity) > 12.0 * 35.0)
		{
			speedMo = SpawnSpeedMobj(pmo);
			if (speedMo)
			{
				speedMo->angles.yaw = pmo->angles.yaw;
				speedMo->translation = pmo->translation;
				speedMo->target = pmo;
				speedMo->special1 = player->pclass;
				if (speedMo->special1 > 2)
				{
					speedMo->special1 = 0;
				}
				speedMo->sprite = pmo->sprite;
				speedMo->floorclip = pmo->floorclip;
#ifdef FIXME
				if (player == &players[consoleplayer])
				{
					speedMo->translucency = 100;
				}
#endif
			}
		}
	}
    
    CalcHeight(player);

    if (player->mo->subsector->sector->special)
    {
		PlayerInSpecialSector(player);
	}
	if ((floorType = P_GetThingFloorType(player->mo)) != FLOOR_SOLID)
	{
		PlayerOnSpecialFlat(player, floorType);
	}
	PlayerInContents(player);

	switch (player->pclass)
	{
		case PCLASS_FIGHTER:
			if (player->mo->velocity.z <= -35.0 * 35.0 && player->mo->velocity.z >= -40.0 * 35.0 &&
				!player->morphTics && !S_GetSoundPlayingInfo(player->mo,
				GetSoundID("PlayerFighterFallingScream")))
			{
				StartSoundName(player->mo, "PlayerFighterFallingScream");
			}
			break;
		case PCLASS_CLERIC:
			if (player->mo->velocity.z <= -35.0 * 35.0 && player->mo->velocity.z >= -40.0 * 35.0 &&
				!player->morphTics && !S_GetSoundPlayingInfo(player->mo,
				GetSoundID("PlayerClericFallingScream")))
			{
				StartSoundName(player->mo, "PlayerClericFallingScream");
			}
			break;
		case PCLASS_MAGE:
			if (player->mo->velocity.z <= -35.0 * 35.0 && player->mo->velocity.z >= -40.0 * 35.0 &&
				!player->morphTics && !S_GetSoundPlayingInfo(player->mo,
				GetSoundID("PlayerMageFallingScream")))
			{
				StartSoundName(player->mo, "PlayerMageFallingScream");
			}
			break;
		default:
			break;
	}

    // Check for weapon change.
	if (player->impulse)
	{
		PlayerImpulse(player);
	}

    // check for use
    if (player->buttons & BT_USE)
    {
		if (!player->usedown)
		{
		    P_UseLines(player);
		    player->usedown = true;
		}
    }
    else
    {
		player->usedown = false;
	}

	// Morph counter
	if (player->morphTics)
	{
		if (!--player->morphTics)
		{ // Attempt to undo the pig
			UndoPlayerMorph(player);
		}
	}

    // cycle psprites
    MovePsprites(player);
    
    // Counters, time dependend power ups.

    if (player->powers[pw_invulnerability])
	{
		if (player->pclass == PCLASS_CLERIC)
		{
#ifdef FIXME
    if (mobj->flags2 & MF2_DONTDRAW)
    {
    	mobj->translucency = 100;
    }
    else if (mobj->flags & MF_SHADOW)
    {
    	mobj->translucency = 66;
    }
    else if (mobj->flags & MF_ALTSHADOW)
    {
    	mobj->translucency = 33;
    }
    else
    {
		mobj->translucency = 0;
    }
			if (!(level->tictime & 7) && player->mo->flags & MF_SHADOW
				&& !(player->mo->flags2 & MF2_DONTDRAW))
			{
				player->mo->flags &= ~MF_SHADOW;
				if (!(player->mo->flags & MF_ALTSHADOW))
				{
					player->mo->flags2 |= MF2_DONTDRAW|MF2_NONSHOOTABLE;
				}
			}
			if (!(level->tictime & 31))
			{
				if (player->mo->flags2 & MF2_DONTDRAW)
				{
					if (!(player->mo->flags & MF_SHADOW))
					{
						player->mo->flags |= MF_SHADOW|MF_ALTSHADOW;
					}
					else
					{
						player->mo->flags2 &= ~(MF2_DONTDRAW|MF2_NONSHOOTABLE);
					}
				}
				else
				{
					player->mo->flags |= MF_SHADOW;
					player->mo->flags &= ~MF_ALTSHADOW;
				}
			}
#endif
		}
		if (!(--player->powers[pw_invulnerability]))
		{
			player->mo->flags2 &= ~(MF2_INVULNERABLE|MF2_REFLECTIVE);
			if (player->pclass == PCLASS_CLERIC)
			{
				player->mo->flags2 &= ~(MF2_NONSHOOTABLE);
				player->mo->translucency = 0;
			}
		}
	}

	if (player->powers[pw_minotaur])
		player->powers[pw_minotaur]--;

    if (player->powers[pw_infrared])
		player->powers[pw_infrared]--;
		
	if (player->powers[pw_speed])
		player->powers[pw_speed]--;

	if (player->powers[pw_flight] && netgame)
	{
		if (!--player->powers[pw_flight])
		{
			if (player->mo->origin.z != player->mo->floorz)
			{
//FIXME				player->centering = true;
			}
			player->mo->flags2 &= ~MF2_FLY;
			player->mo->flags &= ~MF_NOGRAVITY;
		}
	}

	if (player->poisoncount && !(level->tictime & 15))
	{
		player->poisoncount -= 5;
		if (player->poisoncount < 0)
		{
			player->poisoncount = 0;
		}
		P_PoisonDamage(player, player->poisoner, 1, true); 
	}

    if (player->damagecount)
		player->damagecount--;
		
    if (player->bonuscount)
		player->bonuscount--;

    // Handling colormaps.
    if (player->powers[pw_infrared])
    {
		if (player->powers[pw_infrared] <= ftoi(BLINKTHRESHOLD * 35.0))
		{
			if(player->powers[pw_infrared]&8)
			{
				player->fixedcolormap = 0;
			}
			else
			{
				player->fixedcolormap = 1;
			}
		}
		else if (!(level->tictime & 16))
		{
			if (player->newtorch)
			{
				if (player->fixedcolormap + player->newtorchdelta > 7 ||
					player->fixedcolormap + player->newtorchdelta < 1 ||
					player->newtorch == player->fixedcolormap)
				{
					player->newtorch = 0;
				}
				else
				{
					player->fixedcolormap += player->newtorchdelta;
				}
			}
			else
			{
				player->newtorch = (P_Random() & 7) + 1;
				player->newtorchdelta = (player->newtorch == player->fixedcolormap) ?
						0 : ((player->newtorch > player->fixedcolormap) ? 1 : -1);
			}
		}
    }
    else
		player->fixedcolormap = 0;
}

//==========================================================================
//
//	SB_PaletteFlash
//
//	Sets the new palette based upon the current values of
// player->damagecount and player->bonuscount.
//
//==========================================================================

void SB_PaletteFlash(player_t *player)
{
	player->cshifts[CSHIFT_CONTENTS] = 0;
	if (player->mo->waterlevel == 3)
	{
		switch (player->mo->watertype)
		{
		 case CONTENTS_WATER:
 			player->cshifts[CSHIFT_CONTENTS] = CSHIFT(130, 80, 50, 128);
			break;

		 case CONTENTS_SLUDGE:
 			player->cshifts[CSHIFT_CONTENTS] = CSHIFT(128, 160, 128, 150);
			break;

		 case CONTENTS_LAVA:
 			player->cshifts[CSHIFT_CONTENTS] = CSHIFT(255, 80, 0, 150);
			break;
		}
	}

	if (player->poisoncount)
	{
		player->palette = 0;
		player->palette = (player->poisoncount + 7) >> 3;
		if (player->palette >= NUMPOISONPALS)
		{
			player->palette = NUMPOISONPALS - 1;
		}
		player->palette += STARTPOISONPALS;
	}
	else if (player->damagecount)
	{
		player->palette = (player->damagecount + 7) >> 3;
		if (player->palette >= NUMREDPALS)
		{
			player->palette = NUMREDPALS - 1;
		}
		player->palette += STARTREDPALS;
	}
	else if (player->bonuscount)
	{
		player->palette = (player->bonuscount + 7) >> 3;
		if (player->palette >= NUMBONUSPALS)
		{
			player->palette = NUMBONUSPALS - 1;
		}
		player->palette += STARTBONUSPALS;
	}
	else if (player->mo->flags2 & MF2_ICEDAMAGE)
	{ // Frozen player
		player->palette = STARTICEPAL;
	}
	else
	{
		player->palette = player->weapon_palette;
	}
}

//==========================================================================
//
//	SetViewPos
//
//==========================================================================

void SetViewPos(player_t *player)
{
	player->vieworg.x = player->mo->origin.x;
	player->vieworg.y = player->mo->origin.y;

	if (player->localQuakeHappening)
	{
		float 	intensity;

		intensity = itof(player->localQuakeHappening);
		player->vieworg.x += (Random() - 0.5) * (intensity * 4.0);
		player->vieworg.y += (Random() - 0.5) * (intensity * 4.0);
	}

	SB_PaletteFlash(player);
}

//==========================================================================
//
//  ArtiHealRadius
//
//	Do class specific effect for everyone in radius
//
//==========================================================================

boolean ArtiHealRadius(player_t *player)
{
	mobj_t*		mo;
	mobj_t*		pmo=player->mo;
	float		dist;
	int 		effective=false;
	int 		amount;

	mo = NULL;
	while ((mo = NextMobj(mo)))
	{
		if (!mo->player) continue;
		if (mo->health <= 0) continue;
		dist = MobjDist(pmo, mo);
		if (dist > HEAL_RADIUS_DIST)
		{
			// Out of range
			continue;
		}

		switch (player->pclass)
		{
			case PCLASS_FIGHTER:		// Radius armor boost
				if ((P_GiveArmor(mo->player, ARMOR_ARMOR, 1)) ||
					(P_GiveArmor(mo->player, ARMOR_SHIELD, 1)) ||
					(P_GiveArmor(mo->player, ARMOR_HELMET, 1)) ||
					(P_GiveArmor(mo->player, ARMOR_AMULET, 1)))
				{
					effective = true;
					StartSoundName(mo, "MysticIncant");
				}
				break;
			case PCLASS_CLERIC:			// Radius heal
				amount = 50 + (P_Random() % 50);
				if (P_GiveBody(mo->player, amount))
				{
					effective = true;
					StartSoundName(mo, "MysticIncant");
				}
				break;
			case PCLASS_MAGE:			// Radius mana boost
				amount = 50 + (P_Random() % 50);
				if ((P_GiveMana(mo->player, MANA_1, amount)) ||
					(P_GiveMana(mo->player, MANA_2, amount)))
				{
					effective = true;
					StartSoundName(mo, "MysticIncant");
				}
				break;
			case PCLASS_PIG:
			default:
				break;
		}
	}
	return effective;
}

//==========================================================================
//
//  ArtiTele
//
//==========================================================================

void ArtiTele(player_t *player)
{
	int 	i;
	float	destX;
	float	destY;
	angle_t	destAngle;

	if (deathmatch)
	{
		i = P_Random() % level->numdeathmatchstarts;
		destX = level->deathmatchstarts[i].x;
		destY = level->deathmatchstarts[i].y;
		destAngle = ANG45 * (level->deathmatchstarts[i].angle / 45);
	}
	else
	{
		destX = level->playerstarts[0].x;
		destY = level->playerstarts[0].y;
		destAngle = ANG45 * (level->playerstarts[0].angle / 45);
	}
	P_Teleport(player->mo, destX, destY, destAngle, true);
	if (player->morphTics)
	{
		// Teleporting away will undo any morph effects (pig)
		UndoPlayerMorph(player);
	}
	//StartSoundName(NULL, sfx2_wpnup); // Full volume laugh
}

//==========================================================================
//
//	UseArtifact
//
//	Returns true if the artifact was used.
//
//==========================================================================

boolean UseArtifact(player_t *player, int arti)
{
	mobj_t*	mo;
	angle_t angle;
	int 	i;
	int 	count;

	switch (arti)
	{
		case arti_health:
			if (!P_GiveBody(player, 25))
			{
				return false;
			}
			break;
		case arti_superhealth:
			if (!P_GiveBody(player, 100))
			{
				return false;
			}
			break;
		case arti_invulnerability:
			if (!P_GivePower(player, pw_invulnerability))
			{
				return false;
			}
			break;
		case arti_torch:
			if (!P_GivePower(player, pw_infrared))
			{
				return false;
			}
			break;
		case arti_fly:
			if (!P_GivePower(player, pw_flight))
			{
				return false;
			}
			if (player->mo->velocity.z <= -35.0 * 35.0)
			{ // stop falling scream
				S_StopSound(player->mo);
			}
			break;
		case arti_teleport:
			ArtiTele(player);
			break;
		case arti_egg:
			ArtiPig(player->mo);
			break;
		case arti_healingradius:
			if (!ArtiHealRadius(player))
			{
				return false;
			}
			break;
		case arti_summon:
			ArtiSummon(player);
			break;
		case arti_blastradius:
			ArtiBlastRadius(player);
			break;
		case arti_poisonbag:
			ArtiPoisonBag(player);
			break;
		case arti_teleportother:
			ArtiTeleportOther(player);
			break;
		case arti_speed:
			if(!P_GivePower(player, pw_speed))
			{
				return(false);
			}
			break;
		case arti_boostmana:
			if(!P_GiveMana(player, MANA_1, MAX_MANA))
			{
				if(!P_GiveMana(player, MANA_2, MAX_MANA))
				{
					return false;
				}
				
			}
			else 
			{
				P_GiveMana(player, MANA_2, MAX_MANA);
			}
			break;
		case arti_boostarmor:
			count = 0;

			for(i = 0; i < NUMARMOR; i++)
			{
				count += P_GiveArmor(player, i, 1); // 1 point per armor type
			}
			if(!count)
			{
				return false;
			}
			break;

		case arti_puzzskull:
		case arti_puzzgembig:
		case arti_puzzgemred:
		case arti_puzzgemgreen1:
		case arti_puzzgemgreen2:
		case arti_puzzgemblue1:
		case arti_puzzgemblue2:
		case arti_puzzbook1:
		case arti_puzzbook2:
		case arti_puzzskull2:
		case arti_puzzfweapon:
		case arti_puzzcweapon:
		case arti_puzzmweapon:
		case arti_puzzgear1:
		case arti_puzzgear2:
		case arti_puzzgear3:
		case arti_puzzgear4:
			if (P_UsePuzzleItem(player, arti - arti_firstpuzzitem))
			{
				return true;
			}
			else
			{
				centerprint(player, TXT_USEPUZZLEFAILED);
				return false;
			}
			break;
		default:
			return false;
	}
	return true;
}

//==========================================================================
//
//  PlayerUseArtifact
//
//==========================================================================

void PlayerUseArtifact(player_t *player, int arti)
{
	int i;

	for (i = 0; i < player->inventorySlotNum; i++)
	{
		if (player->inventory[i].type == arti)
		{
			// Found match - try to use
			if (UseArtifact(player, arti))
			{
				// Artifact was used - remove it from inventory
				PlayerRemoveArtifact(player, i);
				if (arti < arti_firstpuzzitem)
				{
					StartSoundName(player->mo, "UseArtifact");
				}
				else
				{
					StartSoundName(player->mo, "PuzzleSuccess");
				}
				player->ArtifactFlash = 4;
			}
			else if (arti < arti_firstpuzzitem)
			{
				// Unable to use artifact, advance pointer
				PlayerNextArtifact(player);
			}
			break;
		}
	}
}

//==========================================================================
//
//  PlayerNextArtifact
//
//==========================================================================

void PlayerNextArtifact(player_t *player)
{
	player->inv_ptr--;
	if (player->inv_ptr < 6)
	{
		player->curpos--;
		if (player->curpos < 0)
		{
			player->curpos = 0;
		}
	}
	if (player->inv_ptr < 0)
	{
		player->inv_ptr = player->inventorySlotNum - 1;
		if (player->inv_ptr < 6)
		{
			player->curpos = player->inv_ptr;
		}
		else
		{
			player->curpos = 6;
		}
	}
	player->readyArtifact = player->inventory[player->inv_ptr].type;
}

//==========================================================================
//
//  PlayerRemoveArtifact
//
//==========================================================================

void PlayerRemoveArtifact(player_t *player, int slot)
{
	int i;

	player->artifactCount--;
	if (!(--player->inventory[slot].count))
	{
		// Used last of a type - compact the artifact list
		player->readyArtifact = arti_none;
		player->inventory[slot].type = arti_none;
		for (i = slot+1; i < player->inventorySlotNum; i++)
		{
			player->inventory[i-1].type = player->inventory[i].type;
			player->inventory[i-1].count = player->inventory[i].count;
		}
		player->inventorySlotNum--;
		// Set position markers and get next readyArtifact
		player->inv_ptr--;
		if (player->inv_ptr < 6)
		{
			player->curpos--;
			if (player->curpos < 0)
			{
				player->curpos = 0;
			}
		}
		if (player->inv_ptr >= player->inventorySlotNum)
		{
			player->inv_ptr = player->inventorySlotNum-1;
		}
		if (player->inv_ptr < 0)
		{
			player->inv_ptr = 0;
		}
		player->readyArtifact = player->inventory[player->inv_ptr].type;
	}
}

//==========================================================================
//
//  TeleportToPlayerStarts
//
//==========================================================================

void TeleportToPlayerStarts(mobj_t *victim)
{
	int 	i;
	int		selections = 0;
	float	destX;
	float	destY;
	angle_t destAngle;

	for (i = 0; i < MAXPLAYERS; i++)
	{
	    if (!players[i].active)
	    	continue;
		selections++;
	}
	i = P_Random() % selections;
	destX = level->playerstarts[i].x;
	destY = level->playerstarts[i].y;
	destAngle = ANG45 * (level->playerstarts[i].angle / 45);
	P_Teleport(victim, destX, destY, destAngle, true);
	//StartSoundName(NULL, sfx2_wpnup); // Full volume laugh
}

//==========================================================================
//
//  TeleportToDeathmatchStarts
//
//==========================================================================

void TeleportToDeathmatchStarts(mobj_t *victim)
{
	int 	i;
	float	destX;
	float	destY;
	angle_t destAngle;

	if (level->numdeathmatchstarts)
	{
		i = P_Random() % level->numdeathmatchstarts;
		destX = level->deathmatchstarts[i].x;
		destY = level->deathmatchstarts[i].y;
		destAngle = ANG45 * (level->deathmatchstarts[i].angle / 45);
		P_Teleport(victim, destX, destY, destAngle, true);
		//StartSoundName(NULL, sfx2_wpnup); // Full volume laugh
	}
	else
	{
	 	TeleportToPlayerStarts(victim);
	}
}

//==========================================================================
//
//  TeleportOther
//
//==========================================================================

void TeleportOther(mobj_t *victim)
{
	if (victim->player)
	{
		if (deathmatch)
			TeleportToDeathmatchStarts(victim);
		else
			TeleportToPlayerStarts(victim);
	}
	else
	{
		// If death action, run it upon teleport
		if (victim->flags & MF_COUNTKILL && victim->special)
		{
			P_RemoveMobjFromTIDList(victim);
			ExecuteLineSpecial(victim->special, &victim->args[0],
					NULL, 0, victim);
			victim->special = 0;
		}

		// Send all monsters to deathmatch spots
		TeleportToDeathmatchStarts(victim);
	}
}

//==========================================================================
//
// EV_LineSearchForPuzzleItem
//
//==========================================================================

boolean EV_LineSearchForPuzzleItem(line_t *line, int *args, mobj_t *mo)
{
	player_t 	*player;
	int 		i;
	int 		type,arti;

	if (!mo) return false;
	player = mo->player;
	if (!player) return false;

	// Search player's inventory for puzzle items
	for (i=0; i<player->artifactCount; i++)
	{
		arti = player->inventory[i].type;
		type = arti - arti_firstpuzzitem;
		if (type < 0) continue;
		if (type == line->arg1)
		{
			// A puzzle item was found for the line
			if (UseArtifact(player, arti))
			{
				// A puzzle item was found for the line
				PlayerRemoveArtifact(player, i);
				if (arti < arti_firstpuzzitem)
				{
					StartSoundName(player->mo, "UseArtifact");
				}
				else
				{
					StartSoundName(player->mo, "PuzzleSuccess");
				}
				player->ArtifactFlash = 4;
				return true;
			}
		}
	}
	return false;
}

//==========================================================================
//
//	G_PlayerExitMap
//
//	Called when the player leaves a map.
//
//==========================================================================

void G_PlayerExitMap(int playerNumber, boolean clusterChange)
{
	int			i;
	player_t*	player;
	int			flightPower;

	player = &players[playerNumber];

	// Strip all current powers (retain flight)
	flightPower = player->powers[pw_flight];
    for (i=0; i<NUMPOWERS; i++)
		player->powers[i] = 0;
//	memset(player->powers, 0, sizeof(player->powers));
	player->powers[pw_flight] = flightPower;

	if (deathmatch)
	{
		player->powers[pw_flight] = 0;
	}
	else
	{
		if (clusterChange)
		{ // Entering new cluster
			// Strip all keys
			player->keys = 0;

			// Strip flight artifact
			for(i = 0; i < 25; i++)
			{
				player->powers[pw_flight] = 0;
				PlayerUseArtifact(player, arti_fly);
			}
			player->powers[pw_flight] = 0;
		}
	}

	if (player->morphTics)
	{
		player->readyweapon = player->mo->special1; // Restore weapon
		weaponinfo[player->readyweapon * NUMCLASSES + player->pclass](player);
		player->pspriteSY = PSpriteSY[player->pclass  * NUMWEAPONS + player->readyweapon];
		player->morphTics = 0;
	}
	player->mo->angles.pitch = 0;
	player->mo->translucency = 0; // Remove invisibility
	player->extralight = 0; // Remove weapon flashes
	player->fixedcolormap = 0; // Remove torch
	player->damagecount = 0; // No palette changes
	player->bonuscount = 0;
	player->poisoncount = 0;
}

//==========================================================================
//
//	G_DeathMatchSpawnPlayer
//
//	Spawns a player at one of the random death match spots called at level
// load and each death
//
//==========================================================================

void G_DeathMatchSpawnPlayer(int playernum)
{ 
    int             i;
    int				j;
	 
    for (j=0 ; j<20 ; j++) 
    { 
		i = P_Random() % level->numdeathmatchstarts;
		if (G_CheckSpot(playernum, &level->deathmatchstarts[i]))
		{
	    	level->deathmatchstarts[i].type = playernum + 1;
	    	P_SpawnPlayer(&level->deathmatchstarts[i]);
	    	return;
		}
    } 
 
    // no good spot, so the player will probably get stuck 
    P_SpawnPlayer(&level->playerstarts[playernum]);
} 

//==========================================================================
//
//	NetGameReborn
//
//==========================================================================

void NetGameReborn(int playernum)
{
    int			i;
	boolean		oldWeaponowned[NUMWEAPONS];
	int 		oldKeys;
	int 		oldPieces;
	boolean 	foundSpot;
	int 		bestWeapon;

	// respawn at the start

	// first dissasociate the corpse
	players[playernum].mo->player = NULL;
		 
	// spawn at random spot if in death match
	if (deathmatch)
	{
    	G_DeathMatchSpawnPlayer(playernum);
    	return;
	}
		 
	// Cooperative net-play, retain keys and weapons
	oldKeys = players[playernum].keys;
	oldPieces = players[playernum].pieces;
	for (i = 0; i < NUMWEAPONS; i++)
	{
		oldWeaponowned[i] = players[playernum].weaponowned[i];
	}

	foundSpot = false;
	if (G_CheckSpot(playernum,
		&level->playerstarts[RebornPosition * MAXPLAYERS + playernum]))
	{ // Appropriate player start spot is open
		P_SpawnPlayer(&level->playerstarts[RebornPosition * MAXPLAYERS + playernum]);
		foundSpot = true;
	}
	else
	{
		// Try to spawn at one of the other player start spots
		for(i = 0; i < MAXPLAYERS; i++)
		{
			if (G_CheckSpot(playernum, &level->playerstarts[RebornPosition * MAXPLAYERS + i]))
			{ // Found an open start spot

				// Fake as other player
				level->playerstarts[RebornPosition * MAXPLAYERS + i].type = playernum + 1;
				P_SpawnPlayer(&level->playerstarts[RebornPosition * MAXPLAYERS + i]);

				// Restore proper player type
				level->playerstarts[RebornPosition * MAXPLAYERS + i].type = i + 1;
	
				foundSpot = true;
				break;
			}
		}
	}

	if (!foundSpot)
	{
		// Player's going to be inside something
		P_SpawnPlayer(&level->playerstarts[RebornPosition * MAXPLAYERS + playernum]);
	}

	// Restore keys and weapons
	players[playernum].keys = oldKeys;
	players[playernum].pieces = oldPieces;
	for (bestWeapon = 0, i = 0; i < NUMWEAPONS; i++)
	{
		if (oldWeaponowned[i])
		{
			bestWeapon = i;
			players[playernum].weaponowned[i] = true;
		}
	}
	players[playernum].mana[MANA_1] = 25;
	players[playernum].mana[MANA_2] = 25;
	if (bestWeapon)
	{
		// Bring up the best weapon
		players[playernum].pendingweapon = bestWeapon;
	}
}

//==========================================================================
//
//	SetPlayerColor
//
//==========================================================================

void SetPlayerColor(player_t *player)
{
	player->mo->translation = 0;
	if (player->pclass == PCLASS_FIGHTER)
	{
		// The first type should be blue, and the third should be the
		// Fighter's original gold color
		if (player->color == 0)
		{
	    	player->mo->translation = 2;
		}
		else if (player->color != 2)
		{
	    	player->mo->translation = player->color;
		}
	}
	else if (player->color)
	{
		// Set color translation bits for player sprites
    	player->mo->translation = player->pclass * (MAXPLAYERS - 1) + player->color;
	}
}

//==========================================================================
//
//	SpawnClient
//
//==========================================================================

void SpawnClient(int i)
{
	player_t	*player;
    int			j;
	int			oldKeys;
	int			oldPieces;
	boolean		oldWeaponowned[NUMWEAPONS];
	boolean		playerWasReborn;
	int			bestWeapon;

    player = &players[i];

	player->attacker = NULL;
	player->poisoner = NULL;

	if (netgame && !deathmatch)
	{
		// Cooperative net-play, retain keys and weapons
		oldKeys = player->keys;
		oldPieces = player->pieces;
		for (j = 0; j < NUMWEAPONS; j++)
		{
			oldWeaponowned[j] = player->weaponowned[j];
		}
		playerWasReborn = (player->playerstate == PST_REBORN);
	}

    //	Spawn player
	if (deathmatch)
	{
		G_DeathMatchSpawnPlayer(i);
	}
	else
	{
		P_SpawnPlayer(&level->playerstarts[RebornPosition * MAXPLAYERS + i]);
	}

	if (netgame && !deathmatch && playerWasReborn)
	{
		// Restore keys and weapons when reborn in co-op
		player->keys = oldKeys;
		player->pieces = oldPieces;
		for (bestWeapon = 0, j = 0; j < NUMWEAPONS; j++)
		{
			if (oldWeaponowned[j])
			{
				bestWeapon = j;
				player->weaponowned[j] = true;
			}
		}
		player->mana[MANA_1] = 25;
		player->mana[MANA_2] = 25;
		if (bestWeapon)
		{
			// Bring up the best weapon
			player->pendingweapon = bestWeapon;
		}
	}

	// Destroy all things touching players
	P_TeleportMove(player->mo, player->mo->origin.x, player->mo->origin.y);
}

//==========================================================================
//
//	DisconnectClient
//
//==========================================================================

void DisconnectClient(player_t *player)
{
	if (player->mo)
	{
		player->mo->player = NULL;
		DamageMobj(player->mo, NULL, NULL, 10000);
	}
   	bprint("%s left the game\n", ARR2STR(player->name));
	StartSoundName(NULL, "Chat");
}

//==========================================================================
//
//	UserinfoChanged
//
//==========================================================================

void UserinfoChanged(player_t *player)
{
	if (!player->spawned)
	{
		return;
	}
	SetPlayerColor(player);
}


