//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d3d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
//
//  TDirect3DDrawer::InitLightTables
//
//==========================================================================

void TDirect3DDrawer::InitLightTables(void)
{
	int			i;
	int			j;
	float		light_level;
	float		startmap;

    //	Calculate the light levels to use for each
    // light_level / distance combination.
    for (i = 0; i < LIGHTLEVELS; i++)
    {
        startmap = ((float)LIGHTLEVELS - 1.0 - (float)i) * (2.0 * 256.0 / (float)LIGHTLEVELS);
        for (j = 0; j < MAXLIGHTZ; j++)
        {
            light_level = startmap - 640.0 / ((float)j + 1.0);

            if (light_level < 0)
                light_level = 0;

            if (light_level > 255)
                light_level = 255;

            zlight[i][j] = 255 - light_level;
        }
    }
}

//==========================================================================
//
//	TDirect3DDrawer::EmitVertex
//
//==========================================================================

void TDirect3DDrawer::EmitVertex(D3DLVERTEX &lv, const draw_vertex_t &dv, byte alpha)
{
	draw_vertex_t	tr;

	float light;
	if (fixedlight)
	{
		light = fixedlight;
	}
	else if (use_zlight)
	{
		int lightlevel = (int)(dv.l * LIGHTSEGSCALE + extralight);
		if (lightlevel >= LIGHTLEVELS)
			lightlevel = LIGHTLEVELS - 1;
		if (lightlevel < 0)
			lightlevel = 0;

		int index = (int)(Length(dv - vieworg) * light_z_scale);
		if (index >= MAXLIGHTZ)
			index = MAXLIGHTZ - 1;

		light = zlight[lightlevel][index];
	}
	else
	{
		light = dv.l + extralight * 16.0;
		if (light < 0)
			light = 0;
		if (light > 255)
			light = 255;
	}
	int l = (int)light;

	lv = D3DLVERTEX(D3DVECTOR(dv.x, dv.y, dv.z),
		(alpha << 24) | (l << 16) | (l << 8) | l, 0,
		dv.s * tex_iw, dv.t * tex_ih);
}

//==========================================================================
//
//	TDirect3DDrawer::DrawPolygon
//
//==========================================================================

void TDirect3DDrawer::DrawPolygon(draw_vertex_t *cv, int count, int texture,
	int, const TVec &, bool use_fog)
{
	D3DLVERTEX		out[256];
	int				i;

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}
	if (use_fog)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, TRUE);
	}

	for (i = 0; i < count; i++)
	{
		EmitVertex(out[i], cv[i], 255);
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, count, 0);

	if (use_fog)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, FALSE);
	}
}

//==========================================================================
//
//	TDirect3DDrawer::DrawMaskedPolygon
//
//==========================================================================

void TDirect3DDrawer::DrawMaskedPolygon(draw_vertex_t *cv, int count,
	int texture, int translucency, int, const TVec &, bool use_fog)
{
	D3DLVERTEX		out[256];
	int				i;

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}
	if (translucency || !HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	}
	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
	}

	if (use_fog)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, TRUE);
	}

	int alpha = (100 - translucency) * 255 / 100;
	for (i = 0; i < count; i++)
	{
		EmitVertex(out[i], cv[i], alpha);
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, count, 0);

	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	}
	if (translucency || !HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
	}
	if (use_fog)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, FALSE);
	}
}

//==========================================================================
//
//	TDirect3DDrawer::DrawZPolygon
//
//==========================================================================

void TDirect3DDrawer::DrawZPolygon(draw_vertex_t *cv, int count, int)
{
	D3DLVERTEX		out[256];
	int				i;

	RenderDevice->SetTexture(0, NULL);

	for (i = 0; i < count; i++)
	{
		EmitVertex(out[i], cv[i], 0xff);
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, count, 0);
}

//==========================================================================
//
//	TDirect3DDrawer::DrawSkyPolygon
//
//==========================================================================

void TDirect3DDrawer::DrawSkyPolygon(draw_vertex_t *cv, int count, int texture,
	float xoffs, float yoffs, bool double_sky)
{
	D3DLVERTEX		out[256];
	int				i;

	SetSkyTexture(texture, double_sky);

	for (i = 0; i < count; i++)
	{
		draw_vertex_t v = cv[i];
		v += vieworg;
		v.s += xoffs;
		v.t += yoffs;
		v.l = 255;
		EmitVertex(out[i], v, 0xff);
	}

	RenderDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, FALSE);
	if (double_sky)
	{
		if (HavesAlphaTest)
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		}
		else
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
		}
	}
	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, count, 0);
	if (double_sky)
	{
		if (HavesAlphaTest)
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
		}
		else
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
		}
	}
	RenderDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, TRUE);
}

//==========================================================================
//
//	TDirect3DDrawer::DrawSpritePolygon
//
//==========================================================================

void TDirect3DDrawer::DrawSpritePolygon(draw_vertex_t *cv, int lump,
	const TVec &, int translucency, int translation, bool use_fog)
{
	D3DLVERTEX		out[256];
	int				i;

	SetSpriteLump(lump, translation);

	int alpha = (100 - translucency) * 255 / 100;
	for (i = 0; i < 4; i++)
	{
		EmitVertex(out[i], cv[i], alpha);
	}

	if (translucency || !HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	}
	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
	}
	if (use_fog)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, TRUE);
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, out, 4, 0);

	if (translucency || !HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
	}
	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	}
	if (use_fog)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, FALSE);
	}
}

