//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d3d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TDirect3DDrawer::ToPowerOf2
//
//==========================================================================

int TDirect3DDrawer::ToPowerOf2(int val)
{
	int answer = 1;
	while (answer < val)
		answer <<= 1;
	return answer;
}

//==========================================================================
//
//	TDirect3DDrawer::CreateSurface
//
//==========================================================================

LPDIRECTDRAWSURFACE7 TDirect3DDrawer::CreateSurface(int w, int h)
{
	DDSURFACEDESC2			ddsd;
	LPDIRECTDRAWSURFACE7	surf = NULL;
	int i;

	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS | DDSD_PIXELFORMAT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE | SurfaceMemFlag;
    ddsd.dwWidth  = w;
    ddsd.dwHeight = h;
	memcpy(&ddsd.ddpfPixelFormat, &PixelFormat, sizeof(DDPIXELFORMAT));

	do
	{
		if (DDraw->CreateSurface(&ddsd, &surf, NULL) == DD_OK)
		{
//			DDCOLORKEY ckey;
//			ckey.dwColorSpaceLowValue = 0x03ff;
//			ckey.dwColorSpaceHighValue = 0x03ff;
//			surf->SetColorKey(DDCKEY_SRCBLT, &ckey);

			return surf;
		}

//		con << "Trying to replace\n";
		tscount++;
		for (i = 0; i < numsurfaces; i++)
		{
			int index = (i + tscount) % numsurfaces;
			if (texturesurfaces[index])
			{
//				con << "Releasing " << index << endl;
				SAFE_RELEASE(texturesurfaces[index]);
				break;
			}
		}
	} while (i < numsurfaces);

	cond << "Not enough video memory\n";
	return NULL;
}

//==========================================================================
//
//	TDirect3DDrawer::LockSurface
//
//==========================================================================

word *TDirect3DDrawer::LockSurface(LPDIRECTDRAWSURFACE7 surf)
{
	DDSURFACEDESC2			ddsd;
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_LPSURFACE;
	if (FAILED(surf->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL)))
	{
		cond << "Failed to lock surface\n";
		return NULL;
	}
#if 0
	word *dst = (word*)ddsd.lpSurface;
	for (size_t i = 0; i < ddsd.dwWidth; i++)
 		for (size_t j = 0; j < ddsd.dwHeight; j++)
			*dst++ = 0x03ff;
#else
	memset(ddsd.lpSurface, 0, 2 * ddsd.dwWidth * ddsd.dwHeight);
#endif
	return (word*)ddsd.lpSurface;
}

//==========================================================================
//
//	TDirect3DDrawer::InitTextures
//
//==========================================================================

void TDirect3DDrawer::InitTextures(void)
{
	numsurfaces = numtextures + numflats + numspritelumps +
		MAX_TRANSLATED_SPRITES + MAX_PICS;
	texturesurfaces = (LPDIRECTDRAWSURFACE7*)Z_Calloc(numsurfaces * 4);
	//	Textures
	texturedata = texturesurfaces;
	//	Flats
	flatdata = texturedata + numtextures;
	//	Sprite lumps
	spritedata = flatdata + numflats;
	trsprdata = spritedata + numspritelumps;
	//	2D graphics
	picdata = trsprdata + MAX_TRANSLATED_SPRITES;

	textureiw = (float*)Z_Calloc(numtextures * 4);
	textureih = (float*)Z_Calloc(numtextures * 4);
	spriteiw = (float*)Z_Calloc(numspritelumps * 4);
	spriteih = (float*)Z_Calloc(numspritelumps * 4);
}

//==========================================================================
//
//	TDirect3DDrawer::ReleaseTextures
//
//==========================================================================

void TDirect3DDrawer::ReleaseTextures(void)
{
	for (int i = 0; i < numsurfaces; i++)
	{
		SAFE_RELEASE(texturesurfaces[i]);
	}
}

//==========================================================================
//
//	TDirect3DDrawer::DrawColumnInCache
//
// 	Clip and draw a column from a patch into a flat buffer.
//
//		column - column to draw
//		cache - buffer
//		originx, originy - position of column in the buffer
//		cachewidth, cacheheight - size of the cache
//
//==========================================================================

void TDirect3DDrawer::DrawColumnInCache(column_t* column, word* cache,
	int originx, int originy, int cachewidth, int cacheheight)
{
    int		count;
    int		position;
    byte*	source;
    word*	dest;
	
	// step through the posts in a column
    while (column->topdelta != 0xff)
    {
		source = (byte *)column + 3;
		count = column->length;
		position = originy + column->topdelta;

		//	Clip position
		if (position < 0)
		{
	    	count += position;
			source -= position;
	    	position = 0;
		}
		if (position + count > cacheheight)
		{
	    	count = cacheheight - position;
		}
    	dest = cache + originx + position * cachewidth;

    	while (count-- > 0)
    	{
			*dest = pal8_to16[*source];
			source++;
			dest += cachewidth;
    	}
		
		column = (column_t *)((byte *)column + column->length + 4);
    }
}

//==========================================================================
//
//	TDirect3DDrawer::GenerateTexture
//
// 	Using the texture definition, the composite texture is created from the
// patches, and each column is cached.
//
//==========================================================================

void TDirect3DDrawer::GenerateTexture(int texnum)
{
    word*			block;
    texdef_t*		texture;
    texpatch_t*		patch;	
    patch_t*		realpatch;
    int				x;
    int				x1;
    int				x2;
    int				i;
    column_t*		patchcol;
	int				wtimes;
	int				htimes;
	int				texture_width;
	int				workw;
	int				workh;

    texture = textures[texnum];

	texture_width = texture->width;
	if (texture_width > 256)
	{
		cond << "JL in trouble - texture " << texture->name
			<< " haves width " << texture_width << " greater than 256\n";
		texture_width = 256;
	}
	if (texture_width & (texture_width - 1))
	{
		//	This will happen only with Doom's dummy texture
		con << "Texture width is not a power of 2\n";
		texture_width = 16;
	}
	workw = texture_width;
	if (texture->height & (texture->height - 1))
	{
		//	Texture height is not a power of 2, use 256
		workh = 256;
	}
	else
	{
		workh = texture->height;
	}
	if (square_textures)
	{
		workw = workh = MAX(workw, workh);
	}

	textureiw[texnum] = 1.0 / (float)workw;
	textureih[texnum] = 1.0 / (float)workh;
	texturedata[texnum] = CreateSurface(workw, workh);
	if (!texturedata[texnum])
		return;
	block = LockSurface(texturedata[texnum]);

    // Composite the columns together.
    patch = texture->patches;

	wtimes = workw / texture_width;
	htimes = workh / texture->height;

    for (i = 0; i < texture->patchcount; i++, patch++)
    {
		realpatch = (patch_t*)W_CacheLumpNum(patch->patch, PU_CACHE);
		x1 = patch->originx;
		x2 = x1 + LittleShort(realpatch->width);

		if (x1 < 0)
	    	x = 0;
		else
	    	x = x1;
	
		if (x2 > texture_width)
	    	x2 = texture_width;

		for ( ; x < x2; x++)
		{
	    	patchcol = (column_t *)((byte *)realpatch
				    + LittleLong(realpatch->columnofs[x - x1]));
			for (int ht = 0; ht < htimes; ht++)
			{
				for (int wt = 0; wt < wtimes; wt++)
				{
			    	DrawColumnInCache(patchcol, block + wt * texture_width +
			    		ht * texture->height * workw, x, patch->originy,
			    		workw, texture->height);
				}
			}
		}
    }

	texturedata[texnum]->Unlock(NULL);
}

//==========================================================================
//
// 	TDirect3DDrawer::SetTexture
//
//==========================================================================

void TDirect3DDrawer::SetTexture(int tex)
{
	if (!RenderDevice)
		return;

	tex = texturetranslation[tex];

    if (!texturedata[tex])
		GenerateTexture(tex);

    RenderDevice->SetTexture(0, texturedata[tex]);
	tex_iw = textureiw[tex];
	tex_ih = textureih[tex];
}

//==========================================================================
//
//	GenerateSkyTexture
//
//==========================================================================

#if 0
static void GenerateSkyTexture(int texnum)
{
    byte*			block;
    texdef_t*		texture;
    texpatch_t*		patch;	
    patch_t*		realpatch;
    int				x;
    int				x1;
    int				x2;
    int				i;
    column_t*		patchcol;
	int				wtimes;
	int				htimes;

    texture = textures[texnum];

    block = (byte*)Z_Malloc(1024 * 256, PU_STATIC, (void**)&skytexturedata[texnum]);
	memset(block, TRANSPARENT_COLOR, 1024 * 256);

    // Composite the columns together.
    patch = texture->patches;

	wtimes = 1024 / texture->width;
	htimes = 256 / texture->height;

    for (i = 0; i < texture->patchcount; i++, patch++)
    {
		realpatch = (patch_t*)W_CacheLumpNum(patch->patch, PU_CACHE);
		x1 = patch->originx;
		x2 = x1 + LittleShort(realpatch->width);

		if (x1 < 0)
	    	x = 0;
		else
	    	x = x1;
	
		if (x2 > texture->width)
	    	x2 = texture->width;

		for ( ; x < x2; x++)
		{
	    	patchcol = (column_t *)((byte *)realpatch
				    + LittleLong(realpatch->columnofs[x - x1]));
			for (int ht = 0; ht < htimes; ht++)
			{
				for (int wt = 0; wt < wtimes; wt++)
				{
			    	DrawColumnInCache(patchcol, block + wt * texture->width +
			    		ht * texture->height * 1024, x, patch->originy,
			    		1024, texture->height);
				}
			}
		}
    }

    // Now that the texture has been built in column cache,
    //  it is purgable from zone memory.
    Z_ChangeTag(block, PU_CACHE);
}
#endif

//==========================================================================
//
// 	TDirect3DDrawer::SetSkyTexture
//
//==========================================================================

void TDirect3DDrawer::SetSkyTexture(int tex, bool double_sky)
{
	if (!RenderDevice)
		return;

	word saved;
	if (double_sky)
	{
		saved = pal8_to16[0];
		pal8_to16[0] = 0;
	}
#if 0
	tex = texturetranslation[tex];

    if (!skytexturedata[tex])
		GenerateSkyTexture(tex);

    ds_source = skytexturedata[tex];
#else
	SetTexture(tex);
#endif
	if (double_sky)
	{
		pal8_to16[0] = saved;
	}
}

//==========================================================================
//
//	TDirect3DDrawer::GenerateFlat
//
//==========================================================================

void TDirect3DDrawer::GenerateFlat(int num)
{
	flatdata[num] = CreateSurface(64, 64);
	if (!flatdata[num])
		return;
	word *block = LockSurface(flatdata[num]);
	byte *data = (byte*)W_CacheLumpNum(flatlumps[num], PU_CACHE);

	byte *src = data;
	word *dst = block;
	for (int i = 0; i < 64 * 64; i++)
	{
		*dst++ = pal8_to16[*src++];
	}

	flatdata[num]->Unlock(NULL);
}

//==========================================================================
//
//	TDirect3DDrawer::SetFlat
//
//==========================================================================

void TDirect3DDrawer::SetFlat(int num)
{
	if (!RenderDevice)
		return;

	num &= ~TEXF_FLAT;
	num = flattranslation[num];

	if (!flatdata[num])
	{
		GenerateFlat(num);
	}
    RenderDevice->SetTexture(0, flatdata[num]);
	tex_iw = 1.0 / 64.0;
	tex_ih = 1.0 / 64.0;
}

//==========================================================================
//
//	TDirect3DDrawer::GenerateSprite
//
//==========================================================================

void TDirect3DDrawer::GenerateSprite(int lump)
{
    patch_t	*patch = (patch_t*)W_CacheLumpNum(spritelumps[lump], PU_CACHE);

	spritewidth[lump] = LittleShort(patch->width);
	spriteheight[lump] = LittleShort(patch->height);
	spriteoffset[lump] = LittleShort(patch->leftoffset);
	spritetopoffset[lump] = LittleShort(patch->topoffset);

	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);
	int p2w = ToPowerOf2(w);
	int p2h = ToPowerOf2(h);
	fixed_t xscale;
	if (p2w > maxTexSize)
	{
		xscale = FRACUNIT * w / maxTexSize;
		p2w = maxTexSize;
		if (square_textures)
		{
			p2h = p2w;
		}
		spriteiw[lump] = 1.0 / (float)w;
	}
	else
	{
		xscale = FRACUNIT;
		if (square_textures)
		{
			p2w = p2h = MAX(p2w, p2h);
		}
		spriteiw[lump] = 1.0 / (float)p2w;
	}
	spriteih[lump] = 1.0 / (float)p2h;
	int workw = MIN(w, p2w);

	spritedata[lump] = CreateSurface(p2w, p2h);
	if (!spritedata[lump])
		return;
	word *block = LockSurface(spritedata[lump]);

	for (int x = 0; x < workw; x++)
	{
    	column_t *column = (column_t *)((byte *)patch +
    		LittleLong(patch->columnofs[(x * xscale) >> FRACBITS]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
		    word* dest = block + x + column->topdelta * p2w;
			int count = column->length;

	    	while (count--)
	    	{
				*dest = pal8_to16[*source];
				source++;
				dest += p2w;
	    	}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}

	spritedata[lump]->Unlock(NULL);
}

//==========================================================================
//
//	TDirect3DDrawer::GenerateTranslatedSprite
//
//==========================================================================

void TDirect3DDrawer::GenerateTranslatedSprite(int lump, int slot, int translation)
{
    patch_t	*patch = (patch_t*)W_CacheLumpNum(spritelumps[lump], PU_CACHE);

	spritewidth[lump] = LittleShort(patch->width);
	spriteheight[lump] = LittleShort(patch->height);
	spriteoffset[lump] = LittleShort(patch->leftoffset);
	spritetopoffset[lump] = LittleShort(patch->topoffset);

	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);
	int p2w = ToPowerOf2(w);
	int p2h = ToPowerOf2(h);
	fixed_t xscale;
	if (p2w > maxTexSize)
	{
		xscale = FRACUNIT * w / maxTexSize;
		p2w = maxTexSize;
		if (square_textures)
		{
			p2h = p2w;
		}
		trspriw[slot] = 1.0 / (float)w;
	}
	else
	{
		xscale = FRACUNIT;
		if (square_textures)
		{
			p2w = p2h = MAX(p2w, p2h);
		}
		trspriw[slot] = 1.0 / (float)p2w;
	}
	trsprih[slot] = 1.0 / (float)p2h;
	int workw = MIN(w, p2w);

	trsprdata[slot] = CreateSurface(p2w, p2h);
	if (!trsprdata[slot])
		return;
	word *block = LockSurface(trsprdata[slot]);
	trsprlump[slot] = lump;
	trsprtnum[slot] = translation;

	byte *trtab = translationtables + translation * 256;

	for (int x = 0; x < workw; x++)
	{
    	column_t *column = (column_t *)((byte *)patch +
    		LittleLong(patch->columnofs[(x * xscale) >> FRACBITS]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
		    word* dest = block + x + column->topdelta * p2w;
			int count = column->length;

	    	while (count--)
	    	{
				*dest = pal8_to16[trtab[*source]];
				source++;
				dest += p2w;
	    	}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}

	trsprdata[slot]->Unlock(NULL);
}

//==========================================================================
//
//	TDirect3DDrawer::SetSpriteLump
//
//==========================================================================

void TDirect3DDrawer::SetSpriteLump(int lump, int translation)
{
	if (!RenderDevice)
		return;

	if (translation)
	{
		int i;
		int avail = -1;
		for (i = 0; i <	MAX_TRANSLATED_SPRITES; i++)
		{
			if (trsprdata[i])
			{
				if (trsprlump[i] == lump && trsprtnum[i] == translation)
				{
					RenderDevice->SetTexture(0, trsprdata[i]);
					tex_iw = trspriw[i];
					tex_ih = trsprih[i];
					return;
				}
			}
			else
			{
				if (avail < 0)
					avail = i;
			}
		}
		if (avail < 0)
		{
			Z_Free(trsprdata[0]);
			avail = 0;
		}
		GenerateTranslatedSprite(lump, avail, translation);
		RenderDevice->SetTexture(0, trsprdata[avail]);
		tex_iw = trspriw[avail];
		tex_ih = trsprih[avail];
	}
	else
	{
		if (!spritedata[lump])
		{
			GenerateSprite(lump);
		}
		RenderDevice->SetTexture(0, spritedata[lump]);
		tex_iw = spriteiw[lump];
		tex_ih = spriteih[lump];
	}
}

//==========================================================================
//
//	TDirect3DDrawer::SetPic
//
//==========================================================================

void TDirect3DDrawer::SetPic(int handle)
{
	if (!RenderDevice)
		return;

	if (!picdata[handle])
	{
		switch (pic_list[handle].type)
	 	{
 		 case PIC_PATCH:
			GeneratePicFromPatch(handle);
			break;

		 case PIC_RAW:
			GeneratePicFromRaw(handle);
			break;
		}
	}

	RenderDevice->SetTexture(0, picdata[handle]);
	tex_iw = piciw[handle];
	tex_ih = picih[handle];
}

//==========================================================================
//
//	TDirect3DDrawer::GeneratePicFromPatch
//
//==========================================================================

void TDirect3DDrawer::GeneratePicFromPatch(int handle)
{
	patch_t *patch = (patch_t*)W_CacheLumpName(pic_list[handle].name, PU_CACHE);
	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);
	int p2w = ToPowerOf2(w);
	int p2h = ToPowerOf2(h);
	fixed_t xscale;
	if (p2w > maxTexSize)
	{
		xscale = FRACUNIT * w / maxTexSize;
		p2w = maxTexSize;
		if (square_textures)
		{
			p2h = p2w;
		}
		piciw[handle] = 1.0 / (float)w;
	}
	else
	{
		xscale = FRACUNIT;
		if (square_textures)
		{
			p2w = p2h = MAX(p2w, p2h);
		}
		piciw[handle] = 1.0 / (float)p2w;
	}
	picih[handle] = 1.0 / (float)p2h;
	int workw = MIN(w, p2w);

	picdata[handle] = CreateSurface(p2w, p2h);
	if (!picdata[handle])
		return;
    word *block = LockSurface(picdata[handle]);

	for (int x = 0; x < workw; x++)
	{
    	column_t *column = (column_t *)((byte *)patch +
    		LittleLong(patch->columnofs[(x * xscale) >> FRACBITS]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
		    word* dest = block + x + column->topdelta * p2w;
			int count = column->length;

	    	while (count--)
	    	{
				*dest = pal8_to16[*source];
				source++;
				dest += p2w;
	    	}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}

	picdata[handle]->Unlock(NULL);
}

//==========================================================================
//
//	TDirect3DDrawer::GeneratePicFromRaw
//
//==========================================================================

void TDirect3DDrawer::GeneratePicFromRaw(int handle)
{
	int p2w = 512;
	int p2h = 256;
	fixed_t xscale;
	if (p2w > maxTexSize)
	{
		p2w = maxTexSize;
		xscale = FRACUNIT * 320 / maxTexSize;
		piciw[handle] = 1.0 / 320.0;
	}
	else
	{
		xscale = FRACUNIT;
		piciw[handle] = 1.0 / 512.0;
	}
	if (square_textures)
	{
		p2h = p2w;
	}
	picih[handle] = 1.0 / (float)p2h;
	int workw = MIN(320, p2w);

	byte* raw = (byte*)W_CacheLumpName(pic_list[handle].name, PU_CACHE);
	picdata[handle] = CreateSurface(p2w, p2h);
	if (!picdata[handle])
		return;
    word *block = LockSurface(picdata[handle]);

	for (int y = 0; y < 200; y++)
	{
		byte *src = raw + 320 * y;
		word *dst = block + p2w * y;
		for (int x = 0; x < workw; x++)
		{
			dst[x] = pal8_to16[src[(x * xscale) >> FRACBITS]];
		}
	}

	picdata[handle]->Unlock(NULL);
}

