//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include <allegro.h>
#include "d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern byte		gammatable[5][256];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

BEGIN_COLOR_DEPTH_LIST
	COLOR_DEPTH_8
	COLOR_DEPTH_15
	COLOR_DEPTH_16
END_COLOR_DEPTH_LIST

dword				mmx_mask4 = 0;
dword				mmx_mask8 = 0;
dword				mmx_mask16 = 0;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static BITMAP		*gamebitmap = NULL;

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TSoftwareDrawer::Init
//
//==========================================================================

void TSoftwareDrawer::Init(void)
{
	if (cpu_mmx)
	{
		mmx_mask4 = ~3;
		mmx_mask8 = ~7;
		mmx_mask16 = ~15;
	}
}

//==========================================================================
//
// 	TSoftwareDrawer::SetResolution
//
// 	Set up the video mode
//
//==========================================================================

bool TSoftwareDrawer::SetResolution(int Width, int Height, int bpp)
{
	if (!Width || !Height)
	{
		//	Set defaults
		Width = 320;
		Height = 200;
		bpp = 8;
	}

	if (gamebitmap)
    {
      	destroy_bitmap(gamebitmap);
        gamebitmap = NULL;
    }

	if (zbuffer)
	{
		Z_Free(zbuffer);
		zbuffer = NULL;
	}

    set_color_depth(bpp);
    if (set_gfx_mode(GFX_AUTODETECT, Width, Height, 0, 0))
    {
		con << "Failed to set video mode:\n" << allegro_error << endl;
    	return false;
    }

    gamebitmap = create_bitmap(SCREEN_W, SCREEN_H);
	if (!gamebitmap)
	{
		con << "Failed to create game bitmap:\n";
		return false;
	}

	zbuffer = (word*)Z_Malloc(SCREEN_W * SCREEN_H * 2, PU_VIDEO, 0);
	if (!zbuffer)
	{
		con << "Not enough memory for Z-buffer\n";
		return false;
	}

	ScreenWidth = SCREEN_W;
    ScreenHeight = SCREEN_H;
	ScreenPitch = ScreenWidth;
	ScreenBPP = bpp;

	scrn = (byte*)gamebitmap->dat;
    memset(scrn, 0, SCREEN_W * SCREEN_H * ((bpp + 7) >> 3));

	if (ScreenBPP == 15)
	{
		rshift = _rgb_r_shift_15;
		gshift = _rgb_g_shift_15;
		bshift = _rgb_b_shift_15;
	}
	else if (ScreenBPP == 16)
	{
		rshift = _rgb_r_shift_16;
		gshift = _rgb_g_shift_16;
		bshift = _rgb_b_shift_16;
	}

	return true;
}

//==========================================================================
//
// 	TSoftwareDrawer::SetPalette8
//
// 	Takes full 8 bit values.
//
//==========================================================================

void TSoftwareDrawer::SetPalette8(byte *palette)
{
  	int 	i;
	byte	*table;

	if (ScreenBPP != 8)
		return;

	if ((dword)usegamma > 4)
	{
		con << "Invalid gamma " << usegamma << endl;
		usegamma = 0;
	}
	table = gammatable[usegamma];

#if 0//defined DJGPP || defined __linux__

	//	Wait for vertical retrace
	while ((inportb(0x3da) & 8) != 8);
	while ((inportb(0x3da) & 8) == 8);

  	outportb(0x3c8, 0);
  	for (i = 0; i < 768; i++)
    	outportb(0x3c9, table[*palette++] >> 2);

#else

	PALETTE		pal;
	for (i = 0; i < 256; i++)
	{
		pal[i].r = table[*palette++] >> 2;
		pal[i].g = table[*palette++] >> 2;
		pal[i].b = table[*palette++] >> 2;
	}
	set_palette(pal);

#endif
}

//==========================================================================
//
//	TSoftwareDrawer::Update
//
// 	Blit to the screen / Flip surfaces
//
//==========================================================================

void TSoftwareDrawer::Update(void)
{
	blit(gamebitmap, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
}

//==========================================================================
//
// 	TSoftwareDrawer::Shutdown
//
// 	Restore text mode
//
//==========================================================================

void TSoftwareDrawer::Shutdown(void)
{
	if (gamebitmap)
  		destroy_bitmap(gamebitmap);
    set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);
}

