//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

#define MAX_POLYS		2500
#define MAX_VERTS		5000

// TYPES -------------------------------------------------------------------

struct EdgeT
{
	int		v0;   /* Y(v0) < Y(v1) */
	float	dx;	/* X(v1) - X(v0) */
	float	dy;	/* Y(v1) - Y(v0) */
	fixed_t	fdxdy;	/* dx/dy in fixed-point */
	fixed_t	fsx;	/* first sample point x coord */
	fixed_t	fsy;
	float	adjy;	/* adjust from v[0]->fy to fsy, scaled */
	int		lines;	/* number of lines to be sampled on this edge */
	fixed_t	fx0;	/* fixed pt X of lower endpoint */
};

struct DrawPoly_t
{
	DrawPoly_t		*prev;
	DrawPoly_t		*next;

	draw_vertex_t	*vb;
	EdgeT			eMaj, eTop, eBot;

	int		ltor;		/* true if scanning left-to-right */
	float dzdx, dzdy;      fixed_t fdzdx;
	float drdx, drdy;      fixed_t fdrdx;
	float dsdx, dsdy;
	float dtdx, dtdy;
	float dvdx, dvdy;

	fixed_t fxLeftEdge, fxRightEdge, fdxLeftEdge, fdxRightEdge;
	fixed_t fError, fdError;
	fixed_t fz, fdzOuter, fdzInner;
	fixed_t fr, fdrOuter, fdrInner;
	float sLeft, dsOuter, dsInner;
	float tLeft, dtOuter, dtInner;
	float vLeft, dvOuter, dvInner;

	int subTriangle;
	int lines;
	int iy;

	byte*			colormap;
	word*			colormap16;
	byte*			source;
	spanfunc_t		spanfunc;
	int				masked;
};

struct span_t
{
	span_t		*next;
	int			x;
	int			count;
	int			fz;
	int			fr;
	float		s;
	float		t;
	float		v;
	int			fdzdx;
	int			fdrdx;
	float		dsdx;
	float		dtdx;
	float		dvdx;

	byte		*colormap;
	word		*colormap16;
	byte*		source;
	spanfunc_t	spanfunc;
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static DrawPoly_t		dpolys[MAX_POLYS];
static draw_vertex_t	dverts[MAX_VERTS];
static int				numpolys;
static int				numverts;

static DrawPoly_t		polylist;
static DrawPoly_t		*polystoadd[MAXSCREENHEIGHT];
static DrawPoly_t		maskedpolylist;
static DrawPoly_t		*maskedpolystoadd[MAXSCREENHEIGHT];

static span_t			drawspans[MAXSCREENWIDTH];
span_t					*freespans;
span_t					*spanlist;
static span_t			maskeddrawspans[MAXSCREENWIDTH];
span_t					*maskedfreespans;
span_t					*maskedspanlist;

static TCvarI			d_background("d_background", "0", CVAR_ARCHIVE);

static int				outofverts;
static int				outofpolys;

// CODE --------------------------------------------------------------------

static void D_DummyDrawer(void)
{
}

//==========================================================================
//
//	AddPoly
//
//==========================================================================

static void AddPoly(DrawPoly_t *poly)
{
	DrawPoly_t *prev = &polylist;
	while (prev->next && prev->next > poly)
	{
		prev = prev->next;
	}
	poly->prev = prev;
	poly->next = prev->next;
	if (prev->next)
	{
		prev->next->prev = poly;
	}
	prev->next = poly;
}

//==========================================================================
//
//	RemovePoly
//
//==========================================================================

static void	RemovePoly(DrawPoly_t *poly)
{
	if (poly->next)
	{
		poly->next->prev = poly->prev;
	}
	poly->prev->next = poly->next;
}

//==========================================================================
//
//	D_SetupSubTriangle
//
//==========================================================================

void D_SetupSubTriangle(DrawPoly_t *poly)
{
	EdgeT *eLeft, *eRight;
	int setupLeft, setupRight;

	if (poly->subTriangle == 0)
	{
		/* bottom half */
		if (poly->ltor)
		{
			eLeft = &poly->eMaj;
			eRight = &poly->eBot;
			poly->lines = eRight->lines;
			setupLeft = 1;
			setupRight = 1;
		}
		else
		{
			eLeft = &poly->eBot;
			eRight = &poly->eMaj;
			poly->lines = eLeft->lines;
			setupLeft = 1;
			setupRight = 1;
		}
	}
	else
	{
		/* top half */
		if (poly->ltor)
		{
			eLeft = &poly->eMaj;
			eRight = &poly->eTop;
			poly->lines = eRight->lines;
			setupLeft = 0;
			setupRight = 1;
		}
		else
		{
			eLeft = &poly->eTop;
			eRight = &poly->eMaj;
			poly->lines = eLeft->lines;
			setupLeft = 1;
			setupRight = 0;
		}
		if (poly->lines == 0) return;
	}

	if (setupLeft && eLeft->lines > 0)
	{
		int vLower;
		fixed_t fx;
		fixed_t fdxOuter;
		fixed_t fsx;
		float dxOuter;
		float adjx, adjy;

		fsx = eLeft->fsx;
		fx = (fsx + FRACUNIT - 1) & 0xffff0000;
		poly->fError = fx - fsx - FRACUNIT;
		poly->fxLeftEdge = fsx - 1;
		poly->fdxLeftEdge = eLeft->fdxdy;
		fdxOuter = (poly->fdxLeftEdge - 1) & 0xffff0000;
		poly->fdError = fdxOuter - poly->fdxLeftEdge + FRACUNIT;
		dxOuter = (float)(fdxOuter >> FRACBITS);

		poly->iy = eLeft->fsy >> FRACBITS;

		adjx = (float)(fx - eLeft->fx0);  /* SCALED! */
		adjy = eLeft->adjy;		 /* SCALED! */

		vLower = eLeft->v0;

              /*
               * Now we need the set of parameter (z, color, etc.) values at
               * the point (fx, fy).  This gives us properly-sampled parameter
                * values that we can step from pixel to pixel.  Furthermore,
                * although we might have intermediate results that overflow
                * the normal parameter range when we step temporarily outside
                * the triangle, we shouldn't overflow or underflow for any
                * pixel that's actually inside the triangle.
                */

		float z0 = (poly->vb[vLower].z <= 1.0F) ? 1.0F : 1 / poly->vb[vLower].z;
		poly->fz = (fixed_t)((z0 * FRACUNIT + poly->dzdx * adjx + poly->dzdy * adjy) * 0x8000);
		poly->fdzOuter = (fixed_t)((poly->dzdy + dxOuter * poly->dzdx) * 0x8000 * 0x10000);


		float invW = 1.0F / poly->vb[vLower].z;

		poly->fr = (fixed_t)(((int)poly->vb[vLower].l << FRACBITS) + poly->drdx * adjx + poly->drdy * adjy)
			+ (FRACUNIT >> 1);
		poly->fdrOuter = FX(poly->drdy + dxOuter * poly->drdx);

		poly->sLeft = poly->vb[vLower].s * invW + (poly->dsdx * adjx + poly->dsdy * adjy) * (1.0F/(float)FRACUNIT);
		poly->dsOuter = poly->dsdy + dxOuter * poly->dsdx;

		poly->tLeft = poly->vb[vLower].t * invW + (poly->dtdx * adjx + poly->dtdy * adjy) * (1.0F/(float)FRACUNIT);
		poly->dtOuter = poly->dtdy + dxOuter * poly->dtdx;

		poly->vLeft = invW + (poly->dvdx * adjx + poly->dvdy * adjy) * (1.0F/(float)FRACUNIT);
		poly->dvOuter = poly->dvdy + dxOuter * poly->dvdx;

		/* Rasterize setup */
		poly->fdzInner = poly->fdzOuter + poly->fdzdx;
		poly->fdrInner = poly->fdrOuter + poly->fdrdx;
	    poly->dsInner = poly->dsOuter + poly->dsdx;
	    poly->dtInner = poly->dtOuter + poly->dtdx;
		poly->dvInner = poly->dvOuter + poly->dvdx;
	} /*if setupLeft*/


	if (setupRight && eRight->lines > 0)
	{
		poly->fxRightEdge = eRight->fsx - 1;
		poly->fdxRightEdge = eRight->fdxdy;
	}
}

//==========================================================================
//
//	D_MakeSpans
//
//==========================================================================

#ifdef USEASM
extern "C" span_t *D_MakeSpans(DrawPoly_t *poly);
#else
span_t *D_MakeSpans(DrawPoly_t *poly)
{
	span_t		*s;

	s = freespans;
	if (!s)
	{
		return NULL;
	}

	// initialize the span interpolants to the leftmost value
	int left = poly->fxLeftEdge >> FRACBITS;
	int right = poly->fxRightEdge >> FRACBITS;

	s->count = right - left;
	if (s->count <= 0)
	{
		return NULL;
	}

	s->x = left;
	s->fz = poly->fz;
	s->fr = poly->fr;
	s->s = poly->sLeft;
	s->t = poly->tLeft;
	s->v = poly->vLeft;
	s->fdzdx = poly->fdzdx;
	s->fdrdx = poly->fdrdx;
	s->dsdx = poly->dsdx;
	s->dtdx = poly->dtdx;
	s->dvdx = poly->dvdx;
	s->colormap = poly->colormap;
	s->colormap16 = poly->colormap16;
	s->source = poly->source;
	s->spanfunc = poly->spanfunc;

	freespans = s->next;

	span_t *prev;
	span_t *next;
	int endx = s->x + s->count;
	prev = NULL;
	next = spanlist;
	while (next && next->x < endx)
	{
		if (next->x + next->count > endx)
		{
			span_t *newspan = freespans;
			if (!newspan)
			{
				break;
			}
			freespans = newspan->next;
			int cnt = endx - next->x;
			newspan->count = next->count - cnt;
			newspan->x = endx;
			newspan->fz = next->fz + cnt * next->fdzdx;
			newspan->fr = next->fr + cnt * next->fdrdx;
			newspan->s = next->s + cnt * next->dsdx;
			newspan->t = next->t + cnt * next->dtdx;
			newspan->v = next->v + cnt * next->dvdx;
			newspan->fdzdx = next->fdzdx;
			newspan->fdrdx = next->fdrdx;
			newspan->dsdx = next->dsdx;
			newspan->dtdx = next->dtdx;
			newspan->dvdx = next->dvdx;
			newspan->colormap = next->colormap;
			newspan->colormap16 = next->colormap16;
			newspan->source = next->source;
			newspan->spanfunc = next->spanfunc;
			newspan->next = next->next;
			next->next = newspan;
			next->count = cnt;
		}

		if (next->x >= s->x)
		{
			span_t *tmp;
			tmp = next;
			next = next->next;
			tmp->next = freespans;
			freespans = tmp;
			if (prev)
			{
				prev->next = next;
			}
		}
		else
		{
			prev = next;
			next = next->next;
		}
	}
	s->next = next;
	if (prev)
	{
		prev->next = s;
		int cnt = s->x - prev->x;
		if (cnt < prev->count)
		{
			prev->count = cnt;
		}
	}
	else
	{
		spanlist = s;
	}
	return s;
}
#endif

//==========================================================================
//
//	D_MakeMaskedSpans
//
//==========================================================================

void D_MakeMaskedSpans(DrawPoly_t *poly)
{
	span_t		*s;

	s = maskedfreespans;
	if (!s)
	{
		return;
	}

	// initialize the span interpolants to the leftmost value
	int left = poly->fxLeftEdge >> FRACBITS;
	int right = poly->fxRightEdge >> FRACBITS;

	s->count = right - left;
	if (s->count <= 0)
	{
		return;
	}

	maskedfreespans = s->next;

	s->x = left;
	s->fz = poly->fz;
	s->fr = poly->fr;
	s->s = poly->sLeft;
	s->t = poly->tLeft;
	s->v = poly->vLeft;
	s->fdzdx = poly->fdzdx;
	s->fdrdx = poly->fdrdx;
	s->dsdx = poly->dsdx;
	s->dtdx = poly->dtdx;
	s->dvdx = poly->dvdx;
	s->colormap = poly->colormap;
	s->colormap16 = poly->colormap16;
	s->source = poly->source;
	s->spanfunc = poly->spanfunc;
	s->next = NULL;

	if (maskedspanlist)
	{
		span_t *prev;
		prev = maskedspanlist;
		while (prev->next)
		{
			prev = prev->next;
		}
		prev->next = s;
	}
	else
	{
		maskedspanlist = s;
	}
}

//==========================================================================
//
//	D_ClipMaskedSpans
//
//==========================================================================

void D_ClipMaskedSpans(span_t *clipspan)
{
	int startx = clipspan->x;
	int endx = clipspan->x + clipspan->count;

	span_t *prev = NULL;
	span_t *next = maskedspanlist;
	while (next)
	{
		if (next->x + next->count < startx ||
			next->x >= endx)
		{
			prev = next;
			next = next->next;
			continue;
		}
		if (next->x + next->count > endx)
		{
			span_t *newspan = maskedfreespans;
			if (!newspan)
			{
				goto no_free_spans;
			}
			maskedfreespans = newspan->next;
			int cnt = endx - next->x;
			newspan->count = next->count - cnt;
			newspan->x = endx;
			newspan->fz = next->fz + cnt * next->fdzdx;
			newspan->fr = next->fr + cnt * next->fdrdx;
			newspan->s = next->s + cnt * next->dsdx;
			newspan->t = next->t + cnt * next->dtdx;
			newspan->v = next->v + cnt * next->dvdx;
			newspan->fdzdx = next->fdzdx;
			newspan->fdrdx = next->fdrdx;
			newspan->dsdx = next->dsdx;
			newspan->dtdx = next->dtdx;
			newspan->dvdx = next->dvdx;
			newspan->colormap = next->colormap;
			newspan->colormap16 = next->colormap16;
			newspan->source = next->source;
			newspan->spanfunc = next->spanfunc;
			newspan->next = next->next;
			next->next = newspan;
			next->count = cnt;
		}
	 no_free_spans:
		if (next->x >= startx)
		{
			span_t *tmp;
			tmp = next;
			next = next->next;
			tmp->next = maskedfreespans;
			maskedfreespans = tmp;
			if (prev)
			{
				prev->next = next;
			}
			else
			{
				maskedspanlist = next;
			}
		}
		else
		{
			if (next->x + next->count > startx)
			{
				next->count = startx - next->x;
			}
			prev = next;
			next = next->next;
		}
	}
}

//==========================================================================
//
//	D_DrawPolyLine
//
//==========================================================================

void D_DrawPolyLine(DrawPoly_t *poly)
{
	if (poly->masked)
	{
		D_MakeMaskedSpans(poly);
	}
	else
	{
		span_t *s = D_MakeSpans(poly);
		if (s && maskedspanlist)
		{
			D_ClipMaskedSpans(s);
		}
	}

	//	Advance to the next scan line. Compute the new edge coordinates, and
	// adjust the pixel-center x coordinate so that it stays on or inside
	// the major edge.

	poly->iy++;
	poly->lines--;

	poly->fxLeftEdge += poly->fdxLeftEdge;
	poly->fxRightEdge += poly->fdxRightEdge;

	poly->fError += poly->fdError;
	if (poly->fError >= 0)
	{
		poly->fError -= FRACUNIT;
		poly->fz += poly->fdzOuter;
		poly->fr += poly->fdrOuter;
		poly->sLeft += poly->dsOuter;
		poly->tLeft += poly->dtOuter;
		poly->vLeft += poly->dvOuter;
	}
	else
	{
		poly->fz += poly->fdzInner;
		poly->fr += poly->fdrInner;
		poly->sLeft += poly->dsInner;
		poly->tLeft += poly->dtInner;
		poly->vLeft += poly->dvInner;
	}

	if (!poly->lines)
	{
		poly->subTriangle++;
		if (poly->subTriangle == 2)
		{
			Z_ChangeTag(poly->source, PU_CACHE);
			RemovePoly(poly);
			return;
		}

		D_SetupSubTriangle(poly);
		if (!poly->lines)
		{
			Z_ChangeTag(poly->source, PU_CACHE);
			RemovePoly(poly);
		}
	}
}

#if 0
static TVec UnProjectVertex(const TVec &in)
{
	return TVec((in.x - centerxfrac) / xprojection * in.z,
				(in.y - centeryfrac) / yprojection * in.z,
				in.z);
}

void TransformVector(const TVec &in, TVec &out);

float GetS(const draw_vertex_t &p)
{
	TVec tr_saxis;
	TVec tr_taxis;
	TVec unp;
	float sdist = DotProduct(r_texorg - vieworg, r_saxis);

	TransformVector(r_saxis, tr_saxis);
	TransformVector(r_taxis, tr_taxis);
	unp = UnProjectVertex(p);

	return DotProduct(unp, tr_saxis) - sdist;
}

float GetT(const draw_vertex_t &p)
{
	TVec tr_saxis;
	TVec tr_taxis;
	TVec unp;
	float tdist = DotProduct(r_texorg - vieworg, r_taxis);

	TransformVector(r_saxis, tr_saxis);
	TransformVector(r_taxis, tr_taxis);
	unp = UnProjectVertex(p);

	return DotProduct(unp, tr_taxis) - tdist;
}
#endif

//==========================================================================
//
//	D_SetupTriangle
//
//	Triangle Rasterizer
//
//	This code was designed for the origin to be in the lower-left corner.
//
//	Based on sources of MESA by Brian Paul where it's inspired by triangle
// rasterizer code written by Allen Akin.
//
//==========================================================================

void D_SetupTriangle(draw_vertex_t *vb, int v0, int v1, int v2, int masked)
{
	if (numpolys >= MAX_POLYS)
	{
		outofpolys++;
		return;
	}
	DrawPoly_t *poly = &dpolys[numpolys];

	poly->vb = vb;

	int vMin, vMid, vMax;       /* vertex indexes:  Y(vMin)<=Y(vMid)<=Y(vMax) */
	float oneOverArea;

	/* find the order of the 3 vertices along the Y axis */
	float y0 = vb[v0].y;
	float y1 = vb[v1].y;
	float y2 = vb[v2].y;

	if (y0 <= y1)
	{
		if (y1 <= y2)
		{
			vMin = v0;   vMid = v1;   vMax = v2;   /* y0<=y1<=y2 */
		}
		else if (y2 <= y0)
		{
			vMin = v2;   vMid = v0;   vMax = v1;   /* y2<=y0<=y1 */
		}
		else
		{
			vMin = v0;   vMid = v2;   vMax = v1;   /* y0<=y2<=y1 */
		}
	}
	else
	{
		if (y0 <= y2)
		{
			vMin = v1;   vMid = v0;   vMax = v2;   /* y1<=y0<=y2 */
		}
		else if (y2 <= y1)
		{
			vMin = v2;   vMid = v1;   vMax = v0;   /* y2<=y1<=y0 */
		}
		else
		{
			vMin = v1;   vMid = v2;   vMax = v0;   /* y1<=y2<=y0 */
		}
	}

	/* vertex/edge relationship */
	EdgeT &eMaj = poly->eMaj;
	EdgeT &eTop = poly->eTop;
	EdgeT &eBot = poly->eBot;
	eMaj.v0 = vMin;
	eTop.v0 = vMid;
	eBot.v0 = vMin;

	/* compute deltas for each edge:  vertex[v1] - vertex[v0] */
	eMaj.dx = vb[vMax].x - vb[vMin].x;
	eMaj.dy = vb[vMax].y - vb[vMin].y;
	eTop.dx = vb[vMax].x - vb[vMid].x;
	eTop.dy = vb[vMax].y - vb[vMid].y;
	eBot.dx = vb[vMid].x - vb[vMin].x;
	eBot.dy = vb[vMid].y - vb[vMin].y;

	/* compute oneOverArea */
	float area = eMaj.dx * eBot.dy - eBot.dx * eMaj.dy;
	if (area > -0.05f && area < 0.05f)
	{
		return;  /* very small; CULLED */
	}
	oneOverArea = 1.0F / area;

	/* Edge setup.  For a triangle strip these could be reused... */
	/* fixed point Y coordinates */
	fixed_t vMin_fx = FX(vb[vMin].x + 0.5F);
	fixed_t vMin_fy = FX(vb[vMin].y - 0.5F);
	fixed_t vMid_fx = FX(vb[vMid].x + 0.5F);
	fixed_t vMid_fy = FX(vb[vMid].y - 0.5F);
	fixed_t vMax_fy = FX(vb[vMax].y - 0.5F);

	eMaj.fsy = (vMin_fy + FRACUNIT - 1) & 0xffff0000;
	eMaj.lines = (vMax_fy + FRACUNIT - 1 - eMaj.fsy) >> FRACBITS;
	if (eMaj.lines > 0)
	{
		float dxdy = eMaj.dx / eMaj.dy;
		eMaj.fdxdy = FX(dxdy);
		eMaj.adjy = (float) (eMaj.fsy - vMin_fy);  /* SCALED! */
		eMaj.fx0 = vMin_fx;
		eMaj.fsx = eMaj.fx0 + (fixed_t) (eMaj.adjy * dxdy);
	}
	else
	{
		return;  /*CULLED*/
	}

	eTop.fsy = (vMid_fy + FRACUNIT - 1) & 0xffff0000;
	eTop.lines = (vMax_fy + FRACUNIT - 1 - eTop.fsy) >> FRACBITS;
	if (eTop.lines > 0)
	{
		float dxdy = eTop.dx / eTop.dy;
		eTop.fdxdy = FX(dxdy);
		eTop.adjy = (float) (eTop.fsy - vMid_fy); /* SCALED! */
		eTop.fx0 = vMid_fx;
		eTop.fsx = eTop.fx0 + (fixed_t) (eTop.adjy * dxdy);
	}

	eBot.fsy = (vMin_fy + FRACUNIT - 1) & 0xffff0000;
	eBot.lines = (vMid_fy + FRACUNIT - 1 - eBot.fsy) >> FRACBITS;
	if (eBot.lines > 0)
	{
		float dxdy = eBot.dx / eBot.dy;
		eBot.fdxdy = FX(dxdy);
		eBot.adjy = (float) (eBot.fsy - vMin_fy);  /* SCALED! */
		eBot.fx0 = vMin_fx;
		eBot.fsx = eBot.fx0 + (fixed_t) (eBot.adjy * dxdy);
	}

   /*
    * Conceptually, we view a triangle as two subtriangles
    * separated by a perfectly horizontal line.  The edge that is
    * intersected by this line is one with maximal absolute dy; we
    * call it a ``major'' edge.  The other two edges are the
    * ``top'' edge (for the upper subtriangle) and the ``bottom''
    * edge (for the lower subtriangle).  If either of these two
    * edges is horizontal or very close to horizontal, the
    * corresponding subtriangle might cover zero sample points;
    * we take care to handle such cases, for performance as well
    * as correctness.
    *
    * By stepping rasterization parameters along the major edge,
    * we can avoid recomputing them at the discontinuity where
    * the top and bottom edges meet.  However, this forces us to
    * be able to scan both left-to-right and right-to-left. 
    * Also, we must determine whether the major edge is at the
    * left or right side of the triangle.  We do this by
    * computing the magnitude of the cross-product of the major
    * and top edges.  Since this magnitude depends on the sine of
    * the angle between the two edges, its sign tells us whether
    * we turn to the left or to the right when travelling along
    * the major edge to the top edge, and from this we infer
    * whether the major edge is on the left or the right.
    *
    * Serendipitously, this cross-product magnitude is also a
    * value we need to compute the iteration parameter
    * derivatives for the triangle, and it can be used to perform
    * backface culling because its sign tells us whether the
    * triangle is clockwise or counterclockwise.  In this code we
    * refer to it as ``area'' because it's also proportional to
    * the pixel area of the triangle.
    */

	poly->ltor = (oneOverArea < 0.0F);

	/* compute d?/dx and d?/dy derivatives */
	float izMax = (vb[vMax].z <= 1.0F) ? 1.0F : 1.0F / vb[vMax].z;
	float izMin = (vb[vMin].z <= 1.0F) ? 1.0F : 1.0F / vb[vMin].z;
	float izMid = (vb[vMid].z <= 1.0F) ? 1.0F : 1.0F / vb[vMid].z;

	float eMaj_dz, eBot_dz;
	eMaj_dz = izMax - izMin;
	eBot_dz = izMid - izMin;
	poly->dzdx = oneOverArea * (eMaj_dz * eBot.dy - eMaj.dy * eBot_dz);
	poly->dzdy = oneOverArea * (eMaj.dx * eBot_dz - eMaj_dz * eBot.dx);
	poly->fdzdx = (fixed_t)(poly->dzdx * 0x8000 * 0x10000);

	float eMaj_dr, eBot_dr;
	eMaj_dr = (int)vb[vMax].l - (int)vb[vMin].l;
	eBot_dr = (int)vb[vMid].l - (int)vb[vMin].l;
	poly->drdx = oneOverArea * (eMaj_dr * eBot.dy - eMaj.dy * eBot_dr);
	poly->fdrdx = FX(poly->drdx);
	poly->drdy = oneOverArea * (eMaj.dx * eBot_dr - eMaj_dr * eBot.dx);


	float wMax = 1.0F / vb[vMax].z;
	float wMin = 1.0F / vb[vMin].z;
	float wMid = 1.0F / vb[vMid].z;

#if 0
	TVec tr_saxis;
	TVec tr_taxis;
	float sdist = DotProduct(r_texorg - vieworg, r_saxis);
	float tdist = DotProduct(r_texorg - vieworg, r_taxis);

	TransformVector(r_saxis, tr_saxis);
	TransformVector(r_taxis, tr_taxis);

	float eMaj_ds =
		(vb[vMax].x - vb[vMin].x) / xprojection * tr_saxis.x
		+
		(vb[vMax].y - vb[vMin].y) / yprojection * tr_saxis.y
		-
		sdist * (1.0 / vb[vMax].z - 1.0 / vb[vMin].z);
	float eBot_ds =
		(vb[vMid].x - vb[vMin].x) / xprojection * tr_saxis.x
		+
		(vb[vMid].y - vb[vMin].y) / yprojection * tr_saxis.y
		-
		sdist * (1.0 / vb[vMid].z - 1.0 / vb[vMin].z);
	poly->dsdx = oneOverArea * (eMaj_ds * eBot.dy - eMaj.dy * eBot_ds);
	poly->dsdy = oneOverArea * (eMaj.dx * eBot_ds - eMaj_ds * eBot.dx);
//	poly->dsdx = tr_saxis.x / xprojection;
//	poly->dsdy = tr_saxis.y / yprojection;

	float eMaj_dt =
		(vb[vMax].x - vb[vMin].x) / xprojection * tr_taxis.x
		+
		(vb[vMax].y - vb[vMin].y) / yprojection * tr_taxis.y
		-
		tdist * (1.0 / vb[vMax].z - 1.0 / vb[vMin].z);
	float eBot_dt =
		(vb[vMid].x - vb[vMin].x) / xprojection * tr_taxis.x
		+
		(vb[vMid].y - vb[vMin].y) / yprojection * tr_taxis.y
		-
		tdist * (1.0 / vb[vMid].z - 1.0 / vb[vMin].z);
	poly->dtdx = oneOverArea * (eMaj_dt * eBot.dy - eMaj.dy * eBot_dt);
	poly->dtdy = oneOverArea * (eMaj.dx * eBot_dt - eMaj_dt * eBot.dx);
//	poly->dtdx = tr_taxis.x / xprojection;
//	poly->dtdy = tr_taxis.y / yprojection;

	float eMaj_dv = wMax - wMin;
	float eBot_dv = wMid - wMin;
	poly->dvdx = oneOverArea * (eMaj_dv * eBot.dy - eMaj.dy * eBot_dv);
	poly->dvdy = oneOverArea * (eMaj.dx * eBot_dv - eMaj_dv * eBot.dx);
#else
	float eMaj_ds = vb[vMax].s * wMax - vb[vMin].s * wMin;
	float eBot_ds = vb[vMid].s * wMid - vb[vMin].s * wMin;
	poly->dsdx = oneOverArea * (eMaj_ds * eBot.dy - eMaj.dy * eBot_ds);
	poly->dsdy = oneOverArea * (eMaj.dx * eBot_ds - eMaj_ds * eBot.dx);

	float eMaj_dt = vb[vMax].t * wMax - vb[vMin].t * wMin;
	float eBot_dt = vb[vMid].t * wMid - vb[vMin].t * wMin;
	poly->dtdx = oneOverArea * (eMaj_dt * eBot.dy - eMaj.dy * eBot_dt);
	poly->dtdy = oneOverArea * (eMaj.dx * eBot_dt - eMaj_dt * eBot.dx);

	float eMaj_dv = wMax - wMin;
	float eBot_dv = wMid - wMin;
	poly->dvdx = oneOverArea * (eMaj_dv * eBot.dy - eMaj.dy * eBot_dv);
	poly->dvdy = oneOverArea * (eMaj.dx * eBot_dv - eMaj_dv * eBot.dx);
#endif

      /*
       * We always sample at pixel centers.  However, we avoid
       * explicit half-pixel offsets in this code by incorporating
       * the proper offset in each of x and y during the
       * transformation to window coordinates.
       *
       * We also apply the usual rasterization rules to prevent
       * cracks and overlaps.  A pixel is considered inside a
       * subtriangle if it meets all of four conditions: it is on or
       * to the right of the left edge, strictly to the left of the
       * right edge, on or below the top edge, and strictly above
       * the bottom edge.  (Some edges may be degenerate.)
       *
       * The following discussion assumes left-to-right scanning
       * (that is, the major edge is on the left); the right-to-left
       * case is a straightforward variation.
       *
       * We start by finding the half-integral y coordinate that is
       * at or below the top of the triangle.  This gives us the
       * first scan line that could possibly contain pixels that are
       * inside the triangle.
       *
       * Next we creep down the major edge until we reach that y,
       * and compute the corresponding x coordinate on the edge. 
       * Then we find the half-integral x that lies on or just
       * inside the edge.  This is the first pixel that might lie in
       * the interior of the triangle.  (We won't know for sure
       * until we check the other edges.)
       *
       * As we rasterize the triangle, we'll step down the major
       * edge.  For each step in y, we'll move an integer number
       * of steps in x.  There are two possible x step sizes, which
       * we'll call the ``inner'' step (guaranteed to land on the
       * edge or inside it) and the ``outer'' step (guaranteed to
       * land on the edge or outside it).  The inner and outer steps
       * differ by one.  During rasterization we maintain an error
       * term that indicates our distance from the true edge, and
       * select either the inner step or the outer step, whichever
       * gets us to the first pixel that falls inside the triangle.
       *
       * All parameters (z, red, etc.) as well as the buffer
       * addresses for color and z have inner and outer step values,
       * so that we can increment them appropriately.  This method
       * eliminates the need to adjust parameters by creeping a
       * sub-pixel amount into the triangle at each scanline.
       */

	poly->subTriangle = 0;
	D_SetupSubTriangle(poly);
	if (!poly->lines)
	{
		poly->subTriangle++;
		D_SetupSubTriangle(poly);
	}

	poly->colormap = ds_colormap;
	poly->colormap16 = ds_colormap16;
	poly->source = ds_source;
	poly->spanfunc = spanfunc;
	poly->masked = masked;

	poly->next = polystoadd[poly->iy];
	polystoadd[poly->iy] = poly;

	numpolys++;
	Z_ChangeTag(ds_source, PU_STATIC);
}

//==========================================================================
//
//	D_ClearPolys
//
//==========================================================================

void D_ClearPolys(void)
{
	numpolys = 0;
	numverts = 0;
	polylist.next = NULL;
	memset(polystoadd, 0, sizeof(polystoadd));
	maskedpolylist.next = NULL;
	memset(maskedpolystoadd, 0, sizeof(maskedpolystoadd));
	outofverts = 0;
	outofpolys = 0;
}

//==========================================================================
//
//	D_DrawWorldPolygon
//
//==========================================================================

void D_DrawWorldPolygon(draw_vertex_t *verts, int count, int masked)
{
	if (numverts + count > MAX_VERTS)
	{
		outofverts += count;
		return;
	}
	memcpy(dverts + numverts, verts, count * sizeof(draw_vertex_t));
	for (int i = 2; i < count; i++)
	{
		D_SetupTriangle(dverts + numverts, 0, i - 1, i, masked);
	}
	numverts += count;
}

//==========================================================================
//
//	D_DrawSpans
//
//==========================================================================

static span_t *D_DrawSpans(span_t *s)
{
	span_t	*last = NULL;
	while (s)
	{
		ds_x = s->x;
		ds_count = s->count;
		ds_fz = s->fz;
		ds_fr = s->fr;
		ds_s = s->s;
		ds_t = s->t;
		ds_v = s->v;
		ds_fdzdx = s->fdzdx;
		ds_fdrdx = s->fdrdx;
		ds_dsdx = s->dsdx;
		ds_dtdx = s->dtdx;
		ds_dvdx = s->dvdx;
		ds_colormap = s->colormap;
		ds_colormap16 = s->colormap16;
		ds_source = s->source;

		// need this to accomodate round-off errors
		fixed_t ffrend = ds_fr + (ds_count - 1) * ds_fdrdx;
		if (ffrend < 0) ds_fr -= ffrend;
		if (ds_fr < 0) ds_fr = 0;

		s->spanfunc();
		last = s;
		s = s->next;
	}
	return last;
}

//==========================================================================
//
//	TSoftwareDrawer::WorldDrawing
//
//==========================================================================

void TSoftwareDrawer::WorldDrawing(void)
{
	span_t *s;
	DrawPoly_t *poly;

	if (!freespans)
	{
		for (int i = 0; i < MAXSCREENWIDTH; i++)
		{
			drawspans[i].next = freespans;
			freespans = &drawspans[i];
		}
	}
	if (!maskedfreespans)
	{
		for (int i = 0; i < MAXSCREENWIDTH; i++)
		{
			maskeddrawspans[i].next = maskedfreespans;
			maskedfreespans = &maskeddrawspans[i];
		}
	}

	for (ds_y = 0; ds_y < viewheight; ds_y++)
	{
		if (d_background)
		{
			s = freespans;
			freespans = s->next;
			s->next = NULL;
			s->count = viewwidth;
			s->x = 0;
			s->spanfunc = D_DummyDrawer;
			spanlist = s;
		}
		else
		{
			spanlist = NULL;
		}
		maskedspanlist = NULL;

		poly = polystoadd[ds_y];
		while (poly)
		{
			DrawPoly_t *next;
			next = poly->next;
			AddPoly(poly);
			poly = next;
		}

		poly = polylist.next;
		while (poly)
		{
			DrawPoly_t *next = poly->next;
			D_DrawPolyLine(poly);
			poly = next;
		}

		s = D_DrawSpans(spanlist);
		if (s)
		{
			s->next = freespans;
			freespans = spanlist;
		}
		s = D_DrawSpans(maskedspanlist);
		if (s)
		{
			s->next = maskedfreespans;
			maskedfreespans = maskedspanlist;
		}
	}
	if (outofverts) cond << "Out of " << outofverts << " verts\n";
	if (outofpolys) cond << "Out of " << outofpolys << " polys\n";
}


