//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**	
//**	The actual span/column drawing functions.
//**
//**	All drawing to the view buffer is accomplished in this file.
//**  The other refresh files only know about ccordinates, not the
//**  architecture of the frame buffer.
//**
//**	Here find the main potential for optimization, e.g. inline assembly,
//**  different algorithms.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int				ds_x;
int				ds_y;
fixed_t			ds_fz;
fixed_t			ds_fr;
float			ds_s;
float			ds_t;
float			ds_v;
int				ds_count;
fixed_t			ds_fdzdx;
fixed_t			ds_fdrdx;
float			ds_dsdx;
float			ds_dtdx;
float			ds_dvdx;
byte*			ds_colormap;
word*			ds_colormap16;
byte*			ds_source;
int				ds_tex_width;
int				ds_tex_shift;
byte*			ds_transluc;// For translucent spans
int				ds_transluc16;

int				ylookup[MAXSCREENHEIGHT];

word			*zbuffer = NULL;

spanfunc_t		D_DrawSpan;
spanfunc_t		D_DrawMaskedSpan;
spanfunc_t		D_DrawFuzzSpan;
spanfunc_t		D_DrawAltFuzzSpan;
spanfunc_t		D_DrawSkySpan;
spanfunc_t		D_DrawDoubleSkySpan;
spanfunc_t		D_DrawSpriteSpan;
spanfunc_t		D_DrawFuzzSpriteSpan;
spanfunc_t		D_DrawAltFuzzSpriteSpan;

spanfunc_t		spanfunc;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

#ifndef USEASM

//==========================================================================
//
//	D_DrawZSpan
//
//==========================================================================

extern "C" void D_DrawZSpan(void)
{
    word *zdst = zbuffer + (ylookup[ds_y] + ds_x);
	int count = ds_count;

	fixed_t ffz = ds_fz;
	fixed_t fdzdx = ds_fdzdx;

	if ((long)zdst & 0x02)
	{
		//	Align dest to dword
		*zdst++ = ffz >> 16;
		ffz += fdzdx;
		count--;
	}

	int doublecount = count >> 1;
	if (doublecount > 0)
	{
		//	Process pixels by 2 pixels at time
		do
	    {
			dword zval;

			zval = ffz >> 16;
			ffz += fdzdx;
			zval |= ffz & 0xffff0000;
			ffz += fdzdx;

			*(dword*)zdst = zval;
			zdst += 2;
		} while (--doublecount);
	}

	if (count & 1)
    {
		//	Process remaining pixel (if there is any)
		*zdst = ffz >> 16;
	}
}

#endif	//	USEASM

//==========================================================================
//
//	D_DrawTriangle
//
//	Triangle Rasterizer
//
//	This code was designed for the origin to be in the lower-left corner.
//
//	Based on sources of MESA by Brian Paul where it's inspired by triangle
// rasterizer code written by Allen Akin.
//
//==========================================================================

extern "C" void D_DrawTriangle(draw_vertex_t *vb, int v0, int v1, int v2)
{
	struct EdgeT
	{
		int		v0, v1;   /* Y(v0) < Y(v1) */
		float	dx;	/* X(v1) - X(v0) */
		float	dy;	/* Y(v1) - Y(v0) */
		fixed_t	fdxdy;	/* dx/dy in fixed-point */
		fixed_t	fsx;	/* first sample point x coord */
		fixed_t	fsy;
		float	adjy;	/* adjust from v[0]->fy to fsy, scaled */
		int		lines;	/* number of lines to be sampled on this edge */
		fixed_t	fx0;	/* fixed pt X of lower endpoint */
	};
	EdgeT eMaj, eTop, eBot;
	int vMin, vMid, vMax;       /* vertex indexes:  Y(vMin)<=Y(vMid)<=Y(vMax) */
	float oneOverArea;

	/* find the order of the 3 vertices along the Y axis */
	float y0 = vb[v0].y;
	float y1 = vb[v1].y;
	float y2 = vb[v2].y;

	if (y0 <= y1)
	{
		if (y1 <= y2)
		{
			vMin = v0;   vMid = v1;   vMax = v2;   /* y0<=y1<=y2 */
		}
		else if (y2 <= y0)
		{
			vMin = v2;   vMid = v0;   vMax = v1;   /* y2<=y0<=y1 */
		}
		else
		{
			vMin = v0;   vMid = v2;   vMax = v1;   /* y0<=y2<=y1 */
		}
	}
	else
	{
		if (y0 <= y2)
		{
			vMin = v1;   vMid = v0;   vMax = v2;   /* y1<=y0<=y2 */
		}
		else if (y2 <= y1)
		{
			vMin = v2;   vMid = v1;   vMax = v0;   /* y2<=y1<=y0 */
		}
		else
		{
			vMin = v1;   vMid = v2;   vMax = v0;   /* y1<=y2<=y0 */
		}
	}

	/* vertex/edge relationship */
	eMaj.v0 = vMin;   eMaj.v1 = vMax;   /*TODO: .v1's not needed */
	eTop.v0 = vMid;   eTop.v1 = vMax;
	eBot.v0 = vMin;   eBot.v1 = vMid;

	/* compute deltas for each edge:  vertex[v1] - vertex[v0] */
	eMaj.dx = vb[vMax].x - vb[vMin].x;
	eMaj.dy = vb[vMax].y - vb[vMin].y;
	eTop.dx = vb[vMax].x - vb[vMid].x;
	eTop.dy = vb[vMax].y - vb[vMid].y;
	eBot.dx = vb[vMid].x - vb[vMin].x;
	eBot.dy = vb[vMid].y - vb[vMin].y;

	/* compute oneOverArea */
	float area = eMaj.dx * eBot.dy - eBot.dx * eMaj.dy;
	if (area > -0.05f && area < 0.05f)
	{
		return;  /* very small; CULLED */
	}
	oneOverArea = 1.0F / area;

	/* Edge setup.  For a triangle strip these could be reused... */
	/* fixed point Y coordinates */
	fixed_t vMin_fx = FX(vb[vMin].x + 0.5F);
	fixed_t vMin_fy = FX(vb[vMin].y - 0.5F);
	fixed_t vMid_fx = FX(vb[vMid].x + 0.5F);
	fixed_t vMid_fy = FX(vb[vMid].y - 0.5F);
	fixed_t vMax_fy = FX(vb[vMax].y - 0.5F);

	eMaj.fsy = (vMin_fy + FRACUNIT - 1) & 0xffff0000;
	eMaj.lines = (vMax_fy + FRACUNIT - 1 - eMaj.fsy) >> FRACBITS;
	if (eMaj.lines > 0)
	{
		float dxdy = eMaj.dx / eMaj.dy;
		eMaj.fdxdy = FX(dxdy);
		eMaj.adjy = (float) (eMaj.fsy - vMin_fy);  /* SCALED! */
		eMaj.fx0 = vMin_fx;
		eMaj.fsx = eMaj.fx0 + (fixed_t) (eMaj.adjy * dxdy);
	}
	else
	{
		return;  /*CULLED*/
	}

	eTop.fsy = (vMid_fy + FRACUNIT - 1) & 0xffff0000;
	eTop.lines = (vMax_fy + FRACUNIT - 1 - eTop.fsy) >> FRACBITS;
	if (eTop.lines > 0)
	{
		float dxdy = eTop.dx / eTop.dy;
		eTop.fdxdy = FX(dxdy);
		eTop.adjy = (float) (eTop.fsy - vMid_fy); /* SCALED! */
		eTop.fx0 = vMid_fx;
		eTop.fsx = eTop.fx0 + (fixed_t) (eTop.adjy * dxdy);
	}

	eBot.fsy = (vMin_fy + FRACUNIT - 1) & 0xffff0000;
	eBot.lines = (vMid_fy + FRACUNIT - 1 - eBot.fsy) >> FRACBITS;
	if (eBot.lines > 0)
	{
		float dxdy = eBot.dx / eBot.dy;
		eBot.fdxdy = FX(dxdy);
		eBot.adjy = (float) (eBot.fsy - vMin_fy);  /* SCALED! */
		eBot.fx0 = vMin_fx;
		eBot.fsx = eBot.fx0 + (fixed_t) (eBot.adjy * dxdy);
	}

   /*
    * Conceptually, we view a triangle as two subtriangles
    * separated by a perfectly horizontal line.  The edge that is
    * intersected by this line is one with maximal absolute dy; we
    * call it a ``major'' edge.  The other two edges are the
    * ``top'' edge (for the upper subtriangle) and the ``bottom''
    * edge (for the lower subtriangle).  If either of these two
    * edges is horizontal or very close to horizontal, the
    * corresponding subtriangle might cover zero sample points;
    * we take care to handle such cases, for performance as well
    * as correctness.
    *
    * By stepping rasterization parameters along the major edge,
    * we can avoid recomputing them at the discontinuity where
    * the top and bottom edges meet.  However, this forces us to
    * be able to scan both left-to-right and right-to-left. 
    * Also, we must determine whether the major edge is at the
    * left or right side of the triangle.  We do this by
    * computing the magnitude of the cross-product of the major
    * and top edges.  Since this magnitude depends on the sine of
    * the angle between the two edges, its sign tells us whether
    * we turn to the left or to the right when travelling along
    * the major edge to the top edge, and from this we infer
    * whether the major edge is on the left or the right.
    *
    * Serendipitously, this cross-product magnitude is also a
    * value we need to compute the iteration parameter
    * derivatives for the triangle, and it can be used to perform
    * backface culling because its sign tells us whether the
    * triangle is clockwise or counterclockwise.  In this code we
    * refer to it as ``area'' because it's also proportional to
    * the pixel area of the triangle.
    */

	int		ltor;		/* true if scanning left-to-right */
	float dzdx, dzdy;      fixed_t fdzdx;
	float drdx, drdy;      fixed_t fdrdx;
//	float dldx, dldy;
	float dsdx, dsdy;
	float dtdx, dtdy;
	float dvdx, dvdy;

	ltor = (oneOverArea < 0.0F);

	/* compute d?/dx and d?/dy derivatives */
	float izMax = (vb[vMax].z <= 1.0F) ? 1.0F : 1.0F / vb[vMax].z;
	float izMin = (vb[vMin].z <= 1.0F) ? 1.0F : 1.0F / vb[vMin].z;
	float izMid = (vb[vMid].z <= 1.0F) ? 1.0F : 1.0F / vb[vMid].z;

	float eMaj_dz, eBot_dz;
	eMaj_dz = izMax - izMin;
	eBot_dz = izMid - izMin;
	dzdx = oneOverArea * (eMaj_dz * eBot.dy - eMaj.dy * eBot_dz);
	dzdy = oneOverArea * (eMaj.dx * eBot_dz - eMaj_dz * eBot.dx);
	fdzdx = (fixed_t)(dzdx * 0x8000 * 0x10000);

	float eMaj_dr, eBot_dr;
	eMaj_dr = (int)vb[vMax].l - (int)vb[vMin].l;
	eBot_dr = (int)vb[vMid].l - (int)vb[vMin].l;
	drdx = oneOverArea * (eMaj_dr * eBot.dy - eMaj.dy * eBot_dr);
	fdrdx = FX(drdx);
	drdy = oneOverArea * (eMaj.dx * eBot_dr - eMaj_dr * eBot.dx);


	float wMax = 1.0F / vb[vMax].z;
	float wMin = 1.0F / vb[vMin].z;
	float wMid = 1.0F / vb[vMid].z;

//	float eMaj_dl, eBot_dl;
//	eMaj_dl = vb[vMax].l * wMax - vb[vMin].l * wMin;
//	eBot_dl = vb[vMid].l * wMid - vb[vMin].l * wMin;
//	dldx = oneOverArea * (eMaj_dl * eBot.dy - eMaj.dy * eBot_dl);
//	dldy = oneOverArea * (eMaj.dx * eBot_dl - eMaj_dl * eBot.dx);

	float eMaj_ds, eBot_ds;
	float eMaj_dt, eBot_dt;
	float eMaj_dv, eBot_dv;

	eMaj_ds = vb[vMax].s * wMax - vb[vMin].s * wMin;
	eBot_ds = vb[vMid].s * wMid - vb[vMin].s * wMin;
	dsdx = oneOverArea * (eMaj_ds * eBot.dy - eMaj.dy * eBot_ds);
	dsdy = oneOverArea * (eMaj.dx * eBot_ds - eMaj_ds * eBot.dx);

	eMaj_dt = vb[vMax].t * wMax - vb[vMin].t * wMin;
	eBot_dt = vb[vMid].t * wMid - vb[vMin].t * wMin;
	dtdx = oneOverArea * (eMaj_dt * eBot.dy - eMaj.dy * eBot_dt);
	dtdy = oneOverArea * (eMaj.dx * eBot_dt - eMaj_dt * eBot.dx);

	eMaj_dv = wMax - wMin;
	eBot_dv = wMid - wMin;
	dvdx = oneOverArea * (eMaj_dv * eBot.dy - eMaj.dy * eBot_dv);
	dvdy = oneOverArea * (eMaj.dx * eBot_dv - eMaj_dv * eBot.dx);

      /*
       * We always sample at pixel centers.  However, we avoid
       * explicit half-pixel offsets in this code by incorporating
       * the proper offset in each of x and y during the
       * transformation to window coordinates.
       *
       * We also apply the usual rasterization rules to prevent
       * cracks and overlaps.  A pixel is considered inside a
       * subtriangle if it meets all of four conditions: it is on or
       * to the right of the left edge, strictly to the left of the
       * right edge, on or below the top edge, and strictly above
       * the bottom edge.  (Some edges may be degenerate.)
       *
       * The following discussion assumes left-to-right scanning
       * (that is, the major edge is on the left); the right-to-left
       * case is a straightforward variation.
       *
       * We start by finding the half-integral y coordinate that is
       * at or below the top of the triangle.  This gives us the
       * first scan line that could possibly contain pixels that are
       * inside the triangle.
       *
       * Next we creep down the major edge until we reach that y,
       * and compute the corresponding x coordinate on the edge. 
       * Then we find the half-integral x that lies on or just
       * inside the edge.  This is the first pixel that might lie in
       * the interior of the triangle.  (We won't know for sure
       * until we check the other edges.)
       *
       * As we rasterize the triangle, we'll step down the major
       * edge.  For each step in y, we'll move an integer number
       * of steps in x.  There are two possible x step sizes, which
       * we'll call the ``inner'' step (guaranteed to land on the
       * edge or inside it) and the ``outer'' step (guaranteed to
       * land on the edge or outside it).  The inner and outer steps
       * differ by one.  During rasterization we maintain an error
       * term that indicates our distance from the true edge, and
       * select either the inner step or the outer step, whichever
       * gets us to the first pixel that falls inside the triangle.
       *
       * All parameters (z, red, etc.) as well as the buffer
       * addresses for color and z have inner and outer step values,
       * so that we can increment them appropriately.  This method
       * eliminates the need to adjust parameters by creeping a
       * sub-pixel amount into the triangle at each scanline.
       */

	int subTriangle;
	fixed_t fx, fxLeftEdge = 0, fxRightEdge = 0, fdxLeftEdge = 0, fdxRightEdge = 0;
	fixed_t fdxOuter;
	int idxOuter;
	float dxOuter;
	fixed_t fError = 0, fdError = 0;
	float adjx, adjy;
	fixed_t fy;
	int iy = 0;
	fixed_t fz = 0, fdzOuter = 0, fdzInner;
	fixed_t fr = 0, fdrOuter = 0, fdrInner;
//	float lLeft = 0, dlOuter = 0, dlInner;
	float sLeft = 0, dsOuter = 0, dsInner;
	float tLeft = 0, dtOuter = 0, dtInner;
	float vLeft = 0, dvOuter = 0, dvInner;

	for (subTriangle = 0; subTriangle <= 1; subTriangle++)
	{
		EdgeT *eLeft, *eRight;
		int setupLeft, setupRight;
		int lines;

		if (subTriangle == 0)
		{
			/* bottom half */
			if (ltor)
			{
				eLeft = &eMaj;
				eRight = &eBot;
				lines = eRight->lines;
				setupLeft = 1;
				setupRight = 1;
			}
			else
			{
				eLeft = &eBot;
				eRight = &eMaj;
				lines = eLeft->lines;
				setupLeft = 1;
				setupRight = 1;
			}
		}
		else
		{
			/* top half */
			if (ltor)
			{
				eLeft = &eMaj;
				eRight = &eTop;
				lines = eRight->lines;
				setupLeft = 0;
				setupRight = 1;
			}
			else
			{
				eLeft = &eTop;
				eRight = &eMaj;
				lines = eLeft->lines;
				setupLeft = 1;
				setupRight = 0;
			}
			if (lines == 0) return;
		}

		if (setupLeft && eLeft->lines > 0)
		{
			int vLower;
			fixed_t fsx = eLeft->fsx;
			fx = (fsx + FRACUNIT - 1) & 0xffff0000;
			fError = fx - fsx - FRACUNIT;
			fxLeftEdge = fsx - 1;
			fdxLeftEdge = eLeft->fdxdy;
			fdxOuter = (fdxLeftEdge - 1) & 0xffff0000;
			fdError = fdxOuter - fdxLeftEdge + FRACUNIT;
			idxOuter = fdxOuter >> FRACBITS;
			dxOuter = (float) idxOuter;

			fy = eLeft->fsy;
			iy = fy >> FRACBITS;

			adjx = (float)(fx - eLeft->fx0);  /* SCALED! */
			adjy = eLeft->adjy;		 /* SCALED! */

			vLower = eLeft->v0;

               /*
                * Now we need the set of parameter (z, color, etc.) values at
                * the point (fx, fy).  This gives us properly-sampled parameter
                * values that we can step from pixel to pixel.  Furthermore,
                * although we might have intermediate results that overflow
                * the normal parameter range when we step temporarily outside
                * the triangle, we shouldn't overflow or underflow for any
                * pixel that's actually inside the triangle.
                */

			float z0 = (vb[vLower].z <= 1.0F) ? 1.0F : 1 / vb[vLower].z;
			fz = (fixed_t)((z0 * FRACUNIT + dzdx * adjx + dzdy * adjy) * 0x8000);
			fdzOuter = (fixed_t)((dzdy + dxOuter * dzdx) * 0x8000 * 0x10000);


			float invW = 1.0F / vb[vLower].z;

			fr = (fixed_t)(((int)vb[vLower].l << FRACBITS) + drdx * adjx + drdy * adjy)
                   + (FRACUNIT >> 1);
			fdrOuter = FX(drdy + dxOuter * drdx);

//			lLeft = vb[vLower].l * invW + (dldx * adjx + dldy * adjy) * (1.0F/(float)FRACUNIT);
//			dlOuter = dldy + dxOuter * dldx;

			sLeft = vb[vLower].s * invW + (dsdx * adjx + dsdy * adjy) * (1.0F/(float)FRACUNIT);
			dsOuter = dsdy + dxOuter * dsdx;

			tLeft = vb[vLower].t * invW + (dtdx * adjx + dtdy * adjy) * (1.0F/(float)FRACUNIT);
			dtOuter = dtdy + dxOuter * dtdx;

			vLeft = invW + (dvdx * adjx + dvdy * adjy) * (1.0F/(float)FRACUNIT);
			dvOuter = dvdy + dxOuter * dvdx;
		} /*if setupLeft*/


		if (setupRight && eRight->lines > 0)
		{
			fxRightEdge = eRight->fsx - 1;
			fdxRightEdge = eRight->fdxdy;
		}

		if (lines == 0)
		{
			continue;
		}


		/* Rasterize setup */
		fdzInner = fdzOuter + fdzdx;
		fdrInner = fdrOuter + fdrdx;
//	    dlInner = dlOuter + dldx;
	    dsInner = dsOuter + dsdx;
	    dtInner = dtOuter + dtdx;
		dvInner = dvOuter + dvdx;

		while (lines > 0)
		{
			/* initialize the span interpolants to the leftmost value */
			int left = fxLeftEdge >> FRACBITS;
			int right = fxRightEdge >> FRACBITS;

			ds_count = right - left;
			if (ds_count > 0)
			{
				/* ff = fixed-pt fragment */
				ds_x = left;
				ds_y = iy;
				ds_fz = fz;
				ds_fr = fr;
//				ds_l = lLeft;
				ds_s = sLeft;
				ds_t = tLeft;
				ds_v = vLeft;
				ds_fdzdx = fdzdx;
				ds_fdrdx = fdrdx;
//				ds_dldx = dldx;
				ds_dsdx = dsdx;
				ds_dtdx = dtdx;
				ds_dvdx = dvdx;

				/* need this to accomodate round-off errors */
				fixed_t ffrend = ds_fr + (ds_count - 1) * ds_fdrdx;
				if (ffrend < 0) ds_fr -= ffrend;
				if (ds_fr < 0) ds_fr = 0;

				spanfunc();
			}

			/*
             * Advance to the next scan line.  Compute the
             * new edge coordinates, and adjust the
             * pixel-center x coordinate so that it stays
             * on or inside the major edge.
			 */
			iy++;
			lines--;

			fxLeftEdge += fdxLeftEdge;
			fxRightEdge += fdxRightEdge;


			fError += fdError;
			if (fError >= 0)
			{
				fError -= FRACUNIT;
				fz += fdzOuter;
				fr += fdrOuter;
//				lLeft += dlOuter;
				sLeft += dsOuter;
				tLeft += dtOuter;
				vLeft += dvOuter;
			}
			else
			{
				fz += fdzInner;
				fr += fdrInner;
//				lLeft += dlInner;
				sLeft += dsInner;
				tLeft += dtInner;
				vLeft += dvInner;
			}
		} /*while lines>0*/

	} /* for subTriangle */
}

