//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**	
//**	The actual span/column drawing functions.
//**
//**	All drawing to the view buffer is accomplished in this file.
//**  The other refresh files only know about ccordinates, not the
//**  architecture of the frame buffer.
//**
//**	Here find the main potential for optimization, e.g. inline assembly,
//**  different algorithms.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

#define _MakeCol16(r, g, b)	((((r) << 7) & 0x7c00) | (((g) << 2) & 0x03e0) | (((b) >> 3) & 0x001f))
#define _GetCol16R(col)		(((col) & 0x7c00) >> 7)
#define _GetCol16G(col)		(((col) & 0x03e0) >> 2)
#define _GetCol16B(col)		(((col) & 0x001f) << 3)

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

#ifndef USEASM

//==========================================================================
//
//	D_DrawFuzzSpan_15
//
//==========================================================================

extern "C" void D_DrawFuzzSpan_15(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    word *dest = scrn16 + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	while (count--)
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s = (int)(ss * invQ) & 0xff;
			int t = (int)(tt * invQ) & 0xff;
			byte color = ds_source[(t << 8) + s];
			if (color != TRANSPARENT_COLOR)
			{
				word color16 = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
				byte r1 = _GetCol16R(*dest);
				byte g1 = _GetCol16G(*dest);
				byte b1 = _GetCol16B(*dest);
				byte r2 = _GetCol16R(color16);
				byte g2 = _GetCol16G(color16);
				byte b2 = _GetCol16B(color16);
				byte r = (r1 * ds_transluc16 + r2 * (256 - ds_transluc16)) / 256;
				byte g = (g1 * ds_transluc16 + g2 * (256 - ds_transluc16)) / 256;
				byte b = (b1 * ds_transluc16 + b2 * (256 - ds_transluc16)) / 256;
				*dest = _MakeCol16(r, g, b);
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
}

//==========================================================================
//
//	D_DrawSmoothSpan_15
//
//==========================================================================

extern "C" void D_DrawSmoothSpan_15(void)
{
    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;
	float dsdx = ds_dsdx;
	float dtdx = ds_dtdx;
	float dvdx = ds_dvdx;
    word *dest = scrn16 + ylookup[ds_y] + ds_x;

	fixed_t ffr = ds_fr;
	fixed_t fdrdx = ds_fdrdx;

	do
    {
		byte color;
		word pixel;

		float invQ = (float)FRACUNIT / vv;
		fixed_t s1 = (fixed_t)(ss * invQ);
		fixed_t t1 = (fixed_t)(tt * invQ);
		int s = s1 >> FRACBITS;
		int t = t1 >> FRACBITS;

		color = ds_source[((t & 0xff) << 8) + (s & 0xff)];
		pixel = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
		int r00 = _GetCol16R(pixel);
		int g00 = _GetCol16G(pixel);
		int b00 = _GetCol16B(pixel);

		color = ds_source[((t & 0xff) << 8) + ((s + 1) & 0xff)];
		pixel = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
		int r01 = _GetCol16R(pixel);
		int g01 = _GetCol16G(pixel);
		int b01 = _GetCol16B(pixel);

		color = ds_source[(((t + 1) & 0xff) << 8) + (s & 0xff)];
		pixel = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
		int r10 = _GetCol16R(pixel);
		int g10 = _GetCol16G(pixel);
		int b10 = _GetCol16B(pixel);

		color = ds_source[(((t + 1) & 0xff) << 8) + ((s + 1) & 0xff)];
		pixel = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
		int r11 = _GetCol16R(pixel);
		int g11 = _GetCol16G(pixel);
		int b11 = _GetCol16B(pixel);

		s1 &= 0xffff;
		t1 &= 0xffff;
		fixed_t oms = FRACUNIT - s1;
		fixed_t omt = FRACUNIT - t1;

		#define m(k1, k2, c)	(int)(((__int64)k1 * (__int64)k2 * (__int64)c) >> 32)
		int r = m(oms, omt, r00) + m(s1, omt, r01) + m(oms, t1, r10) + m(s1, t1, r11);
		int g = m(oms, omt, g00) + m(s1, omt, g01) + m(oms, t1, g10) + m(s1, t1, g11);
		int b = m(oms, omt, b00) + m(s1, omt, b01) + m(oms, t1, b10) + m(s1, t1, b11);
		#undef m

		*dest = _MakeCol16(r, g, b);

		ffr += fdrdx;
		ss += dsdx;
		tt += dtdx;
		vv += dvdx;
		dest++;
	}
	while (--count);

	D_DrawZSpan();
}

//==========================================================================
//
//	D_DrawFuzzSpriteSpan_15
//
//==========================================================================

extern "C" void D_DrawFuzzSpriteSpan_15(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    word *dest = scrn16 + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	while (count--)
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ);
			int t =	(int)(tt * invQ);
			byte color = ds_source[t * ds_tex_width + s];
			if (color != TRANSPARENT_COLOR)
			{
				word color16 = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
				byte r1 = _GetCol16R(*dest);
				byte g1 = _GetCol16G(*dest);
				byte b1 = _GetCol16B(*dest);
				byte r2 = _GetCol16R(color16);
				byte g2 = _GetCol16G(color16);
				byte b2 = _GetCol16B(color16);
				byte r = (r1 * ds_transluc16 + r2 * (256 - ds_transluc16)) / 256;
				byte g = (g1 * ds_transluc16 + g2 * (256 - ds_transluc16)) / 256;
				byte b = (b1 * ds_transluc16 + b2 * (256 - ds_transluc16)) / 256;
				*dest = _MakeCol16(r, g, b);
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
}

#endif	//	USEASM

