//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**	
//**	The actual span/column drawing functions.
//**
//**	All drawing to the view buffer is accomplished in this file.
//**  The other refresh files only know about ccordinates, not the
//**  architecture of the frame buffer.
//**
//**	Here find the main potential for optimization, e.g. inline assembly,
//**  different algorithms.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

#define _MakeCol16(r, g, b)	((((r) << 8) & 0xf800) | (((g) << 3) & 0x07e0) | (((b) >> 3) & 0x001f))
#define _GetCol16R(col)		(((col) & 0xf800) >> 8)
#define _GetCol16G(col)		(((col) & 0x07e0) >> 3)
#define _GetCol16B(col)		(((col) & 0x001f) << 3)

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

#ifndef USEASM

//==========================================================================
//
//	D_DrawSpanExact_16
//
//==========================================================================

extern "C" void D_DrawSpanExact_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;
	fixed_t ffr = ds_fr;
    word *dest = scrn16 + ylookup[ds_y] + ds_x;

	do
    {
		float invQ = 1.0f / vv;
		int s = (int)(ss * invQ) & 0xff;
		int t = (int)(tt * invQ) & 0xff;
		byte color = ds_source[(t << 8) + s];
		*dest = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
	}
	while (--count);

	D_DrawZSpan();
}

//==========================================================================
//
//	D_DrawSpan8_16
//
//==========================================================================

extern "C" void D_DrawSpan8_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    word *dest = scrn16 + ylookup[ds_y] + ds_x;
    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

	float invQ = (float)FRACUNIT / vv;
	fixed_t s = (int)(ss * invQ);
	fixed_t t = (int)(tt * invQ);

	float dsdx8 = ds_dsdx * 8;
	float dtdx8 = ds_dtdx * 8;
	float dvdx8 = ds_dvdx * 8;

	fixed_t ffr = ds_fr;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			ss += dsdx8;
			tt += dtdx8;
			vv += dvdx8;

			invQ = (float)FRACUNIT / vv;
			snext = (int)(ss * invQ);
			tnext = (int)(tt * invQ);

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				ss += ds_dsdx * spancountminus1;
				tt += ds_dtdx * spancountminus1;
				vv += ds_dvdx * spancountminus1;

				invQ = (float)FRACUNIT / vv;
				snext = (int)(ss * invQ);
				tnext = (int)(tt * invQ);

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		//	Draw a part of a span. This is like in Doom's original
		// R_DrawSpan but with texture size 256x256 and changing light
		do
		{
			byte color = ds_source[((t >> 8) & 0xff00) | ((s >> 16) & 0xff)];
			*dest = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
			dest++;
			ffr += ds_fdrdx;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);

	D_DrawZSpan();
}

//==========================================================================
//
//	D_DrawSpan16_16
//
//==========================================================================

extern "C" void D_DrawSpan16_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    word *dest = scrn16 + ylookup[ds_y] + ds_x;
    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

	float invQ = (float)FRACUNIT / vv;
	fixed_t s = (int)(ss * invQ);
	fixed_t t = (int)(tt * invQ);

	float dsdx16 = ds_dsdx * 16;
	float dtdx16 = ds_dtdx * 16;
	float dvdx16 = ds_dvdx * 16;

	fixed_t ffr = ds_fr;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 16)
		{
			spancount = 16;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			ss += dsdx16;
			tt += dtdx16;
			vv += dvdx16;

			invQ = (float)FRACUNIT / vv;
			snext = (int)(ss * invQ);
			tnext = (int)(tt * invQ);

			sstep = (snext - s) >> 4;
			tstep = (tnext - t) >> 4;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				ss += ds_dsdx * spancountminus1;
				tt += ds_dtdx * spancountminus1;
				vv += ds_dvdx * spancountminus1;

				invQ = (float)FRACUNIT / vv;
				snext = (int)(ss * invQ);
				tnext = (int)(tt * invQ);

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		//	Draw a part of a span. This is like in Doom's original
		// R_DrawSpan but with texture size 256x256 and changing light
		do
		{
			byte color = ds_source[((t >> 8) & 0xff00) | ((s >> 16) & 0xff)];
			*dest = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
			dest++;
			ffr += ds_fdrdx;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);

	D_DrawZSpan();
}

//==========================================================================
//
//	D_DrawSmoothSpan_16
//
//==========================================================================

extern "C" void D_DrawSmoothSpan_16(void)
{
    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;
	fixed_t ffr = ds_fr;
    word *dest = scrn16 + ylookup[ds_y] + ds_x;

	do
    {
		byte color;
		word pixel;

		float invQ = (float)FRACUNIT / vv;
		fixed_t s1 = (fixed_t)(ss * invQ);
		fixed_t t1 = (fixed_t)(tt * invQ);
		int s = s1 >> FRACBITS;
		int t = t1 >> FRACBITS;

		color = ds_source[((t & 0xff) << 8) + (s & 0xff)];
		pixel = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
		int r00 = _GetCol16R(pixel);
		int g00 = _GetCol16G(pixel);
		int b00 = _GetCol16B(pixel);

		color = ds_source[((t & 0xff) << 8) + ((s + 1) & 0xff)];
		pixel = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
		int r01 = _GetCol16R(pixel);
		int g01 = _GetCol16G(pixel);
		int b01 = _GetCol16B(pixel);

		color = ds_source[(((t + 1) & 0xff) << 8) + (s & 0xff)];
		pixel = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
		int r10 = _GetCol16R(pixel);
		int g10 = _GetCol16G(pixel);
		int b10 = _GetCol16B(pixel);

		color = ds_source[(((t + 1) & 0xff) << 8) + ((s + 1) & 0xff)];
		pixel = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
		int r11 = _GetCol16R(pixel);
		int g11 = _GetCol16G(pixel);
		int b11 = _GetCol16B(pixel);

		s1 &= 0xffff;
		t1 &= 0xffff;
		fixed_t oms = FRACUNIT - s1;
		fixed_t omt = FRACUNIT - t1;

		#define m(k1, k2, c)	(int)(((__int64)k1 * (__int64)k2 * (__int64)c) >> 32)
		int r = m(oms, omt, r00) + m(s1, omt, r01) + m(oms, t1, r10) + m(s1, t1, r11);
		int g = m(oms, omt, g00) + m(s1, omt, g01) + m(oms, t1, g10) + m(s1, t1, g11);
		int b = m(oms, omt, b00) + m(s1, omt, b01) + m(oms, t1, b10) + m(s1, t1, b11);
		#undef m

		*dest = _MakeCol16(r, g, b);

		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
	}
	while (--count);

	D_DrawZSpan();
}

//==========================================================================
//
//	D_DrawMaskedSpan_16
//
//==========================================================================

extern "C" void D_DrawMaskedSpan_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    word *dest = scrn16 + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	do
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ) & 0xff;
			int t =	(int)(tt * invQ) & 0xff;
			byte color = ds_source[(t << 8) + s];
			if (color != TRANSPARENT_COLOR)
			{
				*dest = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
				*zdst = pixelz;
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
	while (--count);
}

//==========================================================================
//
//	D_DrawFuzzSpan_16
//
//==========================================================================

extern "C" void D_DrawFuzzSpan_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    word *dest = scrn16 + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	do
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ) & 0xff;
			int t =	(int)(tt * invQ) & 0xff;
			byte color = ds_source[(t << 8) + s];
			if (color != TRANSPARENT_COLOR)
			{
				word color16 = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
				byte r1 = _GetCol16R(*dest);
				byte g1 = _GetCol16G(*dest);
				byte b1 = _GetCol16B(*dest);
				byte r2 = _GetCol16R(color16);
				byte g2 = _GetCol16G(color16);
				byte b2 = _GetCol16B(color16);
				byte r = (r1 * ds_transluc16 + r2 * (256 - ds_transluc16)) / 256;
				byte g = (g1 * ds_transluc16 + g2 * (256 - ds_transluc16)) / 256;
				byte b = (b1 * ds_transluc16 + b2 * (256 - ds_transluc16)) / 256;
				*dest = _MakeCol16(r, g, b);
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
	while (--count);
}

//==========================================================================
//
//	D_DrawSkySpan_16
//
//==========================================================================

extern "C" void D_DrawSkySpan_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;
    word *dest = scrn16 + ylookup[ds_y] + ds_x;

	do
    {
		float invQ = 1.0f / vv;
		int s = (int)(ss * invQ) & 0x3ff;
		int t = (int)(tt * invQ) & 0xff;
		byte color = ds_source[(t << 10) + s];
		*dest = pal8_to16[color];
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
	}
	while (--count);
}

//==========================================================================
//
//	D_DrawSkySpan8_16
//
//==========================================================================

extern "C" void D_DrawSkySpan8_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    word *dest = scrn16 + ylookup[ds_y] + ds_x;
    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

	float invQ = (float)FRACUNIT / vv;
	fixed_t s = (int)(ss * invQ);
	fixed_t t = (int)(tt * invQ);

	float dsdx8 = ds_dsdx * 8;
	float dtdx8 = ds_dtdx * 8;
	float dvdx8 = ds_dvdx * 8;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			ss += dsdx8;
			tt += dtdx8;
			vv += dvdx8;

			invQ = (float)FRACUNIT / vv;
			snext = (int)(ss * invQ);
			tnext = (int)(tt * invQ);

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				ss += ds_dsdx * spancountminus1;
				tt += ds_dtdx * spancountminus1;
				vv += ds_dvdx * spancountminus1;

				invQ = (float)FRACUNIT / vv;
				snext = (int)(ss * invQ);
				tnext = (int)(tt * invQ);

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		do
		{
			byte color = ds_source[((t >> 6) & 0x3fc00) | ((s >> 16) & 0x3ff)];
			*dest = pal8_to16[color];
			dest++;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);
}

//==========================================================================
//
//	D_DrawDoubleSkySpan_16
//
//==========================================================================

extern "C" void D_DrawDoubleSkySpan_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;
    word *dest = scrn16 + ylookup[ds_y] + ds_x;

	do
    {
		float invQ = 1.0f / vv;
		int s = (int)(ss * invQ) & 0x3ff;
		int t = (int)(tt * invQ) & 0xff;
		byte color = ds_source[(t << 10) + s];
		if (color)
		{
			*dest = pal8_to16[color];
		}
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
	}
	while (--count);
}

//==========================================================================
//
//	D_DrawSpriteSpan_16
//
//==========================================================================

extern "C" void D_DrawSpriteSpan_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    word *dest = scrn16 + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	do
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ);
			int t =	(int)(tt * invQ);
			byte color = ds_source[t * ds_tex_width + s];
			if (color != TRANSPARENT_COLOR)
			{
				*dest = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
				*zdst = pixelz;
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
	while (--count);
}

//==========================================================================
//
//	D_DrawFuzzSpriteSpan_16
//
//==========================================================================

extern "C" void D_DrawFuzzSpriteSpan_16(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    word *dest = scrn16 + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	do
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ);
			int t =	(int)(tt * invQ);
			byte color = ds_source[t * ds_tex_width + s];
			if (color != TRANSPARENT_COLOR)
			{
				word color16 = ds_colormap16[color + ((ffr >> 11) & 0x1f00)];
				byte r1 = _GetCol16R(*dest);
				byte g1 = _GetCol16G(*dest);
				byte b1 = _GetCol16B(*dest);
				byte r2 = _GetCol16R(color16);
				byte g2 = _GetCol16G(color16);
				byte b2 = _GetCol16B(color16);
				byte r = (r1 * ds_transluc16 + r2 * (256 - ds_transluc16)) / 256;
				byte g = (g1 * ds_transluc16 + g2 * (256 - ds_transluc16)) / 256;
				byte b = (b1 * ds_transluc16 + b2 * (256 - ds_transluc16)) / 256;
				*dest = _MakeCol16(r, g, b);
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
	while (--count);
}

#endif	//	USEASM


