//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**	
//**	The actual span/column drawing functions.
//**
//**	All drawing to the view buffer is accomplished in this file.
//**  The other refresh files only know about ccordinates, not the
//**  architecture of the frame buffer.
//**
//**	Here find the main potential for optimization, e.g. inline assembly,
//**  different algorithms.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

#ifndef USEASM

//==========================================================================
//
//	D_DrawSpanExact_8
//
//==========================================================================

extern "C" void D_DrawSpanExact_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;
	float dsdx = ds_dsdx;
	float dtdx = ds_dtdx;
	float dvdx = ds_dvdx;
    byte *dest = scrn + ylookup[ds_y] + ds_x;

	fixed_t ffr = ds_fr;
	fixed_t fdrdx = ds_fdrdx;

	do
    {
		float invQ = 1.0f / vv;
		int s = (int)(ss * invQ) & 0xff;
		int t = (int)(tt * invQ) & 0xff;
		byte color = ds_source[(t << 8) + s];
		*dest = ds_colormap[color + ((ffr >> 11) & 0x1f00)];
		ffr += fdrdx;
		ss += dsdx;
		tt += dtdx;
		vv += dvdx;
		dest++;
	}
	while (--count);

	D_DrawZSpan();
}

//==========================================================================
//
//	D_DrawSpan8_8
//
//==========================================================================

extern "C" void D_DrawSpan8_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    byte *dest = scrn + ylookup[ds_y] + ds_x;
    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

	float invQ = (float)FRACUNIT / vv;
	fixed_t s = (int)(ss * invQ);
	fixed_t t = (int)(tt * invQ);

	float dsdx8 = ds_dsdx * 8;
	float dtdx8 = ds_dtdx * 8;
	float dvdx8 = ds_dvdx * 8;

	fixed_t ffr = ds_fr;
	fixed_t fdrdx = ds_fdrdx;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			ss += dsdx8;
			tt += dtdx8;
			vv += dvdx8;

			invQ = (float)FRACUNIT / vv;
			snext = (int)(ss * invQ);
			tnext = (int)(tt * invQ);

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				ss += ds_dsdx * spancountminus1;
				tt += ds_dtdx * spancountminus1;
				vv += ds_dvdx * spancountminus1;

				invQ = (float)FRACUNIT / vv;
				snext = (int)(ss * invQ);
				tnext = (int)(tt * invQ);

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		//	Draw a part of a span. This is like in Doom's original
		// R_DrawSpan but with texture size 256x256 and changing light
		do
		{
			byte color = ds_source[((t >> 8) & 0xff00) | ((s >> 16) & 0xff)];
			*dest = ds_colormap[color + ((ffr >> 11) & 0x1f00)];
			dest++;
			ffr += fdrdx;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);

	D_DrawZSpan();
}

//==========================================================================
//
//	D_DrawMaskedSpan_8
//
//==========================================================================

extern "C" void D_DrawMaskedSpan_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    byte *dest = scrn + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	while (count--)
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ) & 0xff;
			int t =	(int)(tt * invQ) & 0xff;
			byte color = ds_source[(t << 8) + s];
			if (color != TRANSPARENT_COLOR)
			{
				*dest = ds_colormap[color + ((ffr >> 11) & 0x1f00)];
				*zdst = pixelz;
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
}

//==========================================================================
//
//	D_DrawFuzzSpan_8
//
//==========================================================================

extern "C" void D_DrawFuzzSpan_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    byte *dest = scrn + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	while (count--)
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ) & 0xff;
			int t =	(int)(tt * invQ) & 0xff;
			byte color = ds_source[(t << 8) + s];
			if (color != TRANSPARENT_COLOR)
			{
				color = ds_colormap[color + ((ffr >> 11) & 0x1f00)];
				*dest = ds_transluc[*dest + (color << 8)];
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
}

//==========================================================================
//
//	D_DrawAltFuzzSpriteSpan_8
//
//==========================================================================

extern "C" void D_DrawAltFuzzSpan_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    byte *dest = scrn + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	while (count--)
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ) & 0xff;
			int t =	(int)(tt * invQ) & 0xff;
			byte color = ds_source[(t << 8) + s];
			if (color != TRANSPARENT_COLOR)
			{
				color = ds_colormap[color + ((ffr >> 11) & 0x1f00)];
				*dest = ds_transluc[(*dest << 8) + color];
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
}

//==========================================================================
//
//	D_DrawSkySpan_8
//
//==========================================================================

extern "C" void D_DrawSkySpan_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;
	float dsdx = ds_dsdx;
	float dtdx = ds_dtdx;
	float dvdx = ds_dvdx;
    byte *dest = scrn + ylookup[ds_y] + ds_x;

	do
    {
		float invQ = 1.0f / vv;
		int s = (int)(ss * invQ) & 0x3ff;
		int t = (int)(tt * invQ) & 0xff;
		byte color = ds_source[(t << 10) + s];
		*dest = color;
		ss += dsdx;
		tt += dtdx;
		vv += dvdx;
		dest++;
	} while (--count);
}

//==========================================================================
//
//	D_DrawSkySpan8_8
//
//==========================================================================

extern "C" void D_DrawSkySpan8_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    byte *dest = scrn + ylookup[ds_y] + ds_x;
    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

	float invQ = (float)FRACUNIT / vv;
	fixed_t s = (int)(ss * invQ);
	fixed_t t = (int)(tt * invQ);

	float dsdx8 = ds_dsdx * 8;
	float dtdx8 = ds_dtdx * 8;
	float dvdx8 = ds_dvdx * 8;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			ss += dsdx8;
			tt += dtdx8;
			vv += dvdx8;

			invQ = (float)FRACUNIT / vv;
			snext = (int)(ss * invQ);
			tnext = (int)(tt * invQ);

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				ss += ds_dsdx * spancountminus1;
				tt += ds_dtdx * spancountminus1;
				vv += ds_dvdx * spancountminus1;

				invQ = (float)FRACUNIT / vv;
				snext = (int)(ss * invQ);
				tnext = (int)(tt * invQ);

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		do
		{
			byte color = ds_source[((t >> 6) & 0x3fc00) | ((s >> 16) & 0x3ff)];
			*dest = color;
			dest++;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);
}

//==========================================================================
//
//	D_DrawDoubleSkySpan_8
//
//==========================================================================

extern "C" void D_DrawDoubleSkySpan_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;
	float dsdx = ds_dsdx;
	float dtdx = ds_dtdx;
	float dvdx = ds_dvdx;
    byte *dest = scrn + ylookup[ds_y] + ds_x;

	do
    {
		float invQ = 1.0f / vv;
		int s = (int)(ss * invQ) & 0x3ff;
		int t = (int)(tt * invQ) & 0xff;
		byte color = ds_source[(t << 10) + s];
		if (color)
		{
			*dest = color;
		}
		ss += dsdx;
		tt += dtdx;
		vv += dvdx;
		dest++;
	} while (--count);
}

//==========================================================================
//
//	D_DrawSpriteSpan_8
//
//==========================================================================

extern "C" void D_DrawSpriteSpan_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	dword ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    byte *dest = scrn + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	while (count--)
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ);
			int t =	(int)(tt * invQ);
			byte color = ds_source[t * ds_tex_width + s];
			if (color != TRANSPARENT_COLOR)
			{
				*dest = ds_colormap[color + ((ffr >> 11) & 0x1f00)];
				*zdst = pixelz;
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
}

//==========================================================================
//
//	D_DrawFuzzSpriteSpan_8
//
//==========================================================================

extern "C" void D_DrawFuzzSpriteSpan_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    byte *dest = scrn + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	while (count--)
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ);
			int t =	(int)(tt * invQ);
			byte color = ds_source[t * ds_tex_width + s];
			if (color != TRANSPARENT_COLOR)
			{
				color = ds_colormap[color + ((ffr >> 11) & 0x1f00)];
				*dest = ds_transluc[*dest + (color << 8)];
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
}

//==========================================================================
//
//	D_DrawAltFuzzSpriteSpan_8
//
//==========================================================================

extern "C" void D_DrawAltFuzzSpriteSpan_8(void)
{
#ifdef PARANOID
	if (ds_y < 0 || ds_y >= viewheight ||
		ds_x < 0 || ds_x + ds_count > viewwidth)
	{
		I_Error("Draw span at (%d,%d), count %d", ds_x, ds_y, ds_count);
	}
#endif

    int count = ds_count;
	fixed_t ffz = ds_fz;
	fixed_t ffr = ds_fr;
	float ss = ds_s;
	float tt = ds_t;
	float vv = ds_v;

    byte *dest = scrn + ylookup[ds_y] + ds_x;
    word *zdst = zbuffer + ylookup[ds_y] + ds_x;

	while (count--)
    {
		word pixelz = ffz >> 16;
		if (pixelz >= *zdst)
		{
			float invQ = 1.0f / vv;
			int s =	(int)(ss * invQ);
			int t =	(int)(tt * invQ);
			byte color = ds_source[t * ds_tex_width + s];
			if (color != TRANSPARENT_COLOR)
			{
				color = ds_colormap[color + ((ffr >> 11) & 0x1f00)];
				*dest = ds_transluc[(*dest << 8) + color];
			}
		}
		ffz += ds_fdzdx;
		ffr += ds_fdrdx;
		ss += ds_dsdx;
		tt += ds_dtdx;
		vv += ds_dvdx;
		dest++;
		zdst++;
	}
}

#endif	//	USEASM

