//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

#define MAX_TRANSLATED_SPRITES		256

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static byte**		texturedata;
static byte**		skytexturedata;	//FIXME this is a waste of memory

static byte**		flatdata;

static int*			spritewidthi;
static byte**		spritedata;

static byte*		trsprdata[MAX_TRANSLATED_SPRITES];
static int			trsprlump[MAX_TRANSLATED_SPRITES];
static int			trsprtnum[MAX_TRANSLATED_SPRITES];

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TSoftwareDrawer::InitTextures
//
//==========================================================================

void TSoftwareDrawer::InitTextures(void)
{
	//	Textures
    texturedata = (byte**)Z_Calloc(numtextures * 4);
    skytexturedata = (byte**)Z_Calloc(numtextures * 4);
	//	Flats
    flatdata = (byte**)Z_Calloc((numflats) * 4, PU_STATIC, 0);
	//	Sprite lumps
    spritewidthi = (int*)Z_Calloc(numspritelumps * 4, PU_STATIC, 0);
    spritedata = (byte**)Z_Calloc(numspritelumps * 4, PU_STATIC, 0);
}

//==========================================================================
//
//	DrawColumnInCache
//
// 	Clip and draw a column from a patch into a flat buffer.
//
//		column - column to draw
//		cache - buffer
//		originx, originy - position of column in the buffer
//		cachewidth, cacheheight - size of the cache
//
//==========================================================================

static void DrawColumnInCache(column_t* column, byte* cache,
	int originx, int originy, int cachewidth, int cacheheight)
{
    int		count;
    int		position;
    byte*	source;
    byte*	dest;
	
	// step through the posts in a column
    while (column->topdelta != 0xff)
    {
		source = (byte *)column + 3;
		count = column->length;
		position = originy + column->topdelta;

		//	Clip position
		if (position < 0)
		{
	    	count += position;
			source -= position;
	    	position = 0;
		}
		if (position + count > cacheheight)
		{
	    	count = cacheheight - position;
		}
    	dest = cache + originx + position * cachewidth;

    	while (count-- > 0)
    	{
			*dest = *source;
			source++;
			dest += cachewidth;
    	}
		
		column = (column_t *)((byte *)column + column->length + 4);
    }
}

//==========================================================================
//
//	GenerateTexture
//
// 	Using the texture definition, the composite texture is created from the
// patches, and each column is cached.
//
//==========================================================================

static void GenerateTexture(int texnum)
{
    byte*			block;
    texdef_t*		texture;
    texpatch_t*		patch;	
    patch_t*		realpatch;
    int				x;
    int				x1;
    int				x2;
    int				i;
    column_t*		patchcol;
	int				wtimes;
	int				htimes;
	int				texture_width;

    texture = textures[texnum];

	texture_width = texture->width;
	if (texture_width > 256)
	{
		cond << "JL in trouble - texture " << texture->name
			<< " haves width " << texture_width << " greater than 256\n";
		texture_width = 256;
	}

    block = (byte*)Z_Malloc(256 * 256, PU_STATIC, (void**)&texturedata[texnum]);
	memset(block, TRANSPARENT_COLOR, 256 * 256);

    // Composite the columns together.
    patch = texture->patches;

	wtimes = 256 / texture_width;
	htimes = 256 / texture->height;

    for (i = 0; i < texture->patchcount; i++, patch++)
    {
		realpatch = (patch_t*)W_CacheLumpNum(patch->patch, PU_TEMP);
		x1 = patch->originx;
		x2 = x1 + LittleShort(realpatch->width);

		if (x1 < 0)
	    	x = 0;
		else
	    	x = x1;
	
		if (x2 > texture_width)
	    	x2 = texture_width;

		for ( ; x < x2; x++)
		{
	    	patchcol = (column_t *)((byte *)realpatch
				    + LittleLong(realpatch->columnofs[x - x1]));
			for (int ht = 0; ht < htimes; ht++)
			{
				for (int wt = 0; wt < wtimes; wt++)
				{
			    	DrawColumnInCache(patchcol, block + wt * texture_width +
			    		ht * texture->height * 256, x, patch->originy,
			    		256, texture->height);
				}
			}
		}
    }

    // Now that the texture has been built in column cache,
    //  it is purgable from zone memory.
    Z_ChangeTag(block, PU_CACHE);
}

//==========================================================================
//
// 	TSoftwareDrawer::SetTexture
//
//==========================================================================

void TSoftwareDrawer::SetTexture(int tex)
{
	tex = texturetranslation[tex];

    if (!texturedata[tex])
		GenerateTexture(tex);

    ds_source = texturedata[tex];
}

//==========================================================================
//
//	GenerateSkyTexture
//
//==========================================================================

static void GenerateSkyTexture(int texnum)
{
    byte*			block;
    texdef_t*		texture;
    texpatch_t*		patch;	
    patch_t*		realpatch;
    int				x;
    int				x1;
    int				x2;
    int				i;
    column_t*		patchcol;
	int				wtimes;
	int				htimes;

    texture = textures[texnum];

    block = (byte*)Z_Malloc(1024 * 256, PU_STATIC, (void**)&skytexturedata[texnum]);
	memset(block, TRANSPARENT_COLOR, 1024 * 256);

    // Composite the columns together.
    patch = texture->patches;

	wtimes = 1024 / texture->width;
	htimes = 256 / texture->height;

    for (i = 0; i < texture->patchcount; i++, patch++)
    {
		realpatch = (patch_t*)W_CacheLumpNum(patch->patch, PU_TEMP);
		x1 = patch->originx;
		x2 = x1 + LittleShort(realpatch->width);

		if (x1 < 0)
	    	x = 0;
		else
	    	x = x1;
	
		if (x2 > texture->width)
	    	x2 = texture->width;

		for ( ; x < x2; x++)
		{
	    	patchcol = (column_t *)((byte *)realpatch
				    + LittleLong(realpatch->columnofs[x - x1]));
			for (int ht = 0; ht < htimes; ht++)
			{
				for (int wt = 0; wt < wtimes; wt++)
				{
			    	DrawColumnInCache(patchcol, block + wt * texture->width +
			    		ht * texture->height * 1024, x, patch->originy,
			    		1024, texture->height);
				}
			}
		}
    }

    // Now that the texture has been built in column cache,
    //  it is purgable from zone memory.
    Z_ChangeTag(block, PU_CACHE);
}

//==========================================================================
//
// 	TSoftwareDrawer::SetSkyTexture
//
//==========================================================================

void TSoftwareDrawer::SetSkyTexture(int tex, bool)
{
	tex = texturetranslation[tex];

    if (!skytexturedata[tex])
		GenerateSkyTexture(tex);

    ds_source = skytexturedata[tex];
}

//==========================================================================
//
//	TSoftwareDrawer::SetFlat
//
//==========================================================================

void TSoftwareDrawer::SetFlat(int num)
{
	num &= ~TEXF_FLAT;
	num = flattranslation[num];

	if (!flatdata[num])
	{
		byte *block = (byte*)Z_Malloc(256 * 256, PU_STATIC, (void**)&flatdata[num]);
		byte *data = (byte*)W_CacheLumpNum(flatlumps[num], PU_TEMP);
		for (int j = 0; j < 256; j++)
		{
			for (int i = 0; i < 4; i++)
			{
				memcpy(block + (j << 8) + (i << 6), data + ((j & 63) << 6), 64);
			}
		}
		Z_ChangeTag(block, PU_CACHE);
	}

    ds_source = flatdata[num];
}

//==========================================================================
//
//	GenerateSprite
//
//==========================================================================

static void	GenerateSprite(int lump)
{
#ifdef __GNUC__
	byte patch_buf[W_LumpLength(spritelumps[lump])];
#else
	byte *patch_buf = new byte[W_LumpLength(spritelumps[lump])];
#endif
	W_ReadLump(spritelumps[lump], patch_buf);
    patch_t	*patch = (patch_t*)patch_buf;

	spritewidth[lump] = LittleShort(patch->width);
	spriteheight[lump] = LittleShort(patch->height);
	spriteoffset[lump] = LittleShort(patch->leftoffset);
	spritetopoffset[lump] = LittleShort(patch->topoffset);
	spritewidthi[lump] = LittleShort(patch->width);

	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);

    byte *block = (byte*)Z_Malloc(w * h, PU_CACHE, (void**)&spritedata[lump]);
	memset(block, TRANSPARENT_COLOR, w * h);

	for (int x = 0; x < w; x++)
	{
    	column_t *column = (column_t *)((byte *)patch + LittleLong(patch->columnofs[x]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
		    byte* dest = block + x + column->topdelta * w;
			int count = column->length;

	    	while (count--)
	    	{
				*dest = *source;
				source++;
				dest += w;
	    	}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}
#ifndef __GNUC__
	delete patch_buf;
#endif
}

//==========================================================================
//
//	GenerateTranslatedSprite
//
//==========================================================================

static void	GenerateTranslatedSprite(int lump, int slot, int translation)
{
#ifdef __GNUC__
	byte patch_buf[W_LumpLength(spritelumps[lump])];
#else
	byte *patch_buf = new byte[W_LumpLength(spritelumps[lump])];
#endif
	W_ReadLump(spritelumps[lump], patch_buf);
    patch_t	*patch = (patch_t*)patch_buf;

	spritewidth[lump] = LittleShort(patch->width);
	spriteheight[lump] = LittleShort(patch->height);
	spriteoffset[lump] = LittleShort(patch->leftoffset);
	spritetopoffset[lump] = LittleShort(patch->topoffset);
	spritewidthi[lump] = LittleShort(patch->width);

	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);

    byte *block = (byte*)Z_Malloc(w * h, PU_CACHE, (void**)&trsprdata[slot]);
	memset(block, TRANSPARENT_COLOR, w * h);
	trsprlump[slot] = lump;
	trsprtnum[slot] = translation;

	byte *trtab = translationtables + translation * 256;

	for (int x = 0; x < w; x++)
	{
    	column_t *column = (column_t *)((byte *)patch + LittleLong(patch->columnofs[x]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
		    byte* dest = block + x + column->topdelta * w;
			int count = column->length;

	    	while (count--)
	    	{
				*dest = trtab[*source];
				source++;
				dest += w;
	    	}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}
#ifndef __GNUC__
	delete patch_buf;
#endif
}

//==========================================================================
//
//	TSoftwareDrawer::SetSpriteLump
//
//==========================================================================

void TSoftwareDrawer::SetSpriteLump(int lump, int translation)
{
	if (translation)
	{
		int i;
		int avail = -1;
		for (i = 0; i <	MAX_TRANSLATED_SPRITES; i++)
		{
			if (trsprdata[i])
			{
				if (trsprlump[i] == lump && trsprtnum[i] == translation)
				{
					ds_source = trsprdata[i];
					ds_tex_width = spritewidthi[lump];
					return;
				}
			}
			else
			{
				if (avail < 0)
					avail = i;
			}
		}
		if (avail < 0)
		{
			Z_Free(trsprdata[0]);
			avail = 0;
		}
		GenerateTranslatedSprite(lump, avail, translation);
		ds_source = trsprdata[avail];
		ds_tex_width = spritewidthi[lump];
	}
	else
	{
		if (!spritedata[lump])
		{
			GenerateSprite(lump);
		}
		ds_source = spritedata[lump];
		ds_tex_width = spritewidthi[lump];
	}
}

