//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

#include "asm_i386.h"

#ifdef USEASM

.data
	Align8
mmbuf:
	.long	0
	.long	0

.text

//==========================================================================
//
//	D_DrawZSpan
//
//	EAX - temp data
//	EBX - current z
//	ECX - counter
//	EDX - temp data
//	ESI - z step
//	EDI - dest pointer
//	EBP - saved count
//
//==========================================================================

	Align16
.globl C(D_DrawZSpan)
C(D_DrawZSpan):

	//	Save registers
	pushl	%ebp
	pushl	%edi
	pushl	%esi
	pushl	%ebx

	//	Setup our registers
	movl	C(ds_y),%eax
	movl	C(ylookup)(,%eax,4),%edi
	addl	C(ds_x),%edi
	addl	%edi,%edi
	addl	C(zbuffer),%edi
	movl	C(ds_fz),%ebx
	movl	C(ds_fdzdx),%esi
	movl	C(ds_count),%ecx

	//	Do a single pixel up front, if necessary to dword align the
	// destination
	testl	$2,%edi
	jz		LMiddle
	movl	%ebx,%eax
	addl	%esi,%ebx
	shrl	$16,%eax
	decl	%ecx
	movw	%ax,(%edi)
	addl	$2,%edi

LMiddle:
	movl	%ecx,%ebp

	testl	%ebp,C(mmx_mask16)	// don't bother with MMX with less than 16 pixels
	jnz		LDrawZSpan_mmx		// use MMX version

	// do middle a pair of aligned dwords at a time

	shrl	$1,%ecx				// count / 2
	jz		LLast				// no aligned dwords to do

	shrl	$1,%ecx				// (count / 2) / 2
	jnc		LMiddleLoop		// even number of aligned dwords to do

	movl	%ebx,%edx
	addl	%esi,%ebx
	movl	%ebx,%eax
	addl	%esi,%ebx
	shrl	$16,%edx
	andl	$0xffff0000,%eax
	orl		%edx,%eax
	movl	%eax,(%edi)
	addl	$4,%edi

	andl	%ecx,%ecx
	jz		LLast

	Align16					//	Force alignement
LMiddleLoop:
	movl	%ebx,%edx
	addl	%esi,%ebx
	movl	%ebx,%eax
	addl	%esi,%ebx
	shrl	$16,%edx
	andl	$0xffff0000,%eax
	orl		%edx,%eax
	movl	%eax,(%edi)

	movl	%ebx,%edx
	addl	%esi,%ebx
	movl	%ebx,%eax
	addl	%esi,%ebx
	shrl	$16,%edx
	andl	$0xffff0000,%eax
	orl		%edx,%eax
	movl	%eax,4(%edi)
	addl	$8,%edi

	decl	%ecx
	jnz		LMiddleLoop

LLast:
	//	Do last unaligned pixel, if there is one
	andl	$1,%ebp
	jz		LExit
	shrl	$16,%ebx
	movw	%bx,(%edi)

LExit:
	//	Restore registers
	popl	%ebx
	popl	%esi
	popl	%edi
	popl	%ebp
	ret

	//
	//	MMX version
	//

	Align16
LDrawZSpan_mmx:

	shrl	$1,%ecx				// count / 2

	movl	%ebx,mmbuf
	addl	%esi,%ebx
	movl	%ebx,mmbuf+4
	movq	mmbuf,%mm0

	addl	%esi,%esi
	movl	%esi,mmbuf
	movl	%esi,mmbuf+4
	movq	mmbuf,%mm1

	shrl	$1,%ecx				// (count / 2) / 2
	jnc		LMMXDoQwords		// even number of aligned dwords to do

	movq	%mm0,%mm2
	paddd	%mm1,%mm0
	psrld	$16,%mm2
	packssdw	%mm0,%mm2
	movd	%mm2,(%edi)
	addl	$4,%edi

LMMXDoQwords:
	shrl	$1,%ecx				// ((count / 2) / 2) / 2
	jnc		LMMXMiddleLoop		// even number of aligned qwords to do

	movq	%mm0,%mm2
	paddd	%mm1,%mm0
	movq	%mm0,%mm3
	paddd	%mm1,%mm0
	psrld	$16,%mm2
	psrld	$16,%mm3
	packssdw	%mm3,%mm2
	movq	%mm2,(%edi)
	addl	$8,%edi

	Align16					//	Force alignement
LMMXMiddleLoop:
	movq	%mm0,%mm2
	paddd	%mm1,%mm0
	movq	%mm0,%mm3
	paddd	%mm1,%mm0
	psrld	$16,%mm2
	psrld	$16,%mm3
	packssdw	%mm3,%mm2
	movq	%mm2,(%edi)
	movq	%mm0,%mm2
	paddd	%mm1,%mm0
	movq	%mm0,%mm3
	paddd	%mm1,%mm0
	psrld	$16,%mm2
	psrld	$16,%mm3
	packssdw	%mm3,%mm2
	movq	%mm2,8(%edi)
	addl	$16,%edi
	decl	%ecx
	jz		LMMXLast

	movq	%mm0,%mm2
	paddd	%mm1,%mm0
	movq	%mm0,%mm3
	paddd	%mm1,%mm0
	psrld	$16,%mm2
	psrld	$16,%mm3
	packssdw	%mm3,%mm2
	movq	%mm2,(%edi)
	movq	%mm0,%mm2
	paddd	%mm1,%mm0
	movq	%mm0,%mm3
	paddd	%mm1,%mm0
	psrld	$16,%mm2
	psrld	$16,%mm3
	packssdw	%mm3,%mm2
	movq	%mm2,8(%edi)
	addl	$16,%edi
	decl	%ecx
	jnz		LMMXMiddleLoop

LMMXLast:
	//	Do last unaligned pixel, if there is one
	andl	$1,%ebp
	jz		LMMXExit
	movd	%mm0,%eax
	shrl	$16,%eax
	movw	%ax,(%edi)

LMMXExit:
	emms
	//	Restore registers
	popl	%ebx
	popl	%esi
	popl	%edi
	popl	%ebp
	ret

#endif
