//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

struct rgb_t
{
	byte	r;
	byte	g;
	byte	b;
};

struct rgba_t
{
	byte	r;
	byte	g;
	byte	b;
	byte	a;
};

class draw_vertex_t : public TVec
{
 public:
	draw_vertex_t(void) { }
	draw_vertex_t(float ax, float ay, float az, float as, float at, float al)
		: TVec(ax, ay, az)
	{
		s = as; t = at; l = al;
	}

	float		s;	//	Texture coordinates
	float		t;
	float		l;	//	Light

	draw_vertex_t &operator = (const TVec &v)
	{
		x = v.x;
		y = v.y;
		z = v.z;
		return *this;
	}
};

class TDrawer
{
 public:
	virtual void Init(void) = 0;
	virtual void InitData(void) = 0;
	virtual bool SetResolution(int, int, int) = 0;
	virtual void InitResolution(void) = 0;
	virtual void SetPalette(int) = 0;
	virtual void StartUpdate(void) = 0;
	virtual void Update(void) = 0;
	virtual void Shutdown(void) = 0;
	virtual void* ReadScreen(int*, bool*) = 0;

	//	Screen wipes
	virtual bool InitWipe(void) = 0;
	virtual void DoWipe(int) = 0;

	//	Rendring stuff
	virtual void SetupView(int, int, int, int, float, float) = 0;
	virtual void SetupFrame(void) = 0;
	virtual void WorldDrawing(void) = 0;
	virtual void EndView(void) = 0;

	//	Texture stuff
	virtual void InitTextures(void) = 0;
	virtual void SetTexture(int) = 0;
	virtual void SetSkyTexture(int, bool) = 0;
	virtual void SetFlat(int) = 0;
	virtual void SetSpriteLump(int, int) = 0;

	//	Polygon drawing
	virtual void DrawPolygon(draw_vertex_t *, int, int, int, const TVec &, bool) = 0;
	virtual void DrawMaskedPolygon(draw_vertex_t *, int, int, int, int, const TVec &, bool) = 0;
	virtual void DrawZPolygon(draw_vertex_t *, int, int) = 0;
	virtual void DrawSkyPolygon(draw_vertex_t *, int, int, float, float, bool) = 0;
	virtual void DrawSpritePolygon(draw_vertex_t *, int, const TVec &, int, int, bool) = 0;

	//	Drawing
	virtual void DrawPic(float, float, float, float, float, float, float, float, int, int) = 0;
	virtual void DrawPicShadow(float, float, float, float, float, float, float, float, int, int) = 0;
	virtual void FillRectWithFlat(int, int, int, int, const char*) = 0;
	virtual void ShadeRect(int, int, int, int, int) = 0;
	virtual void DrawConsoleBackground(int) = 0;
	virtual void DrawSpriteLump(int, int, int, int, boolean) = 0;
};

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PUBLIC DATA DECLARATIONS ------------------------------------------------

extern TDrawer			*Drawer;
extern TDrawer			*_SoftwareDrawer;
extern TDrawer			*_OpenGLDrawer;
extern TDrawer			*_Direct3DDrawer;

