//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**	
//**	PROGS execution, asm version.
//**	
//**************************************************************************

//#define PROGS_PROFILE

#include "asm_i386.h"

#ifdef USEASM

//==========================================================================
//
//  RunFunction
//
//	edi		- Current statement pointer
//	esi		- Stack index
//
//==========================================================================

#define OFFS_FUNC_NUM		8
#define OFFS_LOCALS			-256

.text

	Align16
.globl C(RunFunction)
C(RunFunction):
	//	Get local vars
	pushl %ebp
	movl %esp,%ebp
	subl $256,%esp
	//	Save registers
	pushl %edi
	pushl %esi
	pushl %ebx

	//	Set current function num
	movl OFFS_FUNC_NUM(%ebp),%edi
	movl %edi,C(current_func)
	sall $4,%edi
	addl C(pr_functions),%edi
#ifdef PROGS_PROFILE
	call	C(PR_Profile1)
#endif

    //	Check for builtin
	movl 4(%edi),%eax
	testl %eax,%eax
	jge LINTERPRET_FUNCTION

    //	Builtin function call
	sall $2,%eax
	movl C(pr_builtins),%edx
	subl %eax,%edx
	movl (%edx),%eax
	call *%eax
	jmp LEND_RUN_FUNCTION

	//---------------------------------
    //
    //	Interpreter
    //
	//---------------------------------
	Align16
LINTERPRET_FUNCTION:
	//	Set registers
	movl C(pr_stackPtr),%esi
	movl 8(%edi),%eax
	movl 4(%edi),%edi
	//	Check for args
	testl %eax,%eax
	je LSET_STATEMENT_POINTER

    //	Pop args
	movl %eax,%edx
	Align16max7
LPOP_ARGS_LOOP:
	decl %edx
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	movl %eax,OFFS_LOCALS(%ebp,%edx,4)
	testl %edx,%edx
	jne LPOP_ARGS_LOOP

    //	Main function loop
LSET_STATEMENT_POINTER:
	sall $2,%edi
	addl C(pr_statements),%edi
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	//
	//	Opcode table
    //
	Align16
LOPCODE_TABLE:
	.long LOPC_DONE
	.long LOPC_RETURN
	.long LOPC_PUSHNUMBER
	.long LOPC_PUSHPOINTED
	.long LOPC_LOCALADDRESS
	.long LOPC_GLOBALADDRESS
	.long LOPC_ADD
	.long LOPC_SUBTRACT
	.long LOPC_MULTIPLY
	.long LOPC_DIVIDE

	.long LOPC_MODULUS
	.long LOPC_UDIVIDE
	.long LOPC_UMODULUS
	.long LOPC_EQ
	.long LOPC_NE
	.long LOPC_LT
	.long LOPC_GT
	.long LOPC_LE
	.long LOPC_GE
	.long LOPC_ULT

	.long LOPC_UGT
	.long LOPC_ULE
	.long LOPC_UGE
	.long LOPC_ANDLOGICAL
	.long LOPC_ORLOGICAL
	.long LOPC_NEGATELOGICAL
	.long LOPC_ANDBITWISE
	.long LOPC_ORBITWISE
	.long LOPC_XORBITWISE
	.long LOPC_LSHIFT

	.long LOPC_RSHIFT
	.long LOPC_URSHIFT
	.long LOPC_UNARYMINUS
	.long LOPC_BITINVERSE
	.long LOPC_CALL
	.long LOPC_GOTO
	.long LOPC_IFGOTO
	.long LOPC_IFNOTGOTO
	.long LOPC_CASEGOTO
	.long LOPC_DROP

	.long LOPC_ASSIGN
	.long LOPC_ADDVAR
	.long LOPC_SUBVAR
	.long LOPC_MULVAR
	.long LOPC_DIVVAR
	.long LOPC_MODVAR
	.long LOPC_UDIVVAR
	.long LOPC_UMODVAR
	.long LOPC_ANDVAR
	.long LOPC_ORVAR

	.long LOPC_XORVAR
	.long LOPC_LSHIFTVAR
	.long LOPC_RSHIFTVAR
	.long LOPC_URSHIFTVAR
	.long LOPC_PREINC
	.long LOPC_PREDEC
	.long LOPC_POSTINC
	.long LOPC_POSTDEC
	.long LOPC_IFTOPGOTO
	.long LOPC_IFNOTTOPGOTO

	.long LOPC_ASSIGN_DROP
	.long LOPC_ADDVAR_DROP
	.long LOPC_SUBVAR_DROP
	.long LOPC_MULVAR_DROP
	.long LOPC_DIVVAR_DROP
	.long LOPC_MODVAR_DROP
	.long LOPC_UDIVVAR_DROP
	.long LOPC_UMODVAR_DROP
	.long LOPC_ANDVAR_DROP
	.long LOPC_ORVAR_DROP

	.long LOPC_XORVAR_DROP
	.long LOPC_LSHIFTVAR_DROP
	.long LOPC_RSHIFTVAR_DROP
	.long LOPC_URSHIFTVAR_DROP
	.long LOPC_INC_DROP
	.long LOPC_DEC_DROP
	.long LOPC_FADD
	.long LOPC_FSUBTRACT
	.long LOPC_FMULTIPLY
	.long LOPC_FDIVIDE

	.long LOPC_FEQ
	.long LOPC_FNE
	.long LOPC_FLT
	.long LOPC_FGT
	.long LOPC_FLE
	.long LOPC_FGE
	.long LOPC_FUNARYMINUS
	.long LOPC_FADDVAR
	.long LOPC_FSUBVAR
	.long LOPC_FMULVAR

	.long LOPC_FDIVVAR
	.long LOPC_FADDVAR_DROP
	.long LOPC_FSUBVAR_DROP
	.long LOPC_FMULVAR_DROP
	.long LOPC_FDIVVAR_DROP
	.long LOPC_SWAP
	.long LOPC_ICALL
	.long LOPC_VPUSHPOINTED
	.long LOPC_VADD
	.long LOPC_VSUBTRACT

	.long LOPC_VPRESCALE
	.long LOPC_VPOSTSCALE
	.long LOPC_VISCALE
	.long LOPC_VEQ
	.long LOPC_VNE
	.long LOPC_VUNARYMINUS
	.long LOPC_VDROP
	.long LOPC_VASSIGN
	.long LOPC_VADDVAR
	.long LOPC_VSUBVAR

	.long LOPC_VSCALEVAR
	.long LOPC_VISCALEVAR
	.long LOPC_VASSIGN_DROP
	.long LOPC_VADDVAR_DROP
	.long LOPC_VSUBVAR_DROP
	.long LOPC_VSCALEVAR_DROP
	.long LOPC_VISCALEVAR_DROP

	Align16
LINC_STATEMENT_POINTER:
	addl $4,%edi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Empty function or invalid opcode
	Align16
LOPC_DONE:
	call	C(PR_RFInvalidOpcode)

	//	Push number
	Align16
LOPC_PUSHNUMBER:
	movl (%edi),%eax
	movl %eax,C(pr_stack)(,%esi,4)
	addl $4,%edi
	incl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Push pointed
	Align16
LOPC_PUSHPOINTED:
	movl C(pr_stack)-4(,%esi,4),%eax
	movl (%eax),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Push address of a local variable
	Align16
LOPC_LOCALADDRESS:
	movl (%edi),%eax
	sall $2,%eax
	leal OFFS_LOCALS(%ebp,%eax),%eax
	movl %eax,C(pr_stack)(,%esi,4)
	addl $4,%edi
	incl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Push address of a global variable
	Align16
LOPC_GLOBALADDRESS:
	movl (%edi),%eax
	movl C(pr_globaldefs),%edx
	movzwl 2(%edx,%eax,8),%eax
	sall $2,%eax
	addl C(pr_globals),%eax
	movl %eax,C(pr_stack)(,%esi,4)
	addl $4,%edi
	incl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	//	Add
	Align16
LOPC_ADD:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	addl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Subtract
	Align16
LOPC_SUBTRACT:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	subl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Multiply
	Align16
LOPC_MULTIPLY:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	imull C(pr_stack)(,%esi,4),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	//	Divide
	Align16
LOPC_DIVIDE:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	cltd
	idivl C(pr_stack)(,%esi,4)
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Modulus
	Align16
LOPC_MODULUS:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	cltd
	idivl C(pr_stack)(,%esi,4)
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	//	Unsigned divide
	Align16
LOPC_UDIVIDE:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	xorl %edx,%edx
	divl C(pr_stack)(,%esi,4)
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned modulus
	Align16
LOPC_UMODULUS:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	xorl %edx,%edx
	divl C(pr_stack)(,%esi,4)
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Equals
	Align16
LOPC_EQ:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	sete %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Not equals
	Align16
LOPC_NE:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	setne %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Less
	Align16
LOPC_LT:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	setl %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Greater
	Align16
LOPC_GT:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	setg %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Less or equals
	Align16
LOPC_LE:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	setle %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Greater or equals
	Align16
LOPC_GE:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	setge %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned less
	Align16
LOPC_ULT:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	setb %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned greater
	Align16
LOPC_UGT:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	seta %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned less or equals
	Align16
LOPC_ULE:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	setbe %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	//	Unsigned greater or equals
	Align16
LOPC_UGE:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	cmpl %eax,C(pr_stack)-4(,%esi,4)
	setae %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Logical AND
	Align16
LOPC_ANDLOGICAL:
	decl %esi
	cmpl $0,C(pr_stack)-4(,%esi,4)
	je LAND_FALSE
	cmpl $0,C(pr_stack)(,%esi,4)
	je LAND_FALSE
	movl $1,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)
LAND_FALSE:
	movl $0,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Logical OR
	Align16
LOPC_ORLOGICAL:
	decl %esi
	cmpl $0,C(pr_stack)-4(,%esi,4)
	jne LOR_TRUE
	cmpl $0,C(pr_stack)(,%esi,4)
	jne LOR_TRUE
	movl $0,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)
LOR_TRUE:
	movl $1,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Logical negate
	Align16
LOPC_NEGATELOGICAL:
	cmpl $0,C(pr_stack)-4(,%esi,4)
	sete %al
	andl $1,%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bitwise AND
	Align16
LOPC_ANDBITWISE:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	andl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bitwise OR
	Align16
LOPC_ORBITWISE:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	orl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bitwise XOR
	Align16
LOPC_XORBITWISE:
	decl %esi
	movl C(pr_stack)(,%esi,4),%eax
	xorl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Left shift
	Align16
LOPC_LSHIFT:
	decl %esi
	movl C(pr_stack)(,%esi,4),%ecx
	sall %cl,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Right shift
	Align16
LOPC_RSHIFT:
	decl %esi
	movl C(pr_stack)(,%esi,4),%ecx
	sarl %cl,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned right shift
	Align16
LOPC_URSHIFT:
	decl %esi
	movl C(pr_stack)(,%esi,4),%ecx
	shrl %cl,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unary minus
	Align16
LOPC_UNARYMINUS:
	negl C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bit inverse
	Align16
LOPC_BITINVERSE:
	notl C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Call another function
	Align16
LOPC_CALL:
	movl %esi,C(pr_stackPtr)
	addl $-12,%esp
	pushl (%edi)
	addl $4,%edi
	call C(RunFunction)
	addl $16,%esp
	movl OFFS_FUNC_NUM(%ebp),%eax
	movl %eax,C(current_func)
	movl C(pr_stackPtr),%esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Goto
	Align16
LOPC_GOTO:
	movl (%edi),%edi
	//	Set and go to the next statement
	sall $2,%edi
	addl C(pr_statements),%edi
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	If goto
	Align16
LOPC_IFGOTO:
	decl %esi
	cmpl $0,C(pr_stack)(,%esi,4)
	je LINC_STATEMENT_POINTER
	movl (%edi),%edi
	//	Set and go to the next statement
	sall $2,%edi
	addl C(pr_statements),%edi
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	If not goto
	Align16
LOPC_IFNOTGOTO:
	decl %esi
	cmpl $0,C(pr_stack)(,%esi,4)
	jne LINC_STATEMENT_POINTER
	movl (%edi),%edi
	//	Set and go to the next statement
	sall $2,%edi
	addl C(pr_statements),%edi
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Case goto
	Align16
LOPC_CASEGOTO:
	movl (%edi),%eax
	addl $4,%edi
	cmpl C(pr_stack)-4(,%esi,4),%eax
	jne LINC_STATEMENT_POINTER
	movl (%edi),%edi
	decl %esi
	//	Set and go to the next statement
	sall $2,%edi
	addl C(pr_statements),%edi
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Drop from stack
	Align16
LOPC_DROP:
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Assign
	Align16
LOPC_ASSIGN:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	movl %eax,(%edx)
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Add to variable
	Align16
LOPC_ADDVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	addl %eax,(%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Subtract from variable
	Align16
LOPC_SUBVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	subl %eax,(%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Multiply variable
	Align16
LOPC_MULVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	imull (%edx),%eax
	movl %eax,(%edx)
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Divide variable
	Align16
LOPC_DIVVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%ecx
	movl (%ecx),%eax
	cltd
	idivl C(pr_stack)(,%esi,4)
	movl %eax,(%ecx)
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Modulus variable
	Align16
LOPC_MODVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%ecx
	movl (%ecx),%eax
	cltd
	idivl C(pr_stack)(,%esi,4)
	movl %edx,(%ecx)
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned divide variable
	Align16
LOPC_UDIVVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%ecx
	movl (%ecx),%eax
	xorl %edx,%edx
	divl C(pr_stack)(,%esi,4)
	movl %eax,(%ecx)
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned modulus variable
	Align16
LOPC_UMODVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%ecx
	movl (%ecx),%eax
	xorl %edx,%edx
	divl C(pr_stack)(,%esi,4)
	movl %edx,(%ecx)
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bitwise AND variable
	Align16
LOPC_ANDVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	andl %eax,(%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bitwise OR variable
	Align16
LOPC_ORVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	orl %eax,(%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bitwise XOR variable
	Align16
LOPC_XORVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	xorl %eax,(%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Left shift variable
	Align16
LOPC_LSHIFTVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%ecx
	sall %cl,(%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Right shift variable
	Align16
LOPC_RSHIFTVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%ecx
	sarl %cl,(%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned right shift variable
	Align16
LOPC_URSHIFTVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%ecx
	shrl %cl,(%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Preincrement
	Align16
LOPC_PREINC:
	movl C(pr_stack)-4(,%esi,4),%edx
	incl (%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Predecrement
	Align16
LOPC_PREDEC:
	movl C(pr_stack)-4(,%esi,4),%edx
	decl (%edx)
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Postincrement
	Align16
LOPC_POSTINC:
	movl C(pr_stack)-4(,%esi,4),%edx
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	incl (%edx)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Postdecrement
	Align16
LOPC_POSTDEC:
	movl C(pr_stack)-4(,%esi,4),%edx
	movl (%edx),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	decl (%edx)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	If top goto
	Align16
LOPC_IFTOPGOTO:
	cmpl $0,C(pr_stack)-4(,%esi,4)
	je LINC_STATEMENT_POINTER
	movl (%edi),%edi
	//	Set and go to the next statement
	sall $2,%edi
	addl C(pr_statements),%edi
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	If not top goto
	Align16
LOPC_IFNOTTOPGOTO:
	cmpl $0,C(pr_stack)-4(,%esi,4)
	jne LINC_STATEMENT_POINTER
	movl (%edi),%edi
	//	Set and go to the next statement
	sall $2,%edi
	addl C(pr_statements),%edi
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Assign, drop result
	Align16
LOPC_ASSIGN_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	movl %eax,(%edx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Add to variable, drop result
	Align16
LOPC_ADDVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	addl %eax,(%edx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Subtract from variable, drop result
	Align16
LOPC_SUBVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl C(pr_stack)(,%esi,4),%eax
	subl %eax,(%edx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Multiply variable, drop result
	Align16
LOPC_MULVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%edx
	movl (%edx),%eax
	imull C(pr_stack)(,%esi,4),%eax
	movl %eax,(%edx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Divide variable, drop result
	Align16
LOPC_DIVVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%ecx
	movl (%ecx),%eax
	cltd
	idivl C(pr_stack)(,%esi,4)
	movl %eax,(%ecx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Modulus variable, drop result
	Align16
LOPC_MODVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%ecx
	movl (%ecx),%eax
	cltd
	idivl C(pr_stack)(,%esi,4)
	movl %edx,(%ecx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned divide variable, drop result
	Align16
LOPC_UDIVVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%ecx
	movl (%ecx),%eax
	xorl %edx,%edx
	divl C(pr_stack)(,%esi,4)
	movl %eax,(%ecx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned modulus variable, drop result
	Align16
LOPC_UMODVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%ecx
	movl (%ecx),%eax
	xorl %edx,%edx
	divl C(pr_stack)(,%esi,4)
	movl %edx,(%ecx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bitwise AND variable, drop result
	Align16
LOPC_ANDVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	movl C(pr_stack)(,%esi,4),%ecx
	andl %ecx,(%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bitwise OR variable, drop result
	Align16
LOPC_ORVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	movl C(pr_stack)(,%esi,4),%ecx
	orl %ecx,(%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Bitwise XOR variable, drop result
	Align16
LOPC_XORVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	movl C(pr_stack)(,%esi,4),%ecx
	xorl %ecx,(%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Left shift variable, drop result
	Align16
LOPC_LSHIFTVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	movl C(pr_stack)(,%esi,4),%ecx
	sall %cl,(%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Right shift variable, drop result
	Align16
LOPC_RSHIFTVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	movl C(pr_stack)(,%esi,4),%ecx
	sarl %cl,(%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Unsigned right shift variable, drop result
	Align16
LOPC_URSHIFTVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	movl C(pr_stack)(,%esi,4),%ecx
	shrl %cl,(%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Increment, drop result
	Align16
LOPC_INC_DROP:
	movl C(pr_stack)-4(,%esi,4),%edx
	incl (%edx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

    //	Decrement, drop result
	Align16
LOPC_DEC_DROP:
	movl C(pr_stack)-4(,%esi,4),%edx
	decl (%edx)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

//**************************************************************************
	//	Add float
	Align16
LOPC_FADD:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fadds C(pr_stack)(,%esi,4)
	fstps C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	//	Subtract float
	Align16
LOPC_FSUBTRACT:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fsubs C(pr_stack)(,%esi,4)
	fstps C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	//	Multiply float
	Align16
LOPC_FMULTIPLY:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fmuls C(pr_stack)(,%esi,4)
	fstps C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	//	Divide float
	Align16
LOPC_FDIVIDE:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fdivs C(pr_stack)(,%esi,4)
	fstps C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FEQ:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fcomps C(pr_stack)(,%esi,4)
	fnstsw %ax
	sahf
	sete %dl
	andl $1,%edx
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FNE:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fcomps C(pr_stack)(,%esi,4)
	fnstsw %ax
	sahf
	setne %dl
	andl $1,%edx
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FLT:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fcomps C(pr_stack)(,%esi,4)
	fnstsw %ax
	sahf
	setb %dl
	andl $1,%edx
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FGT:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fcomps C(pr_stack)(,%esi,4)
	fnstsw %ax
	sahf
	seta %dl
	andl $1,%edx
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FLE:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fcomps C(pr_stack)(,%esi,4)
	fnstsw %ax
	sahf
	setbe %dl
	andl $1,%edx
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FGE:
	decl %esi
	flds C(pr_stack)-4(,%esi,4)
	fcomps C(pr_stack)(,%esi,4)
	fnstsw %ax
	sahf
	setnb %dl
	andl $1,%edx
	movl %edx,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FUNARYMINUS:
	flds C(pr_stack)-4(,%esi,4)
	fchs
	fstps C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FADDVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	flds (%eax)
	fadds C(pr_stack)(,%esi,4)
	fstps (%eax)
	movl (%eax),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FSUBVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	flds (%eax)
	fsubs C(pr_stack)(,%esi,4)
	fstps (%eax)
	movl (%eax),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FMULVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	flds (%eax)
	fmuls C(pr_stack)(,%esi,4)
	fstps (%eax)
	movl (%eax),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FDIVVAR:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	flds (%eax)
	fdivs C(pr_stack)(,%esi,4)
	fstps (%eax)
	movl (%eax),%eax
	movl %eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FADDVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	flds (%eax)
	fadds C(pr_stack)(,%esi,4)
	fstps (%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FSUBVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	flds (%eax)
	fsubs C(pr_stack)(,%esi,4)
	fstps (%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FMULVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	flds (%eax)
	fmuls C(pr_stack)(,%esi,4)
	fstps (%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

	Align16
LOPC_FDIVVAR_DROP:
	decl %esi
	movl C(pr_stack)-4(,%esi,4),%eax
	flds (%eax)
	fdivs C(pr_stack)(,%esi,4)
	fstps (%eax)
	decl %esi
	//	Go to the next statement
	movl (%edi),%eax
	addl $4,%edi
	jmp *LOPCODE_TABLE(,%eax,4)

//**************************************************************************

	//	Swap elements on top of the stack
	Align16
LOPC_SWAP:
	movl	C(pr_stack)-8(,%esi,4),%eax
	movl	C(pr_stack)-4(,%esi,4),%edx
	movl	%edx,C(pr_stack)-8(,%esi,4)
	movl	%eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

    //	Indirect call another function
	Align16
LOPC_ICALL:
	decl	%esi
	addl	$-12,%esp
	pushl	C(pr_stack)(,%esi,4)
	movl	%esi,C(pr_stackPtr)
	call	C(RunFunction)
	addl	$16,%esp
	movl	OFFS_FUNC_NUM(%ebp),%eax
	movl	%eax,C(current_func)
	movl	C(pr_stackPtr),%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

//**************************************************************************

	//	Push vector
	Align16
LOPC_VPUSHPOINTED:
	addl	$2,%esi
	movl	C(pr_stack)-12(,%esi,4),%edx
	movl	(%edx),%eax
	movl	%eax,C(pr_stack)-12(,%esi,4)
	movl	4(%edx),%eax
	movl	%eax,C(pr_stack)-8(,%esi,4)
	movl	8(%edx),%eax
	movl	%eax,C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Add vectors
	Align16
LOPC_VADD:
	flds	C(pr_stack)-24(,%esi,4)
	fadds	C(pr_stack)-12(,%esi,4)
	fstps	C(pr_stack)-24(,%esi,4)
	flds	C(pr_stack)-20(,%esi,4)
	fadds	C(pr_stack)-8(,%esi,4)
	fstps	C(pr_stack)-20(,%esi,4)
	flds	C(pr_stack)-16(,%esi,4)
	fadds	C(pr_stack)-4(,%esi,4)
	fstps	C(pr_stack)-16(,%esi,4)
	addl	$-3,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Subtract vectors
	Align16
LOPC_VSUBTRACT:
	flds	C(pr_stack)-24(,%esi,4)
	fsubs	C(pr_stack)-12(,%esi,4)
	fstps	C(pr_stack)-24(,%esi,4)
	flds	C(pr_stack)-20(,%esi,4)
	fsubs	C(pr_stack)-8(,%esi,4)
	fstps	C(pr_stack)-20(,%esi,4)
	flds	C(pr_stack)-16(,%esi,4)
	fsubs	C(pr_stack)-4(,%esi,4)
	fstps	C(pr_stack)-16(,%esi,4)
	addl	$-3,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Scale vector float*vector
	Align16
LOPC_VPRESCALE:
	flds	C(pr_stack)-16(,%esi,4)
	fld		%st(0)
	fld		%st(1)
	fmuls	C(pr_stack)-12(,%esi,4)
	fstps	C(pr_stack)-16(,%esi,4)
	fmuls	C(pr_stack)-8(,%esi,4)
	fstps	C(pr_stack)-12(,%esi,4)
	fmuls	C(pr_stack)-4(,%esi,4)
	fstps	C(pr_stack)-8(,%esi,4)
	decl	%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Scale vector vector*float
	Align16
LOPC_VPOSTSCALE:
	flds	C(pr_stack)-4(,%esi,4)
	fld		%st(0)
	fld		%st(1)
	fmuls	C(pr_stack)-16(,%esi,4)
	fstps	C(pr_stack)-16(,%esi,4)
	fmuls	C(pr_stack)-12(,%esi,4)
	fstps	C(pr_stack)-12(,%esi,4)
	fmuls	C(pr_stack)-8(,%esi,4)
	fstps	C(pr_stack)-8(,%esi,4)
	decl	%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Scale vector vector/float
	Align16
LOPC_VISCALE:
	flds	C(pr_stack)-16(,%esi,4)
	fdivs	C(pr_stack)-4(,%esi,4)
	fstps	C(pr_stack)-16(,%esi,4)
	flds	C(pr_stack)-12(,%esi,4)
	fdivs	C(pr_stack)-4(,%esi,4)
	fstps	C(pr_stack)-12(,%esi,4)
	flds	C(pr_stack)-8(,%esi,4)
	fdivs	C(pr_stack)-4(,%esi,4)
	fstps	C(pr_stack)-8(,%esi,4)
	decl	%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Equal vectors
	Align16
LOPC_VEQ:
	flds	C(pr_stack)-24(,%esi,4)
	flds	C(pr_stack)-12(,%esi,4)
	fucompp
	fnstsw	%ax
	sahf
	jne		VEQ_FALSE
	flds	C(pr_stack)-20(,%esi,4)
	flds	C(pr_stack)-8(,%esi,4)
	fucompp
	fnstsw	%ax
	sahf
	jne		VEQ_FALSE
	flds	C(pr_stack)-16(,%esi,4)
	flds	C(pr_stack)-4(,%esi,4)
	fucompp
	fnstsw	%ax
	sahf
	jne		VEQ_FALSE
	movl	$1,C(pr_stack)-24(,%esi,4)
	addl	$-5,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)
VEQ_FALSE:
	movl	$0,C(pr_stack)-24(,%esi,4)
	addl	$-5,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Not equal vectors
	Align16
LOPC_VNE:
	flds	C(pr_stack)-24(,%esi,4)
	flds	C(pr_stack)-12(,%esi,4)
	fucompp
	fnstsw	%ax
	sahf
	jne		LVNE_TRUE
	flds	C(pr_stack)-20(,%esi,4)
	flds	C(pr_stack)-8(,%esi,4)
	fucompp
	fnstsw	%ax
	sahf
	jne		LVNE_TRUE
	flds	C(pr_stack)-16(,%esi,4)
	flds	C(pr_stack)-4(,%esi,4)
	fucompp
	fnstsw	%ax
	sahf
	jne		LVNE_TRUE
	movl	$0,C(pr_stack)-24(,%esi,4)
	addl	$-5,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)
LVNE_TRUE:
	movl	$1,C(pr_stack)-24(,%esi,4)
	addl	$-5,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Vector unary minus
	Align16
LOPC_VUNARYMINUS:
	flds	C(pr_stack)-12(,%esi,4)
	fchs
	fstps	C(pr_stack)-12(,%esi,4)
	flds	C(pr_stack)-8(,%esi,4)
	fchs
	fstps	C(pr_stack)-8(,%esi,4)
	flds	C(pr_stack)-4(,%esi,4)
	fchs
	fstps	C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Drop vector
	Align16
LOPC_VDROP:
	addl	$-3,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Assign vector
	Align16
LOPC_VASSIGN:
	movl	C(pr_stack)-16(,%esi,4),%edx
	movl	C(pr_stack)-12(,%esi,4),%eax
	movl	%eax,(%edx)
	movl	%eax,C(pr_stack)-16(,%esi,4)
	movl	C(pr_stack)-8(,%esi,4),%eax
	movl	%eax,4(%edx)
	movl	%eax,C(pr_stack)-12(,%esi,4)
	movl	C(pr_stack)-4(,%esi,4),%eax
	movl	%eax,8(%edx)
	movl	%eax,C(pr_stack)-8(,%esi,4)
	decl	%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Add vector to vector
	Align16
LOPC_VADDVAR:
	movl	C(pr_stack)-16(,%esi,4),%edx
	flds	(%edx)
	fadds	C(pr_stack)-12(,%esi,4)
	fsts	(%edx)
	fstps	C(pr_stack)-16(,%esi,4)
	flds	4(%edx)
	fadds	C(pr_stack)-8(,%esi,4)
	fsts	4(%edx)
	fstps	C(pr_stack)-12(,%esi,4)
	flds	8(%edx)
	fadds	C(pr_stack)-4(,%esi,4)
	fsts	8(%edx)
	fstps	C(pr_stack)-8(,%esi,4)
	decl	%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Subtract vector from vector
	Align16
LOPC_VSUBVAR:
	movl	C(pr_stack)-16(,%esi,4),%edx
	flds	(%edx)
	fsubs	C(pr_stack)-12(,%esi,4)
	fsts	(%edx)
	fstps	C(pr_stack)-16(,%esi,4)
	flds	4(%edx)
	fsubs	C(pr_stack)-8(,%esi,4)
	fsts	4(%edx)
	fstps	C(pr_stack)-12(,%esi,4)
	flds	8(%edx)
	fsubs	C(pr_stack)-4(,%esi,4)
	fsts	8(%edx)
	fstps	C(pr_stack)-8(,%esi,4)
	decl	%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Scale vector with float
	Align16
LOPC_VSCALEVAR:
	incl	%esi
	movl	C(pr_stack)-12(,%esi,4),%edx
	flds	C(pr_stack)-8(,%esi,4)
	flds	(%edx)
	fmul	%st(1),%st
	fsts	(%edx)
	fstps	C(pr_stack)-12(,%esi,4)
	flds	4(%edx)
	fmul	%st(1),%st
	fsts	4(%edx)
	fstps	C(pr_stack)-8(,%esi,4)
	fmuls	8(%edx)
	fsts	8(%edx)
	fstps	C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Divide vector with float
	Align16
LOPC_VISCALEVAR:
	incl	%esi
	movl	C(pr_stack)-12(,%esi,4),%edx
	flds	C(pr_stack)-8(,%esi,4)
	flds	(%edx)
	fdiv	%st(1),%st
	fsts	(%edx)
	fstps	C(pr_stack)-12(,%esi,4)
	flds	4(%edx)
	fdiv	%st(1),%st
	fsts	4(%edx)
	fstps	C(pr_stack)-8(,%esi,4)
	fdivrs	8(%edx)
	fsts	8(%edx)
	fstps	C(pr_stack)-4(,%esi,4)
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Assign vector, drop result
	Align16
LOPC_VASSIGN_DROP:
	movl	C(pr_stack)-16(,%esi,4),%edx
	movl	C(pr_stack)-12(,%esi,4),%eax
	movl	%eax,(%edx)
	movl	C(pr_stack)-8(,%esi,4),%eax
	movl	%eax,4(%edx)
	movl	C(pr_stack)-4(,%esi,4),%eax
	movl	%eax,8(%edx)
	addl	$-4,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Add vector to vector, drop result
	Align16
LOPC_VADDVAR_DROP:
	movl	C(pr_stack)-16(,%esi,4),%edx
	flds	(%edx)
	fadds	C(pr_stack)-12(,%esi,4)
	fstps	(%edx)
	flds	4(%edx)
	fadds	C(pr_stack)-8(,%esi,4)
	fstps	4(%edx)
	flds	8(%edx)
	fadds	C(pr_stack)-4(,%esi,4)
	fstps	8(%edx)
	addl	$-4,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Subtract vector from vector, drop result
	Align16
LOPC_VSUBVAR_DROP:
	movl	C(pr_stack)-16(,%esi,4),%edx
	flds	(%edx)
	fsubs	C(pr_stack)-12(,%esi,4)
	fstps	(%edx)
	flds	4(%edx)
	fsubs	C(pr_stack)-8(,%esi,4)
	fstps	4(%edx)
	flds	8(%edx)
	fsubs	C(pr_stack)-4(,%esi,4)
	fstps	8(%edx)
	addl	$-4,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Multiply vector with float, drop result
	Align16
LOPC_VSCALEVAR_DROP:
	movl	C(pr_stack)-8(,%esi,4),%edx
	flds	C(pr_stack)-4(,%esi,4)
	flds	(%edx)
	fmul	%st(1),%st
	fstps	(%edx)
	flds	4(%edx)
	fmul	%st(1),%st
	fstps	4(%edx)
	fmuls	8(%edx)
	fstps	8(%edx)
	addl	$-2,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

	//	Divide vector with float, drop result
	Align16
LOPC_VISCALEVAR_DROP:
	movl	C(pr_stack)-8(,%esi,4),%edx
	flds	C(pr_stack)-4(,%esi,4)
	flds	(%edx)
	fdiv	%st(1),%st
	fstps	(%edx)
	flds	4(%edx)
	fdiv	%st(1),%st
	fstps	4(%edx)
	fdivrs	8(%edx)
	fstps	8(%edx)
	addl	$-2,%esi
	//	Go to the next statement
	movl	(%edi),%eax
	addl	$4,%edi
	jmp		*LOPCODE_TABLE(,%eax,4)

//**************************************************************************

    //	End of function
	Align16
LOPC_RETURN:
	movl %esi,C(pr_stackPtr)

LEND_RUN_FUNCTION:
	popl %ebx
	popl %esi
	popl %edi
	addl $256,%esp
	popl %ebp
	ret

//==========================================================================
//
//	sqrt
//
//	On FreeBSD it's not located in libc, so define it here.
//FIXME create a asm file for maths and move it here
//
//==========================================================================

#if !defined DJGPP && !defined __linux__

	.globl	_sqrt
_sqrt:
	fldl	4(%esp)
	fsqrt
	ret

#endif

#endif
