//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "gamedefs.h"
#include "s_local.h"

// MACROS ------------------------------------------------------------------

#define DEFAULT_ARCHIVEPATH		"o:/sound/archive/"

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void S_InitScript(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

boolean				UseSndScript;
char				ArchivePath[128];

sfxinfo_t			*S_sfx;
int					NumSfx;
#ifdef CLIENT

TCvarI				sfx_volume("sfx_volume", "8", CVAR_ARCHIVE);
TCvarI				music_volume("music_volume", "8", CVAR_ARCHIVE);
TCvarI				swap_stereo("swap_stereo", "0", CVAR_ARCHIVE);

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static char			mapSong[12];
static int			mapCDTrack;

static TCvarI		cd_music("use_cd_music", "0", CVAR_ARCHIVE);
static boolean		CDMusic = false;
#endif

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	S_Init
//
// 	Initializes sound stuff, including volume
// 	Sets channels, SFX and music volume,
// allocates channel buffer, sets S_sfx lookup.
//
//==========================================================================

void S_Init(void)
{
#ifdef CLIENT
	S_InitSfx();
	S_InitMusic();
	CD_Init();

#endif
	S_InitScript();
#ifdef CLIENT
	SN_InitSequenceScript();
#endif
}

#ifdef CLIENT

//==========================================================================
//
//	S_Shutdown
//
//	Shuts down all sound stuff
//
//==========================================================================

void S_Shutdown(void)
{
	CD_Shutdown();
	S_ShutdownMusic();
	S_ShutdownSfx();
}

#endif

//==========================================================================
//
//	S_InitScript
//
//	Loads sound script lump or file, if param -devsnd was specified
//
//==========================================================================

static void S_InitScript(void)
{
	int p;
	int i;

    //
    //	Allocate memory for sound info
    //

	S_sfx = (sfxinfo_t*)Z_Malloc(sizeof(sfxinfo_t), PU_STATIC, 0);
	memset(S_sfx, 0, sizeof(sfxinfo_t));
	NumSfx = 1;

    //
    //	Load script SFXINFO
    //

	SC_OpenLump("sfxinfo");

	while (SC_GetString())
	{
		Z_Resize((void**)&S_sfx, (NumSfx + 1) * sizeof(*S_sfx));
		strncpy(S_sfx[NumSfx].tagName, sc_String, 32);

		SC_MustGetString();
		if (*sc_String != '?')
		{
			strcpy(S_sfx[NumSfx].lumpname, sc_String);
		}
        else
        {
			S_sfx[NumSfx].lumpname[0] = 0;
		}

		SC_MustGetNumber();
		S_sfx[NumSfx].priority = sc_Number;

		SC_MustGetNumber();
		S_sfx[NumSfx].numchannels = sc_Number;

		SC_MustGetNumber();
		S_sfx[NumSfx].changePitch = sc_Number;

       	NumSfx++;
	}
	SC_Close();

	//
    //	Load script SNDINFO
    //

	strcpy(ArchivePath, DEFAULT_ARCHIVEPATH);
	p = M_CheckParm("-devsnd");
	if (!p)
	{
		UseSndScript = false;
		SC_OpenLump("sndinfo");
	}
	else
	{
		UseSndScript = true;
		SC_OpenFile(myargv[p + 1]);
	}

	while (SC_GetString())
	{
		if (*sc_String == '$')
		{
			if (!stricmp(sc_String, "$ARCHIVEPATH"))
			{
				SC_MustGetString();
				strcpy(ArchivePath, sc_String);
			}
			else if (!stricmp(sc_String, "$MAP"))
			{
				SC_MustGetNumber();
				SC_MustGetString();
				if(sc_Number)
				{
					P_PutMapSongLump(sc_Number, sc_String);
				}
			}
			continue;
		}
		else
		{
			for (i = 0; i < NumSfx; i++)
			{
				if (!strcmp(S_sfx[i].tagName, sc_String))
				{
					break;
				}
			}
			if (i == NumSfx)
			{
            	//	Not found - add it
            	NumSfx++;
                Z_Resize((void**)&S_sfx, NumSfx * sizeof(*S_sfx));
				strncpy(S_sfx[i].tagName, sc_String, 32);
                //	Default values
                S_sfx[i].priority = 127;
                S_sfx[i].numchannels = -1;
                S_sfx[i].changePitch = 1;
			}

			SC_MustGetString();
			if (*sc_String != '?')
			{
				strcpy(S_sfx[i].lumpname, sc_String);
			}
			else
			{
				strcpy(S_sfx[i].lumpname, "default");
            }
		}
	}
	SC_Close();

	//
    //	Set "default" sound for empty sounds
    //

	for (i = 0; i < NumSfx; i++)
	{
		if (!S_sfx[i].lumpname[0])
		{
			strcpy(S_sfx[i].lumpname, "default");
		}
        S_sfx[i].snd_ptr = NULL;
        S_sfx[i].lumpnum = -1;
	}
}

//==========================================================================
//
// S_GetSoundID
//
//==========================================================================

int S_GetSoundID(char *name)
{
	int i;

	for (i = 0; i < NumSfx; i++)
	{
		if (!strcmp(S_sfx[i].tagName, name))
		{
			return i;
		}
	}
	con << "WARNING! Can't find sound named \"" << name << "\".\n";
	return 0;
}

#ifdef CLIENT

//==========================================================================
//
//	S_StartSong
//
//==========================================================================

void S_StartSong(char* song, int track, boolean loop)
{
	if (CDMusic)
	{
    	if (loop)
			CmdBuf << "CD Loop " << track << "\n";
		else
			CmdBuf << "CD Play " << track << "\n";
	}
	else
	{
    	if (loop)
			CmdBuf << "Music Loop " << song << "\n";
		else
			CmdBuf << "Music Play " << song << "\n";
	}
}

//==========================================================================
//
//	S_PauseSound
//
//==========================================================================

void S_PauseSound(void)
{
	if (CDMusic)
	{
    	CmdBuf << "CD Pause\n";
	}
	else
	{
    	CmdBuf << "Music Pause\n";
	}
}

//==========================================================================
//
//	S_ResumeSound
//
//==========================================================================

void S_ResumeSound(void)
{
	if (CDMusic)
	{
    	CmdBuf << "CD resume\n";
	}
	else
	{
    	CmdBuf << "Music resume\n";
	}
}

//==========================================================================
//
//  StartMusic
//
//==========================================================================

static void StartMusic(void)
{
	S_StartSong(mapSong, mapCDTrack, true);
}

//==========================================================================
//
//	S_Start
//
//	Per level startup code. Kills playing sounds at start of level,
// determines music if any, changes music.
//
//==========================================================================

void S_Start(const mapInfo_t &info)
{
	SN_StopAllSequences();
	S_StopAllSound();

	strcpy(mapSong, info.songLump);
	mapCDTrack = info.cdTrack;

	StartMusic();
}	

//==========================================================================
//
// S_UpdateSounds
//
// Updates music & sounds
//
//==========================================================================

void S_UpdateSounds(void)
{
	if ((int)cd_music && !CDMusic)
    {
    	CmdBuf << "Music Stop\n";
		CDMusic = true;
		StartMusic();
    }
	if (!(int)cd_music && CDMusic)
    {
    	CmdBuf << "CD Stop\n";
		CDMusic = false;
		StartMusic();
    }

	// Update any Sequences
	SN_UpdateActiveSequences();

	S_UpdateSfx();
	S_UpdateMusic();
	CD_Update();
}

#endif

