//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

#ifndef GLVIS_H
#define GLVIS_H

// HEADER FILES ------------------------------------------------------------

#include <math.h>
#include "cmdlib.h"
#include "wadlib.h"

#include "vector.h"
#include "maths.h"

// MACROS ------------------------------------------------------------------

//
// Indicate a leaf.
//
#define	NF_SUBSECTOR	0x8000

// TYPES -------------------------------------------------------------------

//
//  Bounding box
//
enum
{
	BOXTOP,
	BOXBOTTOM,
	BOXLEFT,
	BOXRIGHT
};

typedef TVec vertex_t;

struct line_t : public TPlane
{
    vertex_t	*v1;
    vertex_t	*v2;

    int			flags;

	line_t		*next;
	int			validcount;
};

struct subsector_t
{
    int			numlines;
    int			firstline;
};

struct seg_t : public TPlane
{
    vertex_t	*v1;
    vertex_t	*v2;

    line_t		*linedef;

	seg_t		*partner;
	subsector_t	*subsector;
};

//
// BSP node.
//
struct node_t : public TPlane
{
    // Bounding box for each child.
	float		bbox[2][4];

    // If NF_SUBSECTOR its a subsector.
    word		children[2];
};

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void LoadLevel(int lumpnum, int gl_lumpnum);
void FreeLevel(void);

void BuildPVS(void);

// PUBLIC DATA DECLARATIONS ------------------------------------------------

extern TIWadFile		*mainwad;
extern TIWadFile		*glwad;
extern TOWadFile		outwad;

extern bool				silent_mode;

extern int				numvertexes;
extern vertex_t			*vertexes;

extern int				numsegs;
extern seg_t			*segs;

extern int				numsubsectors;
extern subsector_t		*subsectors;

extern int				numnodes;
extern node_t			*nodes;

extern int				numlines;
extern line_t			*lines;

#endif

