//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "glvis.h"
#include "fmapdefs.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int			numvertexes;
vertex_t	*vertexes;

int			numsegs;
seg_t		*segs;

int			numsubsectors;
subsector_t	*subsectors;

int			numnodes;
node_t		*nodes;

int			numlines;
line_t		*lines;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static vertex_t*		gl_vertexes;

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	LoadVertexes
//
//==========================================================================

static void LoadVertexes(int lump, int gl_lump)
{
    int				i;
	void*			data;
	void*			gldata;
    mapvertex_t*	ml;
    vertex_t*		li;
	int				base_verts;
	int				gl_verts;

	base_verts = mainwad->LumpSize(lump) / sizeof(mapvertex_t);

	gldata = glwad->GetLump(gl_lump);
	outwad.AddLump(glwad->LumpName(gl_lump), gldata, glwad->LumpSize(gl_lump));
	if (!strncmp((char*)gldata, GL_V2_VERTEX_MAGIC, 4))
	{
		gl_verts = (glwad->LumpSize(gl_lump) - 4) / sizeof(gl_mapvertex_t);
	}
	else
	{
		gl_verts = glwad->LumpSize(gl_lump) / sizeof(mapvertex_t);
	}
	numvertexes = base_verts + gl_verts;

    // Allocate zone memory for buffer.
	li = vertexes = new vertex_t[numvertexes];

    // Load data into cache.
	data = mainwad->GetLump(lump);
    ml = (mapvertex_t *)data;

    // Copy and convert vertex, internal representation as vector.
    for (i = 0; i < base_verts; i++, li++, ml++)
    {
		*li = TVec(LittleShort(ml->x), LittleShort(ml->y), 0);
    }

    // Free buffer memory.
    Free(data);

	//	Save pointer to GL vertexes for seg loading
	gl_vertexes = li;

	if (!strncmp((char*)gldata, GL_V2_VERTEX_MAGIC, 4))
	{
	    gl_mapvertex_t*		glml;

	    glml = (gl_mapvertex_t *)((byte*)gldata + 4);

	    // Copy and convert vertex, internal representation as vector.
    	for (i = 0; i < gl_verts; i++, li++, glml++)
	    {
			*li = TVec(FL(LittleLong(glml->x)), FL(LittleLong(glml->y)), 0);
	    }
	}
	else
	{
	    ml = (mapvertex_t *)gldata;

	    // Copy and convert vertex, internal representation as vector.
    	for (i = 0; i < gl_verts; i++, li++, ml++)
	    {
			*li = TVec(LittleShort(ml->x), LittleShort(ml->y), 0);
	    }
	}

   	// Free buffer memory.
    Free(gldata);
}

//==========================================================================
//
//	LoadLineDefs
//
//==========================================================================

template<class maplinedef_t> void LoadLineDefs(int lump)
{
    byte*               data;
    int                 i;
    maplinedef_t		*mld;
    line_t*             ld;

    numlines = mainwad->LumpSize(lump) / sizeof(maplinedef_t);
    lines = new line_t[numlines];
    data = (byte*)mainwad->GetLump(lump);

    mld = (maplinedef_t *)data;
    ld = lines;
    for (i = 0; i < numlines; i++, mld++, ld++)
    {
		ld->flags = LittleShort(mld->flags);
		ld->v1 = &vertexes[LittleShort(mld->v1)];
		ld->v2 = &vertexes[LittleShort(mld->v2)];
		ld->Set2Points(*ld->v1, *ld->v2);
    }

    Free(data);
}

//==========================================================================
//
//	LoadSegs
//
//==========================================================================

static void LoadSegs(int lump)
{
    void*		data;
    int			i;
    mapglseg_t*	ml;
    seg_t*		li;

    numsegs = glwad->LumpSize(lump) / sizeof(mapglseg_t);
    segs = new seg_t[numsegs];
    data = glwad->GetLump(lump);
	outwad.AddLump(glwad->LumpName(lump), data, glwad->LumpSize(lump));

    ml = (mapglseg_t *)data;
    li = segs;

    for (i = 0; i < numsegs; i++, li++, ml++)
    {
		word	v1num =	LittleShort(ml->v1);
		word	v2num =	LittleShort(ml->v2);

		if (v1num & GL_VERTEX)
		{
			v1num ^= GL_VERTEX;
			li->v1 = &gl_vertexes[v1num];
		}
		else
		{
			li->v1 = &vertexes[v1num];
		}
		if (v2num & GL_VERTEX)
		{
			v2num ^= GL_VERTEX;
			li->v2 = &gl_vertexes[v2num];
		}
		else
		{
			li->v2 = &vertexes[v2num];
		}

		int linedef = LittleShort(ml->linedef);
		if (linedef >= 0)
		{
			li->linedef = &lines[linedef];
		}

		int partner = LittleShort(ml->partner);
		if (partner >= 0)
		{
			li->partner = &segs[partner];
		}

		//	Calc seg's plane params
		li->Set2Points(*li->v1, *li->v2);
    }

    Free(data);
}

//==========================================================================
//
//	LoadSubsectors
//
//==========================================================================

static void LoadSubsectors(int lump)
{
    void*				data;
    int					i;
    mapsubsector_t*		ms;
    subsector_t*		ss;

    numsubsectors = glwad->LumpSize(lump) / sizeof(mapsubsector_t);
    subsectors = new subsector_t[numsubsectors];
    data = glwad->GetLump(lump);
	outwad.AddLump(glwad->LumpName(lump), data, glwad->LumpSize(lump));

    ms = (mapsubsector_t *)data;
    ss = subsectors;

    for (i = 0; i < numsubsectors; i++, ss++, ms++)
    {
		ss->numlines = LittleShort(ms->numsegs);
		ss->firstline = LittleShort(ms->firstseg);
    }

    Free(data);
}

//==========================================================================
//
//	LoadNodes
//
//==========================================================================

static void LoadNodes(int lump)
{
	void*		data;
	int         i;
	int         j;
	int			k;
	mapnode_t*	mn;
	node_t*		no;

    numnodes = glwad->LumpSize(lump) / sizeof(mapnode_t);
    nodes = new node_t[numnodes];
    data = glwad->GetLump(lump);
	outwad.AddLump(glwad->LumpName(lump), data, glwad->LumpSize(lump));

    mn = (mapnode_t *)data;
    no = nodes;

    for (i = 0; i < numnodes; i++, no++, mn++)
    {
		no->SetPointDir(TVec(LittleShort(mn->x), LittleShort(mn->y), 0),
			TVec(LittleShort(mn->dx), LittleShort(mn->dy), 0));

		for (j = 0; j < 2; j++)
		{
			no->children[j] = LittleShort(mn->children[j]);
			for (k = 0; k < 4; k++)
			{
				no->bbox[j][k] = LittleShort(mn->bbox[j][k]);
			}
		}
    }
    Free(data);
}

//==========================================================================
//
//	GroupLines
//
//==========================================================================

static void GroupLines(void)
{
	for (int i = 0; i < numsubsectors; i++)
	{
		subsector_t *sub = &subsectors[i];
	    int count = sub->numlines;
    	seg_t *line = &segs[sub->firstline];
	    while (count--)
    	{
			line->subsector = sub;
			line++;
    	}
	}
}

//==========================================================================
//
//	LoadLevel
//
//==========================================================================

void LoadLevel(int lumpnum, int gl_lumpnum)
{
    bool extended = !strcmp(mainwad->LumpName(lumpnum + ML_BEHAVIOR), "BEHAVIOR");

    LoadVertexes(lumpnum + ML_VERTEXES, gl_lumpnum + ML_GL_VERT);
    if (!extended)
    {
		LoadLineDefs<maplinedef1_t>(lumpnum + ML_LINEDEFS);
	}
    else
    {
		LoadLineDefs<maplinedef2_t>(lumpnum + ML_LINEDEFS);
	}
    LoadSegs(gl_lumpnum + ML_GL_SEGS);
    LoadSubsectors(gl_lumpnum + ML_GL_SSECT);
   	LoadNodes(gl_lumpnum + ML_GL_NODES);

    GroupLines();

	cerr << "\nLoaded " << mainwad->LumpName(lumpnum) << ", "
		<< numvertexes << " vertexes, " << numlines << " lines, "
		<< numsegs << " segs, " << numsubsectors << " subsectors, "
		<< numnodes << " nodes\n";
}

//==========================================================================
//
//	FreeLevel
//
//==========================================================================

void FreeLevel(void)
{
	delete vertexes;
	delete lines;
	delete segs;
	delete subsectors;
	delete nodes;
}
