//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include <allegro.h>
#include "vlumpy.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

struct vpic_t
{
	short		width;
	short		height;
	byte		data[4];
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

TOWadFile		outwad;

char			basepath[256];
char			srcpath[256];

PALETTE			ImgPal;
BITMAP			*ImgBmp;
RGB_MAP			rgb_table;

char			lumpname[12];
char			destfile[1024];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	fn
//
//==========================================================================

char *fn(const char *name)
{
	static char filename[1024];
	if (name[0] == '/' || name[0] == '\\' || name[1] == ':')
	{
		//	Absolute path
		strcpy(filename, name);
	}
	else
	{
		sprintf(filename, "%s%s", srcpath, name);
	}
	return filename;
}

//==========================================================================
//
//	AddWadFile
//
//==========================================================================

static void AddWadFile(const char *name)
{
	TIWadFile	inwad;

	inwad.Open(name);
	for (int i = 0; i < inwad.numlumps; i++)
	{
		void *data = inwad.GetLump(i);
		outwad.AddLump(inwad.LumpName(i), data, inwad.LumpSize(i));
		Free(data);
	}
	inwad.Close();
}

//==========================================================================
//
//	AddWad
//
//	$WAD filename
//
//==========================================================================

static void AddWad(const char *name)
{
	char *filename = fn(name);
	DefaultExtension(filename, ".wad");
	AddWadFile(filename);
}

//==========================================================================
//
//	AddMap
//
//	$MAP filename
//
//==========================================================================

static void AddMap(const char *name)
{
	char	*filename;

	filename = fn(name);
	DefaultExtension(filename, ".wad");
	AddWadFile(filename);

	StripExtension(filename);
	strcat(filename, ".gwa");
	FILE *ff = fopen(filename, "rb");
	if (ff)
	{
		fclose(ff);
		AddWadFile(filename);
	}
}

//==========================================================================
//
//	LoadImage
//
//	$LOAD image
//
//==========================================================================

static void LoadImage(void)
{
	SC_MustGetString();
	if (ImgBmp)
	{
		destroy_bitmap(ImgBmp);
	}
	ImgBmp = load_bitmap(fn(sc_String), ImgPal);
    if (!ImgBmp)
    {
    	Error("Can't load image %s\n", sc_String);
    }
	rgb_map = NULL;
}

//==========================================================================
//
//	SetupRGBTable
//
//==========================================================================

void SetupRGBTable(void)
{
	RGB		tmp;

	if (rgb_map)
	{
		return;
	}

	tmp = ImgPal[255];
	ImgPal[255] = ImgPal[0];
	create_rgb_table(&rgb_table, ImgPal, NULL);
	ImgPal[255] = tmp;
	for (int i = 0; i < 32; i++)
	{
		for (int j = 0; j < 32; j++)
		{
			for (int k = 0; k < 32; k++)
			{
				if (rgb_table.data[i][j][k] == 255)
				{
					rgb_table.data[i][j][k] = 0;
				}
			}
		}
	}
	rgb_map = &rgb_table;
}

//==========================================================================
//
//	GrabTranslucencyTable
//
//	lumpname TINTTAB amount
//
//==========================================================================

void GrabTranslucencyTable(void)
{
	byte		table[256 * 256];
	byte		temp[768];
	int			i;
    int			j;
	int			r;
    int			g;
    int			b;
    byte*		p;
    byte*		q;
	int			transluc;

	SC_MustGetNumber();
	transluc = sc_Number;

	SetupRGBTable();

	p = table;
	for (i = 0; i < 256; i++)
    {
    	temp[i * 3]     = ImgPal[i].r * transluc * 255 / 100 / 63;
    	temp[i * 3 + 1] = ImgPal[i].g * transluc * 255 / 100 / 63;
    	temp[i * 3 + 2] = ImgPal[i].b * transluc * 255 / 100 / 63;
    }
	for (i = 0; i < 256; i++)
    {
    	r = ImgPal[i].r * (100 - transluc) * 255 / 100 / 63;
    	g = ImgPal[i].g * (100 - transluc) * 255 / 100 / 63;
    	b = ImgPal[i].b * (100 - transluc) * 255 / 100 / 63;
		q = temp;
        for (j = 0; j < 256; j++)
        {
            *(p++) = makecol8(r + q[0], g + q[1], b + q[2]);
			q += 3;
        }
    }
	outwad.AddLump(lumpname, table, 256 * 256);
}

//==========================================================================
//
//	GrabScaleMap
//
//	lumpname SCALEMAP r g b
//
//==========================================================================

void GrabScaleMap(void)
{
	byte		map[256];
	int			i;
	double		r;
	double		g;
	double		b;

	SC_MustGetFloat();
	r = sc_Float;
	SC_MustGetFloat();
	g = sc_Float;
	SC_MustGetFloat();
	b = sc_Float;

	SetupRGBTable();

	for (i = 0; i < 256; i++)
    {
        double col = (ImgPal[i].r * 0.3 + ImgPal[i].g * 0.5 + ImgPal[i].b * 0.2) * 255.0 / 63.0;
		map[i] = makecol8((int)(r * col), (int)(g * col), (int)(b * col));
    }

	outwad.AddLump(lumpname, map, 256);
}

//==========================================================================
//
//	GrabPic
//
//	lumpname PIC x y width height
//
//==========================================================================

void GrabPic(void)
{
	SC_MustGetNumber();
	int x1 = sc_Number;
	SC_MustGetNumber();
	int y1 = sc_Number;
	SC_MustGetNumber();
	int w = sc_Number;
	SC_MustGetNumber();
	int h = sc_Number;
	int x2 = x1 + w;
	int y2 = y1 + h;

	vpic_t *pic = (vpic_t*)Malloc(4 + w * h);
	pic->width = LittleShort(w);
	pic->height = LittleShort(h);
	byte *dst = pic->data;
	for (int y = y1; y < y2; y++)
	{
		for (int x = x1; x < x2; x++)
		{
			*dst = getpixel(ImgBmp, x, y);
			dst++;
		}
	}

	outwad.AddLump(lumpname, pic, 4 + w * h);
	Free(pic);
}

//==========================================================================
//
//	GrabDecal
//
//	lumpname DECAL x y width height
//
//==========================================================================

void GrabDecal(void)
{
	SC_MustGetNumber();
	int x1 = sc_Number;
	SC_MustGetNumber();
	int y1 = sc_Number;
	SC_MustGetNumber();
	int w = sc_Number;
	SC_MustGetNumber();
	int h = sc_Number;
	int x2 = x1 + w;
	int y2 = y1 + h;

	if ((w & (w - 1)) || (h & (h - 1)))
	{
		SC_ScriptError("Decal size must be a power of 2");
	}

	byte amap[256];
	for (int i = 0; i < 256; i++)
	{
		amap[i] = ~((ImgPal[i].r + ImgPal[i].g + ImgPal[i].b) * 255 / 63 / 3);
	}

	vpic_t *pic = (vpic_t*)Malloc(4 + w * h);
	pic->width = LittleShort(w);
	pic->height = LittleShort(h);
	byte *dst = pic->data;
	for (int y = y1; y < y2; y++)
	{
		for (int x = x1; x < x2; x++)
		{
			*dst = amap[getpixel(ImgBmp, x, y)];
			dst++;
		}
	}

	outwad.AddLump(lumpname, pic, 4 + w * h);
	Free(pic);
}

//==========================================================================
//
//	ParseScript
//
//==========================================================================

void ParseScript(const char *name)
{
	ExtractFilePath(name, basepath);
	strcpy(srcpath, basepath);

	strcpy(destfile, name);
	DefaultExtension(destfile, ".ls");
	SC_Open(destfile);
	StripExtension(destfile);

	bool OutputOpened = false;
	bool GrabMode = false;

	while (SC_GetString())
	{
		if (SC_Compare("$dest"))
		{
			if (OutputOpened)
				SC_ScriptError("Output already opened");
			SC_MustGetString();
			strcpy(destfile, fn(sc_String));
			continue;
		}

		if (SC_Compare("$srcdir"))
		{
			SC_MustGetString();
			sprintf(srcpath, "%s%s", basepath, sc_String);
			if (srcpath[strlen(srcpath) - 1] != '/')
			{
				strcat(srcpath, "/");
			}
			continue;
		}

		if (SC_Compare("$load"))
		{
			LoadImage();
			GrabMode = true;
			continue;
		}

		if (SC_Compare("$files"))
		{
			GrabMode = false;
			continue;
		}

		if (!OutputOpened)
		{
			DefaultExtension(destfile, ".wad");
			outwad.Open(destfile, "PWAD");
			OutputOpened = true;
		}

		if (SC_Compare("$wad"))
		{
			SC_MustGetString();
			AddWad(sc_String);
			continue;
		}

		if (SC_Compare("$map"))
		{
			SC_MustGetString();
			AddMap(sc_String);
			continue;
		}

		if (SC_Compare("$label"))
		{
			SC_MustGetString();
			outwad.AddLump(sc_String, NULL, 0);
			continue;
		}

		if (GrabMode)
		{
			if (strlen(sc_String) > 8)
			{
				SC_ScriptError("Lump name is too long.");
			}
			memset(lumpname, 0, sizeof(lumpname));
			strcpy(lumpname, sc_String);

			SC_MustGetString();
			if (SC_Compare("tinttab"))
			{
				GrabTranslucencyTable();
			}
			else if (SC_Compare("scalemap"))
			{
				GrabScaleMap();
			}
			else if (SC_Compare("pic"))
			{
				GrabPic();
			}
			else if (SC_Compare("decal"))
			{
				GrabDecal();
			}
			else
			{
				SC_ScriptError(va("Unknown command %s", sc_String));
			}
		}
		else
		{
			ExtractFileBase(sc_String, lumpname);
			if (strlen(lumpname) > 8)
			{
				SC_ScriptError("File name too long");
			}
			void *data;
			int size = LoadFile(fn(sc_String), &data);
			outwad.AddLump(lumpname, data, size);
			Free(data);
		}
	}

	if (ImgBmp)
	{
		destroy_bitmap(ImgBmp);
	}

	if (outwad.handle)
	{
		outwad.Close();
	}
	SC_Close();
}

//==========================================================================
//
//	main
//
//==========================================================================

int main(int argc, char *argv[])
{
	if (argc != 2)
	{
		cerr << "Usage: vlumpy <script>\n";
		return 1;
	}

	install_allegro(SYSTEM_NONE, &errno, atexit);
	
	ParseScript(argv[1]);
	return 0;
}
END_OF_MAIN()
